/*
 * DMA on the INCA.
 * Taken from the VxWorks code by Infineon and modified to suit Linux.
 */
#include <linux/config.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/kernel_stat.h>
#include <linux/proc_fs.h>

#include <asm/bootinfo.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <asm/incaip/inca-ip.h>
#include <asm/incaip/irq.h>
#include <asm/incaip/dma.h>

/******************************************
*           debug output macros           *
*******************************************/
typedef enum
{
   DMA_ERR,
   DMA_TRACE,
   DMA_FULL
} DMA_DebugLevel_t;


/* By setting this define the hardware polling in case of a dma hold condition 
   will be enabled. */
/* #define AUTOMATIC_POLLING  */      


#ifndef AUTOMATIC_POLLING
/* When dma polling is disabled it can be configured whether the hold reset 
   command will be set immediately on occurence of hold interrupt (by defining 
   DIRECT_HOLD_RESET) or later on when received data has been processed by software */
#define DIRECT_HOLD_RESET   
#endif

/* With RX_TASK_SCHEDULE it can be configured whether the processing of received 
   data can be done in the ISR or in a scheduled task */

/* #define RX_TASK_SCHEDULE  */ 
#define DMA_USE_PROC

/* #define INCA_DMA_USE_TX_INTERRUPT  */ 

#undef DMA_DEBUG_ENABLE  
/* #define DMA_DEBUG_ENABLE */     
#ifdef DMA_DEBUG_ENABLE
   #define DMA_DEBUG_TX_PRINTOUT
#define  DMA_DEBUG_LEVEL   DMA_ERR  
/* #define  DMA_DEBUG_LEVEL   DMA_FULL  */
/* #define  DMA_DEBUG_LEVEL   DMA_TRACE */  

   #define DMA_DEBUG(level,X)          if(level <= DMA_DEBUG_LEVEL) printk X
   #define DMA_LOG(level,S)		printk(S)
   #define DMA_LOG1(level,S,P1)		printk(S,P1)
   #define DMA_LOG2(level,S,P1,P2)	printk(S,P1,P2)
   #define DMA_LOG3(level,S,P1,P2,P3)	printk(S,P1,P2,P3)
#else
   #define DMA_DEBUG(level,X) 
   #define DMA_LOG(level,S)
   #define DMA_LOG1(level,S,P1)
   #define DMA_LOG2(level,S,P1,P2)
   #define DMA_LOG3(level,S,P1,P2,P3)
#endif

/******************************************
*        static variables                 *
*******************************************/
static IncaDmaTxChannel_t  dmaTxChannel[INCA_DMA_MAX_TX_CHANNEL_ID + 1];
static IncaDmaRxChannel_t  dmaRxChannel[INCA_DMA_MAX_RX_CHANNEL_ID + 1];
static int                 bDMAinit = 0;
#ifdef INCA_DMA_USE_TX_INTERRUPT
static unsigned long       txIrqEnabled = 0; 
#endif
static unsigned long       rxIrqEnabled = 0;

#ifdef RX_TASK_SCHEDULE
static volatile unsigned int    RDMA_HANDLING_FLAG = 0;

#define RDMA0_HANDLING	1
#define RDMA1_HANDLING  2
#endif

#ifndef DIRECT_HOLD_RESET
/* flags to indicate the occurence of an hold interrupt on channel 0 or 1 */
static volatile unsigned int hold_reset_occured[INCA_DMA_MAX_RX_CHANNEL_ID + 1] = { 0, 0 };
#endif

/******************************************
*        local functions and macros       *
*******************************************/
static void incaDMA_RxIsr(int irq, void *dev_id, struct pt_regs *regs);
#ifdef INCA_DMA_USE_TX_INTERRUPT 
static void incaDMA_TxIsr(int irq, void *dev_id, struct pt_regs *regs);
#endif

static void incaDMA_UpdateHOLD(unsigned long channelId);
static void incaDMA_Receive(unsigned long channelId);
static void incaDMA_HandleEop(int ch);

/* to be removed after testing */
#ifdef DMA_DEBUG_ENABLE
void incaDMA_PrintAllInfoRx(unsigned long channelId);
void incaDMA_PrintAllInfoTx(unsigned long channelId);
#endif
#ifdef DMA_USE_PROC
static int inca_dma_read_procmem(char *buf, char **start, off_t offset, 
                                 int count, int *eof, void *data);
static int inca_dma_write_procmem(struct file *file, const char *buffer,
                                  unsigned long count, void *data);
#endif

#define DMA_WRITE_REG(reg, value)   *((volatile unsigned long*)reg) = (unsigned long)value;
#define DMA_READ_REG(reg, value)    value = (unsigned long)*((volatile unsigned long*)reg);

#ifdef RX_TASK_SCHEDULE
DECLARE_TASKLET(incaDMA_Rx_Tasklet0, incaDMA_Receive, 0 );
DECLARE_TASKLET(incaDMA_Rx_Tasklet1, incaDMA_Receive, 1 );
#endif

/****************************************************************************
Description:
   This function initializes registers of the DMA and connects interrupt 
   handlers.
Arguments:
   None
Return Value:
   0   - Initialization succesful.
   <>0 - Initialization failed. See request_irq() for details.
Remarks:
   None
****************************************************************************/
int incaDMA_Init (void)
{
	int retval;

	if( !bDMAinit )
	{
		/* switch off all DMA channels */
		DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR0, INCA_IP_DMA_DMA_RXCCR0_OFF);
		DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR1, INCA_IP_DMA_DMA_RXCCR1_OFF);

		DMA_WRITE_REG (INCA_IP_DMA_DMA_TXCCR0, INCA_IP_DMA_DMA_RXCCR0_OFF);
		DMA_WRITE_REG (INCA_IP_DMA_DMA_TXCCR1, INCA_IP_DMA_DMA_TXCCR1_OFF);
		DMA_WRITE_REG (INCA_IP_DMA_DMA_TXCCR2, INCA_IP_DMA_DMA_TXCCR2_OFF);

		/* setup TX channel polling time */
		DMA_WRITE_REG (INCA_IP_DMA_DMA_TXPOLL, INCA_DMA_TX_POLLING_TIME);
		/* setup RX channel polling time */
		DMA_WRITE_REG (INCA_IP_DMA_DMA_RXPOLL, INCA_DMA_RX_POLLING_TIME);

		/* ERRATA: write reset value into the DMA RX IMR register */
		DMA_WRITE_REG (INCA_IP_DMA_DMA_RXIMR, 0xFFFFFFFF);

		/* install interrupt handlers */
#ifdef INCA_DMA_USE_TX_INTERRUPT
		retval = request_irq(INT_NUM_IM1_IRL27, incaDMA_TxIsr, SA_INTERRUPT, "dma_tx", NULL);
		if (retval)
			return retval;
#endif		
/*		retval = request_irq(INT_NUM_IM1_IRL28, incaDMA_RxIsr, SA_INTERRUPT, "dma_rx", NULL); */
        retval = request_irq(INT_NUM_IM1_IRL28, incaDMA_RxIsr, 0, "dma_rx", NULL);
		if (retval)
			return retval;
		bDMAinit = 1;
		DMA_DEBUG(DMA_TRACE,("incaDMA_Init: DMA initialized successfully\n"));
	}
#ifdef DMA_USE_PROC
        {
            struct proc_dir_entry *inca_dma_proc;
            const char *dev_name = "driver/dma";

            inca_dma_proc = create_proc_entry(dev_name, 0, 0);
            if (!inca_dma_proc) {
                printk(KERN_ERR "could not create /proc/%s!\n", dev_name);
            }
            else
            {
                inca_dma_proc->read_proc  = inca_dma_read_procmem;
                inca_dma_proc->write_proc = inca_dma_write_procmem;
#ifdef DMA_DEBUG_ENABLE
                printk(KERN_NOTICE "/proc/%s created!\n", dev_name);
#endif
            }
        }
#endif

	return 0;
}
EXPORT_SYMBOL(incaDMA_Init);


/* need these for kfree() in the close routine */
static unsigned long *KMRxfirstDescriptorAddress[INCA_DMA_MAX_RX_CHANNEL_ID];
#ifdef INCA_DMA_USE_COPYING
static unsigned char *KMRxBuffers[INCA_DMA_MAX_RX_CHANNEL_ID];
#endif

/****************************************************************************
Description:
   This funciton initializes specified DMA Rx Channel
Arguments:
   channelId      - identifier if the DMA Rx channel (0,1)
   descNum        - number of descriptors for specified channel
   bufSize        - size of data buffer for each descriptor
   device         - pointer to device descriptor
   getNewBuffer   - callback function: it's used to obtain new data buffer
                     for currently processed descriptor
   putRcvdBuffer  - callback function: it's used to pass received data to the 
                     upper layer (for example, to the switch END driver) 
Return Value:
   0  - initialization of the channel is successfully completed
   -1 - there are some error during channel initialization
Remarks:
   None
****************************************************************************/
long incaDMA_InitRxChannel ( unsigned long channelId, unsigned long descNum,
                             unsigned long bufSize,   void *device, 
                             unsigned long (*getNewBuffer)(void *device, unsigned long *info),
                             unsigned long (*putRcvdBuffer)( void *device,
                                                             void *buf, 
                                                             unsigned long bufSize, 
                                                             unsigned long info) )
{
   register unsigned long  regValue;
   unsigned long           *virtFirstDescriptorAddress, *physFirstDescriptorAddress;
   IncaDmaRxDescriptor_t   *virtRxDescriptor;
   IncaDmaRxDescriptor_t   *NCrxDescriptor;
   unsigned long           i;
   unsigned int            flags=0;
#ifdef INCA_DMA_USE_COPYING
   unsigned char           *RxBuffers;
#else
   unsigned long           virtRxDataPtr;
#endif

   /* check parameters */
   if( channelId > INCA_DMA_MAX_RX_CHANNEL_ID ||
       descNum == 0 || 
       bufSize == 0 ||
       getNewBuffer == NULL || 
       putRcvdBuffer == NULL || 
       device == NULL )
   {
      DMA_DEBUG(DMA_ERR,("incaDMA_InitRxChannel: wrong parameters\n"));
      return -1;
   }

   /* init Rx channel structure */
   dmaRxChannel[channelId].descriptorNumber  = descNum;
   dmaRxChannel[channelId].device            = device;
   dmaRxChannel[channelId].bufferSize        = bufSize;
   dmaRxChannel[channelId].getNewBuffer      = getNewBuffer;
   dmaRxChannel[channelId].putRcvdBuffer     = putRcvdBuffer;

   /* initialize Rx descriptors' list */
#ifdef INCA_DMA_USE_COPYING
   KMRxBuffers[channelId] = (unsigned char *)KSEG1ADDR( kmalloc((descNum * bufSize) + 16, GFP_KERNEL) );
   if (KMRxBuffers[channelId] == (unsigned char *) KSEG1ADDR(NULL))
   	return -1;

   /* We don't care about the previous contents, but we don't want */
   /* the cache to write back data at a later time, thus destroying */
   /* receive data (non-relevant for mips4k with write-through cache) */
   dma_cache_inv((unsigned long)KMRxBuffers[channelId], 
                 (descNum * bufSize) + 16);

   RxBuffers = KMRxBuffers[channelId];
   if ((unsigned long)RxBuffers & 0xf)
   	RxBuffers = (unsigned char *)(((unsigned long)RxBuffers + 16) & 0xfffffff0);
#ifndef CONFIG_MIPS_UNCACHED
   RxBuffers = (unsigned char *)ioremap_nocache( virt_to_phys(RxBuffers), descNum * bufSize);
#endif
#endif

   if((KMRxfirstDescriptorAddress[channelId] =  
      (unsigned long*)KSEG1ADDR(kmalloc(descNum * (sizeof(IncaDmaRxDescriptor_t))+16, GFP_KERNEL))) 
                                             == (unsigned long*)KSEG1ADDR(NULL))
   {
      DMA_DEBUG(DMA_ERR,("incaDMA_InitRxChannel: cannot allocate memory for descriptors\n"));
      return -1;
   } 

   dma_cache_inv((unsigned long)KMRxfirstDescriptorAddress[channelId], 
                 descNum * (sizeof(IncaDmaRxDescriptor_t))+16);

   virtFirstDescriptorAddress = KMRxfirstDescriptorAddress[channelId];
   /* Make 16 bytes alingment of descriptor addresses */
   if((unsigned long)virtFirstDescriptorAddress & 0xF)
   {
      virtFirstDescriptorAddress += 4;
      virtFirstDescriptorAddress = (unsigned long*)((unsigned long)virtFirstDescriptorAddress & 0xFFFFFFF0);
   }

#ifndef CONFIG_MIPS_UNCACHED
   virtFirstDescriptorAddress = (unsigned long*)
       ioremap_nocache(virt_to_phys(virtFirstDescriptorAddress), 
                       descNum * sizeof(IncaDmaRxDescriptor_t));
#endif
   dmaRxChannel[channelId].virtFirstDescriptorAddr = virtFirstDescriptorAddress;
   physFirstDescriptorAddress = (unsigned long*)virt_to_phys(virtFirstDescriptorAddress);

   /* clear memory area */
#ifdef INCA_DMA_USE_COPYING
   memset(RxBuffers, 0, descNum * bufSize);
#endif
   memset(virtFirstDescriptorAddress, 0, descNum * (sizeof(IncaDmaRxDescriptor_t)));

   for(i = 0, virtRxDescriptor = (IncaDmaRxDescriptor_t *)virtFirstDescriptorAddress;
       i < descNum;
       i++, virtRxDescriptor++)
   {

      NCrxDescriptor = (IncaDmaRxDescriptor_t *)KSEG1ADDR(virtRxDescriptor);
      /* set maximum size of receive buffer */
      NCrxDescriptor->params.field.NFB =  dmaRxChannel[channelId].bufferSize /* TODO: use correct value */;

      /* set the offset of the receive buffer */
      /* offset mechanism is not used when copying data */
      NCrxDescriptor->params.field.offset = 0;

      /* check if it is the last descriptor */
      if( i == (descNum - 1) )
      {
         /* let the last descriptor point to the first one */
         NCrxDescriptor->virtNextRxDescPtr = (unsigned long)virtFirstDescriptorAddress;
         NCrxDescriptor->physNextRxDescPtr = (unsigned long)physFirstDescriptorAddress;
      }
      else
      {
         /* set the address of the next descriptor */
         NCrxDescriptor->virtNextRxDescPtr = (unsigned long)(virtRxDescriptor + 1);
         NCrxDescriptor->physNextRxDescPtr = (unsigned long)virt_to_phys(virtRxDescriptor + 1);
      }

#ifdef INCA_DMA_USE_COPYING
      NCrxDescriptor->virtRxDataPtr = (unsigned long)RxBuffers + i * bufSize;
      NCrxDescriptor->physRxDataPtr = (unsigned long)virt_to_phys((void*)(NCrxDescriptor->virtRxDataPtr));
#else
      /* Allocate memory for data buffers */
      virtRxDataPtr = getNewBuffer(device, &NCrxDescriptor->info);
      if(virtRxDataPtr == 0)
      {
         DMA_DEBUG(DMA_ERR,("incaDMA_InitRxChannel: cannot allocate buffer for descriptor\n"));
         return -1;
      }

      NCrxDescriptor->params.field.offset = virtRxDataPtr & 3;

      if (virtRxDataPtr & 3)
      {
          virtRxDataPtr &= ~3;
      }
      NCrxDescriptor->virtRxDataPtr = virtRxDataPtr;
      NCrxDescriptor->physRxDataPtr = (unsigned long)virt_to_phys((void*)virtRxDataPtr);
#endif

      NCrxDescriptor->params.field.ICpt = 1;
      NCrxDescriptor->params.field.IEop = 0;
      NCrxDescriptor->params.field.HOLD = 0;
   }

   /* set HOLD bit in last decriptor.
      and store last descriptor in the channel object */
   dmaRxChannel[channelId].virtHoldDescriptor = (virtRxDescriptor - 1);
   dmaRxChannel[channelId].virtHoldDescriptor->params.field.HOLD = 1;
   dmaRxChannel[channelId].virtFirstDescriptor =
       (IncaDmaRxDescriptor_t*)virtFirstDescriptorAddress;
   dmaRxChannel[channelId].virtCurrentDescriptor =
       (IncaDmaRxDescriptor_t*)virtFirstDescriptorAddress;

   save_and_cli(flags);

   DMA_READ_REG (INCA_IP_DMA_DMA_RXIMR, regValue);
   
   /* 
    * enable receive interrupts for polling/no polling and 
    * debug/no debug modes 
    */   
#ifndef AUTOMATIC_POLLING
   /* dma polling timer disabled -> 
    * enable hold interrupt and process by software 
    */ 
#ifdef DMA_DEBUG_ENABLE
   /* enable the EOP, HOLD, CMDCPT, CPT and RDERR interrupts */
   regValue &= ~( INCA_IP_DMA_RXIMR_CMDCPT(channelId) |
                  INCA_IP_DMA_RXIMR_CPT(channelId)    |
                  INCA_IP_DMA_RXIMR_RDERR(channelId)  |
                  INCA_IP_DMA_RXIMR_EOP(channelId)    |
                  INCA_IP_DMA_RXIMR_CPT(channelId)    |
                  INCA_IP_DMA_RXIMR_HLD(channelId)  );
#else
   /* enable the completion and hold interrupt  */  
   regValue &= ~( INCA_IP_DMA_RXIMR_CPT(channelId)    |
                  INCA_IP_DMA_RXIMR_HLD(channelId) );
#endif 
#else 
   /* use timer controlled polling -> no hold interrupt necessary */
#ifdef DMA_DEBUG_ENABLE
   /* enable the EOP, CMDCPT, CPT and RDERR interrupts */
   regValue &= ~( INCA_IP_DMA_RXIMR_CMDCPT(channelId) |
                  INCA_IP_DMA_RXIMR_CPT(channelId)    |
                  INCA_IP_DMA_RXIMR_RDERR(channelId)  |
                  INCA_IP_DMA_RXIMR_EOP(channelId)    |
                  INCA_IP_DMA_RXIMR_CPT(channelId) );
#else
   /* enable the completion interrupt  */
   regValue &= ~( INCA_IP_DMA_RXIMR_CPT(channelId) );
#endif     
#endif   
   DMA_WRITE_REG (INCA_IP_DMA_DMA_RXIMR, regValue);
   restore_flags(flags);
   
   /*
    * Set the first receive descriptor address, 
    * enable the hold poll timer and start initialization of DMA channel
    */
   switch(channelId)
   {
      case 0:
         DMA_WRITE_REG (INCA_IP_DMA_DMA_RXFRDA0, physFirstDescriptorAddress);
         DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR0, INCA_IP_DMA_DMA_RXCCR0_INIT );
/* #ifdef AUTOMATIC_POLLING  */
         /* 
          * For any unknown reason the hold poll timer has to be enabled 
          * although the hold reset command is given by software !!!
          */  		 
         DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR0, INCA_IP_DMA_DMA_RXCCR0_INIT | INCA_IP_DMA_DMA_RXCCR0_HPEN);  
/* #endif  */
         break;

      case 1:
         DMA_WRITE_REG (INCA_IP_DMA_DMA_RXFRDA1, physFirstDescriptorAddress);
         DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR1, INCA_IP_DMA_DMA_RXCCR1_INIT );
/* #ifdef AUTOMATIC_POLLING  */  			 
         /* 
          * For any unknown reason the hold poll timer has to be enabled 
          * although the hold reset command is given by software !!!
          */  		 
         DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR1, INCA_IP_DMA_DMA_RXCCR1_INIT | INCA_IP_DMA_DMA_RXCCR1_HPEN);  
/* #endif  */
         break;
      
      default: 
         break;
   }

   /* enable interrupts */
   if(rxIrqEnabled == 0)
   {
      /* enable interrupts */
      DMA_INTERRUPTS_CLEAR(INCA_IRQ_LINE_DMARIR);
      DMA_INTERRUPTS_ENABLE(INCA_IRQ_LINE_DMARIR);
   }
   rxIrqEnabled++;

   DMA_DEBUG(DMA_TRACE,("incaDMA_InitRxChannel: Rx channel %ld initialized\n",channelId));
   DMA_DEBUG(DMA_TRACE,("                       Number of Descriptors: %ld\n",descNum));

   return 0;
}
EXPORT_SYMBOL(incaDMA_InitRxChannel);

/* need this for kfree() in the close routine */
static unsigned long *KMTxfirstDescriptorAddress[INCA_DMA_MAX_TX_CHANNEL_ID];

/****************************************************************************
Description:
   This funciton initializes specified DMA Tx Channel
Arguments:
   channelId      - identifier of the DMA Tx channel (0,1,2)
   descNum        - number of descriptors for specified channel
   priority       - priority of the specified channel
   device         - pointer to device descriptor
   txComplete     - callback function: it's used to inform upper layer
                     about completed transmission
Return Value:
   0  - initialization of the channel is successfully completed
   -1 - there are some error during channel initialization
Remarks:
   None
****************************************************************************/
long incaDMA_InitTxChannel ( unsigned long channelId, unsigned long descNum, 
                             unsigned long priority,  void *device, 
                             unsigned long (*txComplete)(void* device, unsigned long info) )
{
   register unsigned long  regValue;
   unsigned long           i, memsize;
   unsigned long           *virtFirstDescriptorAddress, *physFirstDescriptorAddress;
   IncaDmaTxDescriptor_t   *virtTxDescriptor;
   IncaDmaTxDescriptor_t   *NCtxDescriptor;
   unsigned int            flags=0;


#ifdef DMA_DEBUG_ENABLE
   printk("incaDMA_InitTxChannel: \n");
#endif
   /* check parameters */
   if(   channelId > INCA_DMA_MAX_TX_CHANNEL_ID || 
         priority > INCA_DMA_MAX_TX_PRIORITY    ||
         device == NULL                         ||
         txComplete == NULL                     ||
         descNum == 0   )
   {
      DMA_DEBUG(DMA_ERR,("incaDMA_InitTxChannel: wrong parameters\n"));
      return -1;
   }

   /* init Tx channel structure */
   dmaTxChannel[channelId].descriptorNumber  = descNum;
   dmaTxChannel[channelId].device            = device;
   dmaTxChannel[channelId].txComplete        = txComplete;

   /* initialize Tx descriptors' list */
#ifdef INCA_DMA_USE_COPYING
   memsize = descNum * (sizeof(IncaDmaTxDescriptor_t) + 1520)+16;
#else
   memsize = descNum * sizeof(IncaDmaTxDescriptor_t) + 16;
#endif

   if((KMTxfirstDescriptorAddress[channelId] =
       (unsigned long*)KSEG1ADDR( kmalloc(memsize, GFP_KERNEL)) ) == (unsigned long*)KSEG1ADDR(NULL))
   {
      DMA_DEBUG(DMA_ERR,("incaDMA_InitTxChannel: cannot allocate memory for descriptors\n"));
      return -1;
   }

   /* we don't want the cache to be flushed later, so invalidate it */
   dma_cache_inv((unsigned long)KMTxfirstDescriptorAddress[channelId], memsize);

   virtFirstDescriptorAddress = KMTxfirstDescriptorAddress[channelId];
   /* Make 16 bytes alingment of descriptor addresses */
   if((unsigned long)virtFirstDescriptorAddress & 0xF)
   {
       memsize -= ((unsigned long)virtFirstDescriptorAddress & 0xF);
      virtFirstDescriptorAddress += 4;
      virtFirstDescriptorAddress = (unsigned long*)((unsigned long)virtFirstDescriptorAddress & 0xFFFFFFF0);
   }
#ifndef CONFIG_MIPS_UNCACHED
   virtFirstDescriptorAddress = (unsigned long*)
       ioremap_nocache(virt_to_phys(virtFirstDescriptorAddress), memsize);
#endif
   physFirstDescriptorAddress = (unsigned long*)virt_to_phys(virtFirstDescriptorAddress);
   dmaTxChannel[channelId].virtFirstDescriptorAddr = virtFirstDescriptorAddress;
   dmaTxChannel[channelId].virtCleanupDescriptor   = 
       (IncaDmaTxDescriptor_t *)virtFirstDescriptorAddress;

   /* clear memory area */
   memset(virtFirstDescriptorAddress, 0, descNum * sizeof(IncaDmaTxDescriptor_t));

   /* Build a linked list of  Tx descriptors */
   for(i = 0, virtTxDescriptor = (IncaDmaTxDescriptor_t*)virtFirstDescriptorAddress;
       i < descNum;
       i++, virtTxDescriptor++)
   {
      NCtxDescriptor = (IncaDmaTxDescriptor_t *)KSEG1ADDR(virtTxDescriptor);

      NCtxDescriptor->params.dword       = 0;
      NCtxDescriptor->params.field.HOLD  = 1; /* next descriptor not available */
      NCtxDescriptor->C                  = 1; /* to be set by DMA hardware  */

#ifdef INCA_DMA_USE_COPYING
      NCtxDescriptor->virtTxDataPtr = ((unsigned long)virtFirstDescriptorAddress + 
                                   (descNum * sizeof(IncaDmaTxDescriptor_t))+
                                   (1520*i));
      NCtxDescriptor->physTxDataPtr = ((unsigned long)physFirstDescriptorAddress + 
                                   (descNum * sizeof(IncaDmaTxDescriptor_t))+
                                   (1520*i));
#endif      

      /* check if it is the last descriptor */
      if( i == (descNum - 1) )
      {
         /* let the last descriptor point to the first one */
          NCtxDescriptor->virtNextTxDescPtr = (unsigned long)virtFirstDescriptorAddress;
          NCtxDescriptor->physNextTxDescPtr = (unsigned long)physFirstDescriptorAddress;
      }
      else
      {
          /* set the address of the next descriptor */
          NCtxDescriptor->virtNextTxDescPtr = (unsigned long)(virtTxDescriptor + 1);
          NCtxDescriptor->physNextTxDescPtr = (unsigned long)virt_to_phys(virtTxDescriptor + 1);
      }
   }

   /* Store the current hold descriptor in the channel object */
   dmaTxChannel[channelId].virtHoldDescriptor = virtTxDescriptor - 1;

   save_and_cli(flags);
   /* Writing a priority of the channel (0-3)  */
   DMA_READ_REG (INCA_IP_DMA_DMA_TXWGT, regValue);
   regValue |= INCA_IP_DMA_TXWGT_TXPRI(channelId,priority);
   DMA_WRITE_REG (INCA_IP_DMA_DMA_TXWGT, regValue);

   /* Writing to the MASK register: unmask the HLD, CMDCPT, CPT
                                    mask the EOP, RDERR */
   DMA_READ_REG (INCA_IP_DMA_DMA_TXIMR, regValue);

#ifdef DMA_DEBUG_ENABLE
   /* enable the EOP, HLD, CMDCPT, CPT and RDERR interrupts */      
   regValue &= ~( INCA_IP_DMA_TXIMR_HLD(channelId)    |    
                  INCA_IP_DMA_TXIMR_CMDCPT(channelId) |
                  INCA_IP_DMA_TXIMR_CPT(channelId)    | 
                  INCA_IP_DMA_TXIMR_EOP(channelId)    | 
                  INCA_IP_DMA_TXISR_RDERR(channelId)  );
#else
  /* enable the end of packet (EOP) interrupt  */        
   regValue &= ~( INCA_IP_DMA_TXIMR_EOP(channelId) );
#endif      

   restore_flags(flags);   
#ifdef INCA_DMA_USE_TX_INTERRUPT 
   if(txIrqEnabled == 0)
   {
      /* enable interrupts */
      DMA_INTERRUPTS_CLEAR(INCA_IRQ_LINE_DMATIR);
      DMA_INTERRUPTS_ENABLE(INCA_IRQ_LINE_DMATIR);
   }
   txIrqEnabled++;
#endif
   /* write first receive descriptor address to appr. register */
   switch(channelId)
   {
      case 0:
         /* Writing to the FRDA of CHANNEL */
         DMA_WRITE_REG (INCA_IP_DMA_DMA_TXFRDA0, physFirstDescriptorAddress);
         break;

      case 1:
         /* Writing to the FRDA of CHANNEL */
         DMA_WRITE_REG (INCA_IP_DMA_DMA_TXFRDA1, physFirstDescriptorAddress);
         break;

      case 2:
         /* Writing to the FRDA of CHANNEL */
         DMA_WRITE_REG (INCA_IP_DMA_DMA_TXFRDA2, physFirstDescriptorAddress);
         break;

      default: 
         break;
   }

   DMA_DEBUG(DMA_TRACE,("incaDMA_InitTxChannel: Tx channel %ld initialized\n",channelId));
   DMA_DEBUG(DMA_TRACE,("                       Number of Descriptors: %ld\n",descNum));
   DMA_DEBUG(DMA_TRACE,("                       Priority: %ld\n",priority));

   return 0;
}
EXPORT_SYMBOL(incaDMA_InitTxChannel);


/****************************************************************************
Description:
   Interrupt service routine to handle DMA_RX interrupt
Arguments:
   irq     -
   *dev_id -
   *regs   -
Return Value:
   None
Remarks:
   None
****************************************************************************/
void incaDMA_RxIsr(int irq, void *dev_id, struct pt_regs *regs)
{       
    register unsigned long regValue;
#ifdef RX_TASK_SCHEDULE	
	register unsigned long imr_value;
#endif
    unsigned int flags = 0;

    save_and_cli(flags);

    /* Check DMA_RXISR register */
    DMA_READ_REG( INCA_IP_DMA_DMA_RXISR, regValue );

    DMA_LOG1(DMA_TRACE,"incaDMA_RxIsr: (0x%08lX)\n",regValue);

#ifdef RX_TASK_SCHEDULE
    if( (regValue & INCA_IP_DMA_RXISR_CPT_0) && 
       !(RDMA_HANDLING_FLAG & RDMA0_HANDLING) )       
    {
        /* process End Of Packet interrupt on channel 0 */
        DMA_LOG(DMA_TRACE,"  CPT (0)\n");
        /* mask CPT 0 interrupt */           
        DMA_READ_REG (INCA_IP_DMA_DMA_RXIMR, imr_value);
        imr_value |= INCA_IP_DMA_RXIMR_CPT(0);
        DMA_WRITE_REG (INCA_IP_DMA_DMA_RXIMR, imr_value);           
        
		tasklet_schedule( &incaDMA_Rx_Tasklet0);
        RDMA_HANDLING_FLAG |= RDMA0_HANDLING;
    }
#else
    if( regValue & INCA_IP_DMA_RXISR_CPT_0 )	
    {
        /* process completion interrupt on channel 0 */
        DMA_LOG(DMA_TRACE,"  CPT (0)\n");
        incaDMA_Receive(0);
    }
#endif	

#ifdef RX_TASK_SCHEDULE
    if( (regValue & INCA_IP_DMA_RXISR_CPT_1) && 
       !(RDMA_HANDLING_FLAG & RDMA1_HANDLING) )       
    {
        /* process End Of Packet interrupt on channel 0 */
        DMA_LOG(DMA_TRACE,"  CPT (1)\n");
        /* mask CPT1 interrupt */
        DMA_READ_REG (INCA_IP_DMA_DMA_RXIMR, imr_value);
        imr_value |= INCA_IP_DMA_RXIMR_CPT(1);
        DMA_WRITE_REG (INCA_IP_DMA_DMA_RXIMR, imr_value); 
        tasklet_schedule(&incaDMA_Rx_Tasklet1);
        RDMA_HANDLING_FLAG |= RDMA1_HANDLING;           
    }
#else
    if( (regValue & INCA_IP_DMA_RXISR_CPT_1) )
    {
        /* process completion interrupt on channel 1 */
        DMA_LOG(DMA_TRACE,"  CPT (1)\n");
        incaDMA_Receive(1);
    }
#endif
    if(regValue & INCA_IP_DMA_RXISR_HLD_0)
    {
        /* process hold interrupt on channel 0 */
        DMA_LOG(DMA_TRACE,"  HOLD (0)\n");
	incaDMA_UpdateHOLD(0);
    }
    if(regValue & INCA_IP_DMA_RXISR_HLD_1)
    {
       /* process hold interrupt on channel 1 */
        DMA_LOG(DMA_TRACE,"  HOLD (1)\n");
        incaDMA_UpdateHOLD(1);
    }

#ifdef DMA_DEBUG_ENABLE
    if(regValue & INCA_IP_DMA_RXISR_RDERR_0)
    {
        /* S-Bus Read Error */
        /* Do nothing: just collect statistics ??? */
        DMA_LOG(DMA_TRACE,"  RDERR (0)\n");
    }
    if(regValue & INCA_IP_DMA_RXISR_RDERR_1)
    {
        /* S-Bus Read Error */
        DMA_LOG(DMA_TRACE,"  RDERR (1)\n");
    }

    if(regValue & INCA_IP_DMA_RXISR_CMDCPT_0)
    {
        /* Command Complete */
        /* Do nothing: just collect statistics ??? */
        DMA_LOG(DMA_TRACE,"  CMDCPT (0)\n");
    }
    if(regValue & INCA_IP_DMA_RXISR_CMDCPT_1)
    {
        /* Command Complete */
        /* Do nothing: just collect statistics ??? */
        DMA_LOG(DMA_TRACE,"  CMDCPT (1)\n");
    }

    if(regValue & INCA_IP_DMA_RXISR_EOP_0)
    {
        /* process End Of Packet interrupt on channel 0 */
        DMA_LOG(DMA_TRACE,"  EOP (0)\n");
    }

    if(regValue & INCA_IP_DMA_RXISR_EOP_1)
    {
        /* process End Of Packet interrupt on channel 1 */
        DMA_LOG(DMA_TRACE,"  EOP (1)\n");
    }

#endif /* DMA_DEBUG_ENABLE */

    /* Acknowledge interrupts */
    DMA_WRITE_REG(INCA_IP_DMA_DMA_RXISR,regValue);
    restore_flags(flags);

    return;
}

#ifdef INCA_DMA_USE_TX_INTERRUPT
/****************************************************************************
Description:
Arguments:
Return Value:
Remarks:
   None
****************************************************************************/
static void incaDMA_TxIsr(int irq, void *dev_id, struct pt_regs *regs)
{
   register unsigned long regValue, regValue_imr;

   /* Acknowledge the interrupt */
   /*DMA_INTERRUPTS_CLEAR(INCA_IRQ_LINE_DMATIR);*/
   /* Check DMA_TXISR register */

   DMA_READ_REG(INCA_IP_DMA_DMA_TXIMR,regValue_imr);
/*
   printk("INCA_IP_DMA_DMA_TXIMR: 0x%08lX\n",regValue_imr);             
*/
   do {
   
      DMA_READ_REG(INCA_IP_DMA_DMA_TXISR,regValue);
/*
      printk("incaDMA_TxIsr: (0x%08lX)\n",regValue);    
*/   
      DMA_LOG1(DMA_TRACE,"incaDMA_TxIsr: (0x%08lX)\n",regValue);
   
      regValue &= ~regValue_imr;
      /* acknowledge interrupt */
      DMA_WRITE_REG(INCA_IP_DMA_DMA_TXISR,regValue); 
/*    DMA_WRITE_REG(INCA_IP_DMA_DMA_TXISR,0xFFFFFFFF);  */

      if(regValue == 0)
         return;

      /* What kind of interrupt we've got */
      if(regValue & INCA_IP_DMA_TXISR_EOP_0)
      {
         /* End Of Packet */
         /* Packet is transmitted so transmit next one */
      
/*       printk("  EOP (0)\n");  */
         DMA_LOG(DMA_TRACE,"  EOP (0)\n");
         if(dmaTxChannel[0].txComplete)
         {
            DMA_LOG1(DMA_TRACE,"     func (0x%08X)\n",(int)dmaTxChannel[0].txComplete);
            incaDMA_HandleEop(0);
/*           dmaTxChannel[0].txComplete(dmaTxChannel[0].device, 0); */
         }
         dmaTxChannel[0].txAckCounter ++;
      }

      if(regValue & INCA_IP_DMA_TXISR_EOP_1)
      {
         /* End Of Packet */
         /* Packet is transmitted so transmit next one */
   /*      printk("  EOP (1)\n");  */
         DMA_LOG(DMA_TRACE,"  EOP (1)\n");
         if(dmaTxChannel[1].txComplete)
         {
            DMA_LOG1(DMA_TRACE,"     func (0x%08X)\n",(int)dmaTxChannel[1].txComplete);
            incaDMA_HandleEop(1);
/*           dmaTxChannel[1].txComplete(dmaTxChannel[1].device, 0); */
         }
         dmaTxChannel[1].txAckCounter ++;
      }
      if(regValue & INCA_IP_DMA_TXISR_EOP_2)
      {
         /* End Of Packet */
         /* Packet is transmitted so transmit next one */
/*       printk("  EOP (2)\n");      */
         DMA_LOG(DMA_TRACE,"  EOP (2)\n");
         if(dmaTxChannel[2].txComplete)
         {
            DMA_LOG1(DMA_TRACE,"     func (0x%08X)\n",(int)dmaTxChannel[2].txComplete);
            incaDMA_HandleEop(2);
/*          dmaTxChannel[2].txComplete(dmaTxChannel[2].device, 0); */
         }
         dmaTxChannel[2].txAckCounter ++;
      }

#ifdef DMA_DEBUG_ENABLE
   if(regValue & INCA_IP_DMA_TXISR_CMDCPT_0)
   {
      DMA_LOG(DMA_TRACE,"  CMDCPT (0)\n");
   }
   if(regValue & INCA_IP_DMA_TXISR_CMDCPT_1)
   {
      DMA_LOG(DMA_TRACE,"  CMDCPT (1)\n");
   }
   if(regValue & INCA_IP_DMA_TXISR_CMDCPT_2)
   {
      DMA_LOG(DMA_TRACE,"  CMDCPT (2)\n");
   }

   if(regValue & INCA_IP_DMA_TXISR_CPT_0)
   {
      /* Current Descriptor Used */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  CPT (0)\n");
   }
   if(regValue & INCA_IP_DMA_TXISR_CPT_1)
   {
      /* Current Descriptor Used */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  CPT (1)\n");
   }
   if(regValue & INCA_IP_DMA_TXISR_CPT_2)
   {
      /* Current Descriptor Used */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  CPT (2)\n");
   }

  if(regValue & INCA_IP_DMA_TXISR_HLD_0)
   {
      /* Current Descriptor Used */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  HOLD (0)\n");
   }
   if(regValue & INCA_IP_DMA_TXISR_HLD_1)
   {
      /* Current Descriptor Used */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  HOLD (1)\n");
   }
   if(regValue & INCA_IP_DMA_TXISR_HLD_2)
   {
      /* Current Descriptor Used */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  HOLD (2)\n");
   }

 
   if(regValue & INCA_IP_DMA_TXISR_RDERR_0)
   {
      /* S-Bus Read Error */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  RDERR (0)\n");
   }
   if(regValue & INCA_IP_DMA_TXISR_RDERR_1)
   {
      /* S-Bus Read Error */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  RDERR (1)\n");
   }
   if(regValue & INCA_IP_DMA_TXISR_RDERR_2)
   {
      /* S-Bus Read Error */
      /* Do nothing: just collect statistics ??? */
      DMA_LOG(DMA_TRACE,"  RDERR (2)\n");
   }
#endif /* DMA_DEBUG_ENABLE */
   } while(regValue != 0);
}
#endif


/****************************************************************************
Description: 
   This function raises the hold reset bit of the rx hold descriptor's 
   predecessor to re-enable DMA data transfer
Arguments:
   channelId      - identifier of the DMA Rx channel (0,1)
Return Value:
Remarks:
   None
****************************************************************************/
static void incaDMA_UpdateHOLD(unsigned long channelId)
{
#ifndef AUTOMATIC_POLLING      
#ifdef DIRECT_HOLD_RESET   
   register unsigned long  regValue;
#endif
#endif

#ifndef AUTOMATIC_POLLING      
#ifdef DIRECT_HOLD_RESET   

   /* give a HOLD RESET command to restart DMA processing */
   if( channelId == 0 )
   {
      DMA_READ_REG (INCA_IP_DMA_DMA_RXCCR0, regValue);
      regValue |= INCA_IP_DMA_DMA_RXCCR0_HR;
      DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR0, regValue);
   }
   else
   {
      DMA_READ_REG (INCA_IP_DMA_DMA_RXCCR1, regValue);
      regValue |= INCA_IP_DMA_DMA_RXCCR1_HR;
      DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR1, regValue);
   }
#else
    /* set flag to indicate the occurence of an hold interrupt */
    hold_reset_occured[channelId] = 1;
#endif
#endif   
}

/****************************************************************************
Description:
   This function passes received data buffer to the upper layer and updates
   RX descriptors list
Arguments:
   channelId - identification of the DMA channel
Return Value:
   None
Remarks:
   None
****************************************************************************/
static void incaDMA_Receive(unsigned long channelId)
{
   IncaDmaRxDescriptor_t   *pvirtCurrentDesc;
   unsigned long           physNewBuffer, virtNewBuffer, result, counter;
   unsigned long           info;
#ifdef RX_TASK_SCHEDULE    
   unsigned int            flags=0;
#endif
#if ( !defined DIRECT_HOLD_RESET | defined RX_TASK_SCHEDULE)	  
   register unsigned long  regValue;   
#endif
   
   DMA_DEBUG( DMA_TRACE,("incaDMA_Receive(%d)\n", channelId) );

   /* get current rx descriptor */
   pvirtCurrentDesc = dmaRxChannel[channelId].virtCurrentDescriptor;
   if (pvirtCurrentDesc == NULL)
   {
       printk("incaDMA_Receive(): No pvirtCurrentDesc available !!!\n"); 
	   return;
   }

   if(dmaRxChannel[channelId].getNewBuffer == NULL ||
      dmaRxChannel[channelId].putRcvdBuffer == NULL )
   {
	  return;
   }
   
   counter = 0;
   
   /*
    * process all descriptors used up by DMA hardware
    */
   while( pvirtCurrentDesc->status.dword & INCA_DMA_RX_C )
   {
#ifdef INCA_DMA_USE_COPYING
      virtNewBuffer = pvirtCurrentDesc->virtRxDataPtr;   
      physNewBuffer = pvirtCurrentDesc->physRxDataPtr;   
#else
      /* allocate memory for received data */
      virtNewBuffer = dmaRxChannel[channelId].getNewBuffer(dmaRxChannel[channelId].device,
                                                            &info);
      physNewBuffer = virt_to_phys((void*)virtNewBuffer);
#endif
      if(virtNewBuffer == 0)
      {
          printk("incaDMA_Receive(): No buffer available !!!\n"); 
		  return;
      }

      DMA_LOG2(DMA_ERR,"DMA_RX: 0x%08lX %d\n",pvirtCurrentDesc->virtRxDataPtr,
               pvirtCurrentDesc->status.field.NBT);      

#undef DMA_DEBUG_RX_PRINTOUT
/*  #define DMA_DEBUG_RX_PRINTOUT */
#ifdef DMA_DEBUG_RX_PRINTOUT
      for (result = 0; result < pvirtCurrentDesc->status.field.NBT; result++)
      {
         printk("R%02X ",((unsigned char*)pvirtCurrentDesc->virtRxDataPtr)[result]);
         if(((result+1)%16) == 0)
            printk("\n");
      }
      printk("\n");
#endif
      /* pass received data to upper layer */
      result = dmaRxChannel[channelId].
                  putRcvdBuffer( dmaRxChannel[channelId].device,
                                 (void*)pvirtCurrentDesc->virtRxDataPtr,
                                 pvirtCurrentDesc->status.field.NBT,
                                 pvirtCurrentDesc->info);
      if(result == -1)
      {
		 printk("incaDMA_Receive: up layer error \n");		 
		 DMA_LOG(DMA_ERR,"incaDMA_Receive: up layer error \n");
      }                                            

      /*
       * update RX hold descriptor, free current descriptor
       */
      pvirtCurrentDesc->params.field.HOLD = 1;
      pvirtCurrentDesc->status.dword &= ~(INCA_DMA_RX_EOP | INCA_DMA_RX_SOP | INCA_DMA_RX_C);
      dmaRxChannel[channelId].virtHoldDescriptor->params.field.HOLD = 0;
      /* store current descriptor as hold descriptor*/
      dmaRxChannel[channelId].virtHoldDescriptor = pvirtCurrentDesc;
	  
#ifndef DIRECT_HOLD_RESET	  
      /* give a HOLD RESET command to restart DMA processing 
         in case of occurence of hold interrupt */
      if(hold_reset_occured[channelId])
      {
          if( channelId == 0 )
          {
              DMA_READ_REG (INCA_IP_DMA_DMA_RXCCR0, regValue);
              regValue |= INCA_IP_DMA_DMA_RXCCR0_HR;
              DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR0, regValue);
          }
          else
          {
              DMA_READ_REG (INCA_IP_DMA_DMA_RXCCR1, regValue);
              regValue |= INCA_IP_DMA_DMA_RXCCR1_HR;
              DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR1, regValue);
          }
          hold_reset_occured[channelId] = 0;  /* delete flag */	  
      }	  
#endif	  

/*
      DMA_LOG2(DMA_ERR,"DMA_RX: 0x%08lX %d\n",pvirtCurrentDesc->virtRxDataPtr,
               pvirtCurrentDesc->status.field.NBT);
*/
      /* fill new current descriptor with data */
      pvirtCurrentDesc->params.field.offset = 
          (unsigned long)virtNewBuffer & 3;
      virtNewBuffer &= ~3;
      pvirtCurrentDesc->virtRxDataPtr = virtNewBuffer;
      pvirtCurrentDesc->physRxDataPtr = physNewBuffer;
      pvirtCurrentDesc->info          = info;
      pvirtCurrentDesc = (IncaDmaRxDescriptor_t*)pvirtCurrentDesc->virtNextRxDescPtr;
      counter++;
      dmaRxChannel[channelId].rxCounter ++;      
   }  /*  end of  while( pvirtCurrentDesc->status.dword & INCA_DMA_RX_C) )  */

   /* update current rx descriptor */
   dmaRxChannel[channelId].virtCurrentDescriptor = (IncaDmaRxDescriptor_t*)pvirtCurrentDesc;

   if( counter > dmaRxChannel[channelId].maxRxInOneIrq )
   {
      dmaRxChannel[channelId].maxRxInOneIrq = counter;
   }
#ifdef RX_TASK_SCHEDULE
   /* enable CPT interrupt */   
   if( channelId ==0 )
   {
      save_and_cli(flags);
	  
      DMA_READ_REG (INCA_IP_DMA_DMA_RXIMR, regValue);
      regValue &= ~INCA_IP_DMA_RXIMR_CPT(0);
      DMA_WRITE_REG (INCA_IP_DMA_DMA_RXIMR, regValue);	  
	  
      RDMA_HANDLING_FLAG &= ~RDMA0_HANDLING;
      restore_flags(flags); 
   }
   else
   {
      save_and_cli(flags);	
      DMA_READ_REG (INCA_IP_DMA_DMA_RXIMR, regValue);
      regValue &= ~INCA_IP_DMA_RXIMR_CPT(1);
      DMA_WRITE_REG (INCA_IP_DMA_DMA_RXIMR, regValue);	 	  
	  
      RDMA_HANDLING_FLAG &= ~RDMA1_HANDLING;
      restore_flags(flags);
   }   
#endif   
   
}


/****************************************************************************
Description:
   This function releases all tx descriptors that are no longer used 
   by DMA hardware.
Arguments:
   ch             - identifier of the DMA Tx channel (0,1,2)
Return Value:
Remarks:
   None
****************************************************************************/
static void incaDMA_HandleEop(int ch)
{
    unsigned long         info = 0;
    int                   eop = 0;
    IncaDmaTxDescriptor_t *virtCleanupDescriptor;

    do {
        virtCleanupDescriptor = dmaTxChannel[ch].virtCleanupDescriptor;
        info = 0;
        eop = 0;

        if (!virtCleanupDescriptor->C)
        {
            DMA_LOG(DMA_ERR, "incaDma_HandleEop: no C-Bit set in CleanupDescriptor\n");
#ifdef DMA_DEBUG_ENABLE
            incaDMA_PrintAllInfoTx(ch);
#endif
        }
        else
        {
            info = virtCleanupDescriptor->info;
            eop = virtCleanupDescriptor->params.field.Eop;
            virtCleanupDescriptor->C = 0;
            dmaTxChannel[ch].virtCleanupDescriptor =
                (IncaDmaTxDescriptor_t *)virtCleanupDescriptor->virtNextTxDescPtr;
        }

        if(dmaTxChannel[ch].txComplete && eop)
        {
            dmaTxChannel[ch].txComplete(dmaTxChannel[ch].device, info);
        }
    } while (dmaTxChannel[ch].virtCleanupDescriptor->C);
}



/****************************************************************************
Description:
   This function puts data buffer into current available TX descriptor and 
   updates TX descriptors list
Arguments:
   channelId   - identifier of the DMA Tx channel (0,1,2)
   buffer      - pointer to the data buffer
   bufLen      - length of the data buffer
   lastBuffer  - flag that shows is it last buffer of the message
   info        - adress of    
Return Value:
   0   -  data buffer is succesfully connected to the current TX descriptor
          and will be transmitted on the next interrupt
   -1  -  data buffer cannot be transmitted because of lack of descriptors or
          some other reasons
Remarks:
   None
****************************************************************************/
unsigned long incaDMA_Transmit( unsigned long channelId, void *buffer, 
                                unsigned long bufLen,    int  lastBuffer, 
                                unsigned long info )
{
   register unsigned long   regValue, command;
   int                      firstBuffer = 0;
   unsigned int             flags = 0;

   if(buffer == NULL)
   {
      DMA_DEBUG(DMA_ERR,("incaDMA_Transmit: wrong parameters\n"));
      return -1;
   }

#ifndef INCA_DMA_USE_TX_INTERRUPT 
   incaDMA_HandleEop(channelId);  /* release tx descriptors no longer used */
#endif 
          
    /* get the current hold descriptor */
   if(dmaTxChannel[channelId].virtCurrentDescriptor == NULL)
   {
      /* this is a first frame of the packet */
      /* get the next free descriptor */
      dmaTxChannel[channelId].virtCurrentDescriptor = 
         (IncaDmaTxDescriptor_t*)dmaTxChannel[channelId].virtHoldDescriptor->virtNextTxDescPtr;
      firstBuffer = 1;

      dmaTxChannel[channelId].curPacketSize = 0;

#ifdef INCA_DMA_USE_COPYING
      dmaTxChannel[channelId].virtCurrentBufferPtr = 
          (unsigned char*)dmaTxChannel[channelId].virtCurrentDescriptor->virtTxDataPtr;
   }
#else
   }
   else
   {
      /* get the next free descriptor */
      dmaTxChannel[channelId].virtCurrentDescriptor = 
         (IncaDmaTxDescriptor_t*)dmaTxChannel[channelId].virtCurrentDescriptor->virtNextTxDescPtr;
   }
/*
   DMA_DEBUG(DMA_TRACE,("incaDMA_Transmit: dmaTxChannel[%d].virtCurrentDescriptor->C: %d\n", 
             channelId, dmaTxChannel[channelId].virtCurrentDescriptor->C));   
*/   

   /* Check if there is a free descriptor */
   if( dmaTxChannel[channelId].virtCurrentDescriptor->virtNextTxDescPtr == 
       (unsigned long)dmaTxChannel[channelId].virtCleanupDescriptor) 
   {
      DMA_DEBUG(DMA_ERR,("incaDMA_Transmit: there is no free descriptor\n"));
      dmaTxChannel[channelId].virtCurrentDescriptor = NULL;
      return -1;
   }

   save_and_cli(flags);
   
#endif

   /* delete Complete bit so that it can be set again by DMA */
   dmaTxChannel[channelId].virtCurrentDescriptor->C = 0;
   dmaTxChannel[channelId].virtCurrentDescriptor->info = info;

#ifdef INCA_DMA_USE_COPYING
   memcpy((void*)dmaTxChannel[channelId].virtCurrentDescriptor->virtTxDataPtr, 
          buffer, bufLen);
#else
   /* Set the TX data pointer */
   dma_cache_wback_inv((unsigned long)buffer, bufLen);
   if ((unsigned long)buffer & 3) /* unaligned buffer */
   {
       int rest = 4 - ((unsigned long)buffer & 3);

       DMA_LOG(DMA_ERR, ("incaDMA_Transmit: unaligned data buffers (0x%08X)"
                         "are not supported !\n", buffer));
/*       printk("incaDMA_Transmit: unaligned data buffers (0x%08X)"
              "are not supported !\n", buffer);
*/       
       if (rest >= bufLen)
       {
           DMA_DEBUG(DMA_ERR,("incaDMA_Transmit: Don't try to transmit unaligned"
                              "data with less than 4 bytes !\n"));
           return -1;
       }

       if (dmaTxChannel[channelId].virtCurrentDescriptor->virtNextTxDescPtr ==
           (unsigned long)dmaTxChannel[channelId].virtCleanupDescriptor)
       {
           DMA_DEBUG(DMA_ERR,("incaDMA_Transmit: there is no second free "
                              "descriptor for unaligned data\n"));
           dmaTxChannel[channelId].virtCurrentDescriptor = NULL;
           return -1;
       }

       dmaTxChannel[channelId].virtCurrentDescriptor->virtTxDataPtr = (unsigned long)
           &dmaTxChannel[channelId].virtCurrentDescriptor->unalignedData;
       memcpy(&dmaTxChannel[channelId].virtCurrentDescriptor->unalignedData,
              buffer, rest);
       dmaTxChannel[channelId].virtCurrentDescriptor->unalignedData = 0x12345678;
       buffer += rest;
       bufLen -= rest;

       dmaTxChannel[channelId].virtCurrentDescriptor->physTxDataPtr =
           virt_to_phys((void*)dmaTxChannel[channelId].virtCurrentDescriptor->virtTxDataPtr);

       dmaTxChannel[channelId].virtCurrentDescriptor->params.field.NBA = rest;

       dmaTxChannel[channelId].virtCurrentDescriptor->params.dword &= 
           ~( INCA_DMA_TX_EOP | INCA_DMA_TX_IEOP | 
              INCA_DMA_TX_HOLD | INCA_DMA_TX_ICPT );

       /* Now set current pointer to next descriptor to set it up for 
          the remaining aligned data */
       dmaTxChannel[channelId].virtCurrentDescriptor = (IncaDmaTxDescriptor_t *)
           dmaTxChannel[channelId].virtCurrentDescriptor->virtNextTxDescPtr;
       
       dmaTxChannel[channelId].virtCurrentDescriptor->params.field.Sop = 0;
       /* delete Complete bit so that it can be set again by DMA */
       dmaTxChannel[channelId].virtCurrentDescriptor->C = 0;
       dmaTxChannel[channelId].virtCurrentDescriptor->info = info;
   }

   dmaTxChannel[channelId].virtCurrentDescriptor->virtTxDataPtr = (unsigned long)(buffer);
   dmaTxChannel[channelId].virtCurrentDescriptor->physTxDataPtr = virt_to_phys(buffer);
#endif 

   /* Set the buffer size */
   dmaTxChannel[channelId].virtCurrentDescriptor->params.field.NBA = bufLen;
      
#undef DMA_DEBUG_TX_PRINTOUT
/*  #define DMA_DEBUG_TX_PRINTOUT */
#ifdef DMA_DEBUG_TX_PRINTOUT
   for (command = 0; command < bufLen; command++)
   {
      printk("T%02X ",((unsigned char*)dmaTxChannel[channelId].virtCurrentDescriptor->TxDataPtr)[command]);
      if(((command+1)%16) == 0)printk("\n");
   }
   printk("\n");
#endif

   if(firstBuffer == 1)
   {
      dmaTxChannel[channelId].virtCurrentDescriptor->params.field.Sop    = 1;
/*      DMA_DEBUG(DMA_TRACE,("incaDMA_Transmit(%d): Tx first frame of the packet\n", channelId));*/
   }
   else
   {
      dmaTxChannel[channelId].virtCurrentDescriptor->params.field.Sop    = 0;
   }

   if(lastBuffer == 1)
   {

	   /* DMA_DEBUG(DMA_TRACE,("incaDMA_Transmit(%d): Tx last frame of the packet\n", channelId));*/
      /* If it is the last buffer of a frame 
         then set FE , HOLD , THI bits
         so that the DT stops after transmit */
      dmaTxChannel[channelId].virtCurrentDescriptor->params.dword |= 
                  ( INCA_DMA_TX_EOP | INCA_DMA_TX_IEOP |
                     INCA_DMA_TX_HOLD | INCA_DMA_TX_ICPT );
      /* Reset the HOLD bit */
      dmaTxChannel[channelId].virtHoldDescriptor->params.field.HOLD = 0;

      /* update the TX hold descriptor */
      dmaTxChannel[channelId].virtHoldDescriptor = dmaTxChannel[channelId].virtCurrentDescriptor;
      dmaTxChannel[channelId].virtCurrentDescriptor = NULL;

   
      /* set the command to force a re-read of the hold descriptor by the DMA */
      /* and give a HOLD RESET command */
      if(dmaTxChannel[channelId].bInit == 0)
      {
         command = INCA_IP_DMA_DMA_TXCCR0_INIT | INCA_IP_DMA_DMA_TXCCR0_HPEN;
         switch(channelId)
         {
            case 0:
               DMA_READ_REG (INCA_IP_DMA_DMA_TXCCR0, regValue);
               regValue |= command;
               DMA_WRITE_REG (INCA_IP_DMA_DMA_TXCCR0, regValue);
               break;

            case 1:
               DMA_READ_REG (INCA_IP_DMA_DMA_TXCCR1, regValue);
               regValue |= command;
               DMA_WRITE_REG (INCA_IP_DMA_DMA_TXCCR1, regValue);
               break;

            case 2:
               DMA_READ_REG (INCA_IP_DMA_DMA_TXCCR2, regValue);
               regValue |= command;
               DMA_WRITE_REG (INCA_IP_DMA_DMA_TXCCR2, regValue);
               break;

            default:
               break;
         }

         dmaTxChannel[channelId].bInit = 1;
      }

      dmaTxChannel[channelId].txCounter ++;
      if(dmaTxChannel[channelId].curPacketSize > dmaTxChannel[channelId].maxPacketSize)
      {
         dmaTxChannel[channelId].maxPacketSize = dmaTxChannel[channelId].curPacketSize;
      }
  
   }
   else
   {
      dmaTxChannel[channelId].virtCurrentDescriptor->params.dword &= 
                  ~( INCA_DMA_TX_EOP | INCA_DMA_TX_IEOP | 
                     INCA_DMA_TX_HOLD | INCA_DMA_TX_ICPT );
   }

#ifdef DMA_DEBUG_ENABLE
/*
   printk("DMA_TX end:\n");
   if(dmaTxChannel[channelId].currentDescriptor)
      incaDMA_PrintTxInfo( dmaTxChannel[channelId].currentDescriptor );
   else
      incaDMA_PrintTxInfo( dmaTxChannel[channelId].holdDescriptor );*/
#endif

/*   DMA_DEBUG(DMA_TRACE,("incaDMA_Transmit: Id:%d, size:%d\n", channelId, bufLen));*/
   dmaTxChannel[channelId].useCounter ++;
   restore_flags(flags);	
   
   return 0;
}
EXPORT_SYMBOL(incaDMA_Transmit);


/****************************************************************************
Description:
   This function disables specified DMA Tx Channel
Arguments:
   channelId      - identifier of the DMA Tx channel (0,1,2)
Return Value:
   0   - channel is successfully closed
   -1  - wrong channel id passed as a parameter
Remarks:
   None
****************************************************************************/
long incaDMA_CloseTxChannel ( unsigned long channelId )
{
   /* close channel */
   switch(channelId)
   {
      case 0:
         DMA_WRITE_REG (INCA_IP_DMA_DMA_TXCCR0, INCA_IP_DMA_DMA_TXCCR0_OFF);
         break;

      case 1:
         DMA_WRITE_REG (INCA_IP_DMA_DMA_TXCCR1, INCA_IP_DMA_DMA_TXCCR1_OFF);
         break;

      case 2:
         DMA_WRITE_REG (INCA_IP_DMA_DMA_TXCCR2, INCA_IP_DMA_DMA_TXCCR2_OFF);
         break;

      default:
         return -1;
   }

   /* release all resources */
   if(dmaTxChannel[channelId].virtFirstDescriptorAddr)
   {
      kfree( (void*)KSEG0ADDR(KMTxfirstDescriptorAddress[channelId]) );
      KMTxfirstDescriptorAddress[channelId] = NULL;

      /* uninit Tx channel structure */
      memset(&dmaTxChannel[channelId],0,sizeof(IncaDmaTxChannel_t));
   }
#ifdef INCA_DMA_USE_TX_INTERRUPT 
   txIrqEnabled--;
   if(txIrqEnabled == 0)
   {
      DMA_INTERRUPTS_DISABLE(INCA_IRQ_LINE_DMATIR);
   }
#endif

   return 0;
}
EXPORT_SYMBOL(incaDMA_CloseTxChannel);

/****************************************************************************
Description:
   This function disables specified DMA Rx Channel
Arguments:
   channelId      - identifier of the DMA Rx channel (0,1)
Return Value:
   OK    - channel is successfully closed
   ERROR - wrong channel id passed as a parameter
Remarks:
   None
****************************************************************************/
long incaDMA_CloseRxChannel ( unsigned long channelId )
{
   /* close channel */
   switch(channelId)
   {
      case 0:
         DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR0, 0x00000000);
         break;

      case 1:
         DMA_WRITE_REG (INCA_IP_DMA_DMA_RXCCR1, 0x00000000);
         break;

      default:
         return -1;
   }

   /* release all resources */
   if (dmaRxChannel[channelId].virtFirstDescriptorAddr)
   {
      kfree((void*)KSEG0ADDR(KMRxfirstDescriptorAddress[channelId]));
      KMRxfirstDescriptorAddress[channelId] = NULL;
#ifdef INCA_DMA_USE_COPYING
      kfree((void*)KSEG0ADDR(KMRxBuffers[channelId]));
      KMRxBuffers[channelId] = NULL;
#endif

      /* uninit Rx channel structure */
      memset(&dmaRxChannel[channelId],0,sizeof(IncaDmaRxChannel_t));
   }

   rxIrqEnabled--;
   if(rxIrqEnabled == 0)
   {
      DMA_INTERRUPTS_DISABLE(INCA_IRQ_LINE_DMARIR);
   }

   return 0;
}
EXPORT_SYMBOL(incaDMA_CloseRxChannel);


#if defined(DMA_DEBUG_ENABLE) || defined(DMA_USE_PROC)

void incaDMA_PrintTxInfo( IncaDmaTxDescriptor_t *desc );
void incaDMA_PrintRxInfo( IncaDmaRxDescriptor_t *desc );
/****************************************************************************
Description:
   Prints information on all rx descriptors.
Arguments:
   channelId      - identifier of the DMA Rx channel (0,1)
Return Value:
   None
Remarks:
   virtCurrentDescriptor is marked with '*'
   virtHoldDescriptor is marked with '+'
****************************************************************************/
void incaDMA_PrintAllInfoRx(unsigned long channelId)
{
   IncaDmaRxDescriptor_t   *rxDescriptor;
   unsigned long                  i;

   printk("------------------------  RX DESCRIPTORS  --------------------------\n");
   for(i = 0, rxDescriptor = dmaRxChannel[channelId].virtFirstDescriptor;
       i < dmaRxChannel[channelId].descriptorNumber;
       i++, rxDescriptor = (IncaDmaRxDescriptor_t*)rxDescriptor->virtNextRxDescPtr)
   {
       if (rxDescriptor == dmaRxChannel[channelId].virtCurrentDescriptor)
           printk("*");
       if (rxDescriptor == dmaRxChannel[channelId].virtHoldDescriptor)
           printk("+");
      incaDMA_PrintRxInfo( rxDescriptor );
   }
   printk("--------------------------------------------------------------------\n");
}

/****************************************************************************
Description:
   Prints information on all tx descriptors.
Arguments:
   channelId      - identifier of the DMA Tx channel (0,1,2)
Return Value:
   None
Remarks:
   virtCurrentDescriptor is marked with '*'
   virtHoldDescriptor is marked with '+'
   virtCleanupDescriptor is marked with '-'
****************************************************************************/
void incaDMA_PrintAllInfoTx(unsigned long channelId)
{
   IncaDmaTxDescriptor_t   *txDescriptor;                     
   unsigned long           i;

   printk("------------------------  TX DESCRIPTORS  --------------------------\n");
   for(i = 0, txDescriptor = (IncaDmaTxDescriptor_t*)dmaTxChannel[channelId].virtHoldDescriptor->virtNextTxDescPtr;
       i < dmaTxChannel[channelId].descriptorNumber;
       i++, txDescriptor = (IncaDmaTxDescriptor_t*)txDescriptor->virtNextTxDescPtr)
   {
       if (txDescriptor == dmaTxChannel[channelId].virtCurrentDescriptor)
           printk("*");
       if (txDescriptor == dmaTxChannel[channelId].virtHoldDescriptor)
           printk("+");
       if (txDescriptor == dmaTxChannel[channelId].virtCleanupDescriptor)
           printk("-");
      incaDMA_PrintTxInfo( txDescriptor );
   }
   printk("--------------------------------------------------------------------\n");
}

/****************************************************************************
Description:
   Prints information on the passed tx descriptor
Arguments:
   *desc - pointer to tx descriptor to be printed
Return Value:
   None
Remarks:
   None
****************************************************************************/
void incaDMA_PrintTxInfo( IncaDmaTxDescriptor_t *desc )
{
   printk("0x%08lX> HOLD(%d), Eop(%d), Sop(%d), ICpt(%d), IEop(%d), NBA(%d)\n",
         (unsigned long)desc,
         desc->params.field.HOLD,
         desc->params.field.Eop,
         desc->params.field.Sop,
         desc->params.field.ICpt,
         desc->params.field.IEop,
         desc->params.field.NBA);

   printk("            next(0x%08lX/0x%08lX),  data(0x%08lX/0x%08lX), C(%d)\n",
          desc->physNextTxDescPtr, desc->virtNextTxDescPtr,
          desc->physTxDataPtr, desc->virtTxDataPtr,
          desc->C);
}

/****************************************************************************
Description:
   Prints information on the passed rx descriptor
Arguments:
   *desc - pointer to rx descriptor to be printed
Return Value:
   None
Remarks:
   None
****************************************************************************/
void incaDMA_PrintRxInfo( IncaDmaRxDescriptor_t *desc )
{
   printk("0x%08lX> HOLD(%d), ICpt(%d), IEop(%d), Sop(%d), Eop(%d), NFB(%d), NBT(%d), O(%d)\n",
         (unsigned long)desc,
         desc->params.field.HOLD,
         desc->params.field.ICpt,
         desc->params.field.IEop,

         desc->status.field.Sop,
         desc->status.field.Eop,
         desc->params.field.NFB,
         desc->status.field.NBT,
         desc->params.field.offset);

   printk("            next(0x%08lX/0x%08lX),  data(0x%08lX/0x%08lX), C(%d)\n",
          desc->physNextRxDescPtr, desc->virtNextRxDescPtr,
         desc->physRxDataPtr, desc->virtRxDataPtr,
         desc->status.field.C);
}

#endif /* DMA_DEBUG_ENABLE || DMA_USE_PROC */

#ifdef DMA_USE_PROC
static int inca_dma_read_procmem(char *buf, char **start, off_t offset, 
                                        int count, int *eof, void *data)
{
    int len = 0;

    incaDMA_PrintAllInfoTx(0); 
    incaDMA_PrintAllInfoRx(0); 
    
    *eof = 1;	// No more data available
    
    return len;
}

static int inca_dma_write_procmem(struct file *file, const char *buffer,
					unsigned long count, void *data)
{
#if 0
    const char *buf = buffer;
    const char *cmd;
    const char *parg_string;
    const char *arg;
    int     slen, c;
    long    cmd_arg[10]={}; /* max. 10 command arguments (integer/long/string) */
    char    cmd_arg_strings[100]; /* max 100 characters as string parameters */
    int     cmd_arg_str_offset = 0;
    int     argc = 0;
    int     ret = 0;
    
    cmd = get_string(&buf, &slen);
    if ((cmd == NULL) || (slen == 0))
    {
        printk("No/invalid command specified (%s)\n", buffer);
        return -EINVAL;
    }
    
    for (c = 0; c < (sizeof(commands)/sizeof(commands[0])); c++)
    {
        if (strncmp(cmd, commands[c].name, slen) == 0)
        {
            /* yeah, we found our command */
/*              printk("Command found: name=\"%s\", fct=%p, param=\"%s\"\n", */
/*                     commands[c].name, commands[c].function, commands[c].param); */
            break;
        }
    }
    
    if (c >= (sizeof(commands)/sizeof(commands[0])))
    {
        int i;
        printk("Command not found (");
        for (i = 0; i < slen; i++) 
            printk("%c", cmd[i]);
        printk(")\n");
        return -EINVAL;
    }
    
    parg_string = commands[c].param;
    while ((*parg_string) && (argc < (sizeof(cmd_arg)/sizeof(cmd_arg[0]))))
    {
        arg = get_string(&buf, &slen);
        if ((arg == NULL) || (slen == 0))
        {
            printk("Too few parameters for command \"%s\", required are \"%s\"\n",
                   commands[c].name, commands[c].param);
            printk("arg=%p, slen=%d\n", arg, slen);
            return -EINVAL;
        }
        
        switch (*parg_string)
        {
            case 'i':
            case 'l':
                cmd_arg[argc++] = atoi(arg, slen);
                break;
            case 's':
                memcpy(&cmd_arg_strings[cmd_arg_str_offset], arg, slen);
                cmd_arg_strings[cmd_arg_str_offset + slen] = '\0';
                cmd_arg[argc++] = (long)&cmd_arg_strings[cmd_arg_str_offset];
                cmd_arg_str_offset += slen+1;
                break;
            default:
                printk("Invalid format string in module, please fix and re-compile\n");
                return -EINVAL;
        }
        
        parg_string++;
    }

    arg = get_string(&buf, &slen);
    if ((arg != NULL) && (slen != 0))
    {
        printk("Too many parameters for command \"%s\", required are \"%s\"\n",
               commands[c].name, commands[c].param);
        return -EINVAL;
    }
    
    /* This instructions casts the void pointer commands.function to a function that
     returns int and has 10 long parameters. Though the function actually called has
     a different argument list it works if the parameter are set up accordingly
     (e.g. int/long or string pointer) */
    ret = ((int (*)(long,long,long,long,long,long,long,long,long,long))(commands[c].function))
        (cmd_arg[0], cmd_arg[1], cmd_arg[2], cmd_arg[3], 
         cmd_arg[4], cmd_arg[5], cmd_arg[6], cmd_arg[7], 
         cmd_arg[8], cmd_arg[9]);
    printk("return value is: %d (0x%08x)\n", ret, ret);
#endif    
    return count;
	
}
#endif
