#include <linux/module.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/slab.h> // former "malloc.h"
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/proc_fs.h>
#include <linux/ioport.h>

#include <asm/incaip/inca-ip.h>
#include <asm/incaip/mux.h>
#include <asm/incaip/ledmatrix.h>


#define INCA_MUX_LED_INV    0	/* to be removed when mux module exists! */
#define INCA_MUX_LED_NCOL   4	/* to be removed when mux module exists! */

static int 	inca_led_ioctl		(struct inode *inode, 
					 struct file *filp, 
					 unsigned int cmd, unsigned long arg);

/*>>>>>>>>>>>>>>>>>>>>>>>>> MODULE Specific Stuff <<<<<<<<<<<<<<<<<<<<<<<<<<<*/
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Andreas Schmidt");
MODULE_DESCRIPTION("Inca-IP LED Matrix Driver");
/*
 * Parameter Definitions for Module
 */
static int led_pins = -1;	// Number of LED pins (5..10) 
MODULE_PARM(led_pins, "i");
MODULE_PARM_DESC(led_pins, "Number of LED pins (5..10)");

static int ncol = -1;		// Number of columns in LED array (2 or 4)
MODULE_PARM(ncol, "i");
MODULE_PARM_DESC(ncol, "Number of columns in LED array (2 or 4)");

static int inv  = -1;		// Invert Polarity of Column Driver Signal
MODULE_PARM(inv, "i");
MODULE_PARM_DESC(inv, "Invert Polarity of Column Driver Signal (0 for external NPN transistor, 1 for direct LED drive)");

static int led_eth_lan = 0;	// LEDs controlled by Ethernet LAN Port
MODULE_PARM(led_eth_lan, "i");
MODULE_PARM_DESC(led_eth_lan, "LEDs controlled by Ethernet LAN Port (bits[4:0] = [SPD, ACT, DPX, STA, TL])");

static int led_eth_pc = 0;	// LEDs controlled by Ethernet PC Port
MODULE_PARM(led_eth_pc, "i");
MODULE_PARM_DESC(led_eth_pc, "LEDs controlled by Ethernet PC Port (bits[4:0] = [SPD, ACT, DPX, STA, TL])");

static int major = 0;	// Major ID of ledmatrix device (default: 0=dynamic)
MODULE_PARM(major, "i");
MODULE_PARM_DESC(major, "Major ID of ledmatrix device (default: 0=dynamic)");


/*>>>>>>>>>>>>>>>>>>>>>>>>>>>> Global Variables <<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
static u32 led_mask = 0x00FFFFFF;
static int activate_led9 = 0;

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> PROC Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
static struct proc_dir_entry *inca_led_proc;

static int inca_led_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	long	len = 0;

	len  = sprintf(buf+len,"Inca-IP LED_REG register: 0x%08X\n", 
					*INCA_IP_TSF_LED_REG);
	len += sprintf(buf+len,"  Mask of available LEDs: 0x%08X\n", 
					led_mask);
	len += sprintf(buf+len,"    LEDMUX_CONF register: 0x%08X\n", 
					*INCA_IP_TSF_LEDMUX_CONF);

	*eof = 1;	// No more data available
	return len;
}

static u32 hexstr2uint(char* s)
{
	int	digit = 0;
	u32	temp_val = 0;

	while (*s != 0xA) { /* last character should be <LF> */
		if (digit >= 8) {
			printk(KERN_ERR "Enter a Hex value (max 32 bit)!\n");
			return 0; /* results in keeping LEDs as-is! */
		}
		
		if (*s >= '0' && *s <= '9') {
			temp_val = (temp_val << 4) + (*s - '0');
		}
		else if (*s >= 'a' && *s <= 'f') {
			temp_val = (temp_val << 4) + (*s - 'a') + 10;
		}
		else if (*s >= 'A' && *s <= 'F') {
			temp_val = (temp_val << 4) + (*s - 'A') + 10;
		}
		else {
			printk(KERN_ERR "Enter a Hex value!\n");
			return 0; /* results in keeping LEDs as-is! */
		} 
		digit++;
		s++;
	}

	return temp_val;
}

static int inca_led_write_procmem(struct file *file, const char *buffer,unsigned long count, void *data)
{
	char*	ibuf = (char*) buffer;
	u32 	val = 0;
	int	err = 0;

	if (strncmp(ibuf, "led_set 0x", 10) == 0) {
		printk(KERN_INFO "...decoding led_set command...\n");
		ibuf += 10;
		val = hexstr2uint(ibuf);
		if (val == 0) 
			err = -EINVAL;
		else
			err = inca_led_on(val);
	}
	else if	(strncmp(ibuf, "led_clr 0x", 10) == 0) {
		printk(KERN_INFO "...decoding led_clr command...\n");
		ibuf += 10;
		val = hexstr2uint(ibuf);
		if (val == 0) 
			err = -EINVAL;
		else
			err = inca_led_off(val);
	}
	else if	(strncmp(ibuf, "led_eth 0x", 10) == 0) {
		printk(KERN_INFO "...decoding led_eth command...\n");
		ibuf += 10;
		val = hexstr2uint(ibuf);
		inca_led_eth_ctrl(LED_ETHPORT_PC, val & 0x1F);
		inca_led_eth_ctrl(LED_ETHPORT_LAN, (val >> 16) & 0x1F);
		err = 0;
	}
	else {
		printk(KERN_ERR "command syntax unknown!\n");
		err = -EINVAL;
	}
	
	if (err) {
		printk(KERN_ERR "Usage: led_set|led_clr|led_eth 0x<hex_value>\n"
			"Currently available LEDs: 0x%08X (Mask)\n", led_mask); 
	}
	
	printk(KERN_INFO "Inca-IP LED_REG register: 0x%08X\n", 
							*INCA_IP_TSF_LED_REG);
	return count;
	
}

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> IOCTL Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
struct file_operations led_fops = {
	ioctl : inca_led_ioctl
};
static int inca_led_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg) {
	int err;
	
	switch (cmd) {
		case LED_ON:
			err = inca_led_on((u32) arg);
			break;
		case LED_OFF:
			err = inca_led_off((u32) arg);
			break;
		case LED_ETH_PC:
			err = inca_led_eth_ctrl(LED_ETHPORT_PC, (u8) arg);
			break;
		case LED_ETH_LAN:
			err = inca_led_eth_ctrl(LED_ETHPORT_LAN, (u8) arg);
			break;
		default:
			err = -ENOIOCTLCMD;
	}
	return err;
}

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> Functions <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

//static u32 inca_led_check(u32 indexLED)
//{
//	/* check if all selected LED pins are supported by the hardware */
//	return indexLED & ~led_mask;
//}
#define inca_led_check(indexLED) (indexLED & ~led_mask)

/****************************************************************************
Description :
   Switch off the LED(s) of the Matrix. All set bits in the 'indexLED' are
   switched off. Other bits are ignored.  
Arguments   :
   indexLED - bit vector as in register LED_REG (see INCA-IP User's Manual)
Return      :
   returns 0 when the clearing of the LED was successful. If the chip
   or the implementation doesn't support to clear one or more LEDs, the 
   function is returning the indexLED bits that lead to the error.
Remarks     :
   This routine doesn't switch on any LED of the Matrix
****************************************************************************/
u32 inca_led_off(u32 indexLED)
{
	int res;
	
	/* check if all selected LED pins are supported by the hardware */
	res = inca_led_check(indexLED);
	if (res == 0) {
		/* clear the bit in the LED register. */
		/* All other LED are not changed */
		*INCA_IP_TSF_LED_REG &= ~indexLED;
	}
	return res;
}

/****************************************************************************
Description :
   Switch on the LED(s) of the Matrix. All set bits in the 'indexLED' are
   switched on. Other bits are ignored.  
Arguments   :
   indexLED - bit vector as in register LED_REG (see INCA-IP User's Manual)
Return      :
   returns 0 when the setting of the LED was successful. If the chip
   or the implementation doesn't support to set one or more LEDs, the 
   function is returning the indexLED bits that lead to the error.
Remarks     :
   This routine doesn't switch off any LED of the Matrix
****************************************************************************/
u32 inca_led_on(u32 indexLED)
{
	int res;

	/* check if all selected LED pins are supported by the hardware */
	res = inca_led_check(indexLED);
	if (res == 0) {
		/* set the bit in the LED register. */
		/* All other LED are not changed */
		*INCA_IP_TSF_LED_REG |= indexLED;
	}
	return res;
}

u8 inca_led_eth_ctrl (int eth_port, u8 eth_leds)
{
	switch (eth_port) {
		case LED_ETHPORT_PC:
			*INCA_IP_TSF_LEDMUX_CONF &= 0x08E00003;
			*INCA_IP_TSF_LEDMUX_CONF |= ((eth_leds & (LED_ETH_SPD |
								  LED_ETH_ACT |
								  LED_ETH_DPX |
								  LED_ETH_STA |
								  LED_ETH_TL) 
								  ) << 16);
			led_mask &= ~((u32) eth_leds &
					(LED(0,4)|LED(1,4)|LED(2,4)|LED(3,4)) );
			led_mask |= (~((u32) eth_leds) &
					(LED(0,4)|LED(1,4)|LED(2,4)|LED(3,4)) );
			break;
		case LED_ETHPORT_LAN:
			*INCA_IP_TSF_LEDMUX_CONF &= 0x001F0003;
			*INCA_IP_TSF_LEDMUX_CONF |= ((eth_leds & (LED_ETH_SPD |
								  LED_ETH_ACT |
								  LED_ETH_DPX |
								  LED_ETH_STA |
								  LED_ETH_TL) 
								  ) << 21);
			led_mask &= ~((u32) (eth_leds << 4) & 
					(LED(0,5)|LED(1,5)|LED(2,5)|LED(3,5)) );
			led_mask |= (~((u32) (eth_leds << 4)) & 
					(LED(0,5)|LED(1,5)|LED(2,5)|LED(3,5)) );
			break;
		default:
			printk(KERN_ERR "Invalid Ethernet Port Number!\n");
			return eth_leds;
	}	
	return 0;
}

/*>>>>>>>>>>>>>>>>>>>>>>>>>>> Module Init/Exit <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/****************************************************************************
Description :
   This function initializes the LED multiplexer hardware of the INCA-IP 
   device. This function is called in the sysHWInit after kernel
   initialization.  
Arguments   :
Return      :
Remarks     :
   This function might use the A0 pin of the EBU. 
****************************************************************************/
static int __init inca_led_init(void)
{
	int res = 0;
	unsigned long flags;
	
	if (led_pins == -1) 
		led_pins = inca_mux_led_pins();
	else if (led_pins == 10 && inca_mux_led_pins() < 10)
		activate_led9 = 1;
		
	if (led_pins >= 5) {
		led_mask = ~((u32) 0xFFFFFFFF << (4*(led_pins-4)));
	}
	else {	/* too few LED pins requested by user! */
		printk(KERN_ERR "too few LED pins requested by user "
							"(min: 5)!\n");
		return -EINVAL;
	}
	
	/* allocate Linux resources */
	if (request_region((u32) INCA_IP_TSF_LEDMUX_CONF, 0x08, 
					"INCA-IP LED Matrix") == NULL) {
		printk(KERN_ERR "request_region() failed!\n");
		res = -EBUSY;
		goto FAIL_REQUEST_REGION;
	}
	
#ifndef DONE_ELSEWHERE
	/* enable the clock of the LED module in the Power Management Unit */
 	save_flags(flags); cli();
	*INCA_IP_PMU_PM_GEN |= INCA_IP_PMU_PM_GEN_EN9;   
 	restore_flags(flags);
#endif

	if (activate_led9) {
		/* use A0 pin of the EBU interface for the LED multiplexer */
 		save_flags(flags); cli();
		*INCA_IP_Ports_P0_ALTSEL |= 1;
	 	restore_flags(flags);
	}

	if (inv == -1) inv = INCA_MUX_LED_INV;
	if (inv) {
		*INCA_IP_TSF_LEDMUX_CONF |= INCA_IP_TSF_LEDMUX_CONF_INV;
	} 
	else {
		*INCA_IP_TSF_LEDMUX_CONF &= ~INCA_IP_TSF_LEDMUX_CONF_INV;
	}

	/* clear all LEDs */
	*INCA_IP_TSF_LED_REG = 0x00000000;

	/* define which LEDs are controlled autonomously by Ethernet HW */ 
	inca_led_eth_ctrl(LED_ETHPORT_LAN, led_eth_lan);
	inca_led_eth_ctrl(LED_ETHPORT_PC,  led_eth_pc);
	
	if (ncol == -1) ncol = INCA_MUX_LED_NCOL;
	if (ncol == 2) {
		*INCA_IP_TSF_LEDMUX_CONF |= INCA_IP_TSF_LEDMUX_CONF_NCOL;
		
		/* LEDs L24, L34, L25, L35, ..., L29 & L39 not available! */
		led_mask &= 0x00333333;
	} 
	else {
		*INCA_IP_TSF_LEDMUX_CONF &= ~INCA_IP_TSF_LEDMUX_CONF_NCOL;
	}
	
	/* enable the LED multiplexer on the board */
 	save_flags(flags); cli();
	*INCA_IP_TSF_TSF_CONF |= INCA_IP_TSF_TSF_CONF_LEDEN;
 	restore_flags(flags);
   
   	/* register ledmatrix device */
   	res = register_chrdev(major, "ledmatrix", &led_fops);
	if (res < 0)  {
		printk(KERN_ERR "Could not register ledmatrix device!\n");
		goto FAIL_REGISTER_DEVICE;
	}
	if (major == 0) major = res;
	
	SET_MODULE_OWNER(&led_fops);
	
	printk(KERN_NOTICE "Ledmatrix device registered. major: %d.\n", major);
	
	
	/* create PROC entry */
	inca_led_proc = create_proc_entry( "driver/ledmatrix", 0, 0);
	if (!inca_led_proc) {
		printk(KERN_ERR "could not create /proc/driver/ledmatrix!\n");
		res = -ENOENT;
		goto FAIL_CREATE_PROC;
	}
	inca_led_proc->read_proc  = inca_led_read_procmem;
	inca_led_proc->write_proc = inca_led_write_procmem;

	printk(KERN_NOTICE "/proc/driver/ledmatrix created!\n");
	
	/* success */
	printk(KERN_NOTICE "Inca-IP LED Matrix initialized with pins "
					"LED[%d:0].\n", led_pins-1);
	
	return 0; 


	/* if initialization fails, release all resources aquired so far! */
FAIL_CREATE_PROC:
	unregister_chrdev(major, "ledmatrix");
FAIL_REGISTER_DEVICE:
	release_region((u32) INCA_IP_TSF_LEDMUX_CONF, 0x08);
FAIL_REQUEST_REGION:
	return res;
}

static void __exit inca_led_exit(void)
{
	unsigned long flags;
	
	/* clear all LEDs */
	*INCA_IP_TSF_LED_REG = 0x00000000;

	/* reset configuration */
	*INCA_IP_TSF_LEDMUX_CONF = 0x00000000;

	/* disable the LED module */
 	save_flags(flags); cli();
	if (activate_led9) *INCA_IP_Ports_P0_ALTSEL &= ~1;
#ifndef DONE_ELSEWHERE
	*INCA_IP_PMU_PM_GEN      &= ~INCA_IP_PMU_PM_GEN_EN9;
#endif
	*INCA_IP_TSF_TSF_CONF    &= ~INCA_IP_TSF_TSF_CONF_LEDEN;
 	restore_flags(flags);

	/* Release aquired resources (in inverse order) */
	remove_proc_entry("driver/ledmatrix", 0);

	unregister_chrdev(major, "ledmatrix");

	release_region((u32) INCA_IP_TSF_LEDMUX_CONF, 0x08);
	
	printk(KERN_NOTICE "/proc/driver/ledmatrix removed!\n");

}

#ifdef MODULE
module_init(inca_led_init);
module_exit(inca_led_exit);
#endif
