/*
 * ########################################################################
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * ########################################################################
 *
 * port.c
 *
 * Inca-IP port driver
 *
 * Andre Messerschmidt 
 * Infineon Technologies AG (C) 2003
 *
 * For questions about this file please contact your local Infineon 
 * representative.
 *
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/ioctl.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>
#include <asm/incaip/inca-ip.h>
#include <asm/incaip/mux.h>
#include <asm/incaip/port.h>
#include "port_defs.h"

//DISP_GS_240x128
/* PINS */
#define PORT1 1
#define PORT2 2

#define GS_CS 9  //2.9
#define GS_D0 13 //1.13
#define GS_D1 14 //1.14
#define GS_D2 15 //1.15
#define GS_D3 16 //1.16
#define GS_D4 12 //2.12
#define GS_D5 13 //2.13
#define GS_D6 14 //2.14
#define GS_D7 15 //2.15
#define GS_DC 6  //2.6
#define GS_WR 10 //2.10
#define GS_RD 2  //2.2
#define GS_RST 8 //2.8

#define DELAY //for(i=0;i<300;i++);

//SNOM360
#define SDA 10
#define CLK 11
#define CS 9
#define RS 6
#define RST 8
#define LGT 7

#define SERIAL_DELAY ; //for(i=0;i<50;i++)

#define MAX_PORTS	3	// Number of ports in system
#define PINS_PER_PORT	32	// Number of available pins per port
//#define PORT_DEV_MAJOR 251

static int port_major = 0;
static int inca_port_pin_usage[MAX_PORTS][PINS_PER_PORT]; // Map for pin usage
static u32 inca_port_bases[MAX_PORTS] 
	= {INCA_IP_Ports, \
	   INCA_IP_Ports + 0x00000020, \
	   INCA_IP_Ports + 0x00000040}; // Base addresses for ports

static struct semaphore port_sem;

/***************************************************************************/
/**                                                                       **/
/** Port usage                                                            **/
/**                                                                       **/
/***************************************************************************/

/**
 *	inca_port_reserve_pin -	Reserve port pin for usage
 *	@port:      Port number
 *	@pin:       Pin to be reserved
 *	@module_id: Module ID to identify the owner of a pin
 *
 *	This function reserves a given pin for usage by the given module.
 *
 *  	Return Value:	
 *	@EBUSY = Pin allready used.
 *	@EINVAL = Invalid port or pin provided.
 *	@OK = OK, pin reserved.
 */
int inca_port_reserve_pin(int port, int pin, int module_id)
{

	if (module_id != PORT_MODULE_ID) {
		if (down_trylock(&port_sem) != 0) return -EBUSY;
	}
	
	if (port < 0 || pin <0) return -EINVAL;
	if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

	if (inca_port_pin_usage[port][pin] == 0) {
		inca_port_pin_usage[port][pin] = module_id;
	}
	else {
		return -EBUSY;
	}
	
	return OK;
}

/**
 *	inca_port_free_pin -	Free port pin
 *	@port:      Port number
 *	@pin:       Pin to be reserved
 *	@module_id: Module ID to identify the owner of a pin
 *
 *	This functions frees a port pin and thus clears the entry in the 
 *	usage map.
 *
 *	Return Value:	
 *	@EBUSY = Pin used by another module than the given ID.
 *	@EINVAL = Invalid port or pin provided.
 *	@OK = OK, pin freed.
 */
int inca_port_free_pin(int port, int pin, int module_id)
{

	if (port < 0 || pin < 0) return -EINVAL;
	if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

	if (inca_port_pin_usage[port][pin] != module_id) return -EBUSY;
	inca_port_pin_usage[port][pin] = 0;

	if (module_id != PORT_MODULE_ID) {
		up(&port_sem);
	}
	return OK;
}

/**
 *	inca_port_set_open_drain - Enable Open Drain for given pin
 *	@port:      Port number
 *	@pin:       Pin to be reserved
 *	@module_id: Module ID to identify the owner of a pin
 *
 *	This function sets Open Drain mode for the given pin.
 *
 *	Return Value:	
 *	@EBUSY = Pin used by another module.
 *	@EINVAL = Invalid port or pin provided.
 *	@OK = OK
 */
int inca_port_set_open_drain(int port, int pin, int module_id)
{
	u32	reg;

	if (port < 0 || pin < 0) return -EINVAL;
	if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

	if (inca_port_pin_usage[port][pin] != module_id) return -EBUSY;

	/* Check for valid port */
	if ((port == 0x00) || (port == 0x01))  return -EINVAL;
	
	/* Check for valid pin */
	if ((pin != 0x00) && (pin != 0x01))  return -EINVAL;
	
	PORT_READ_REG(inca_port_bases[port] + INCA_PORT_OD_REG,reg);
	reg |= (1 << pin);
	PORT_WRITE_REG(inca_port_bases[port] + INCA_PORT_OD_REG,reg);

	return OK;
}

/**
 *	inca_port_clear_open_drain - Disable Open Drain for given pin
 *	@port:      Port number
 *	@pin:       Pin to be reserved
 *	@module_id: Module ID to identify the owner of a pin
 *
 *	This function clears Open Drain mode for the given pin.
 *
 *  	Return Value:	
 *	@EBUSY = Pin used by another module.
 *	@EINVAL = Invalid port or pin provided.
 *	@OK = OK
 */
int inca_port_clear_open_drain(int port, int pin, int module_id)
{
	u32	reg;

	if (port < 0 || pin < 0) return -EINVAL;
	if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

	if (inca_port_pin_usage[port][pin] != module_id) return -EBUSY;

	/* Check for valid port */
	if ((port == 0x00) || (port == 0x01))  return -EINVAL;
	
	/* Check for valid pin */
	if ((pin != 0x00) && (pin != 0x01))  return -EINVAL;

	PORT_READ_REG(inca_port_bases[port] + INCA_PORT_OD_REG,reg);
	reg &= ~(1 << pin);
	PORT_WRITE_REG(inca_port_bases[port] + INCA_PORT_OD_REG,reg);

	return OK;
}

/**
 *	inca_port_set_pudsel - Enable PUDSEL for given pin
 *	@port:      Port number
 *	@pin:       Pin to be reserved
 *	@module_id: Module ID to identify the owner of a pin
 *
 *	This function sets the PUDSEL bit for the given pin.
 *
 *  	Return Value:	
 *	@EBUSY = Pin used by another module.
 *	@EINVAL = Invalid port or pin provided.
 *	@OK = OK
 */
int inca_port_set_pudsel(int port, int pin, int module_id)
{
	u32	reg;

	if (port < 0 || pin < 0) return -EINVAL;
	if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

	if (inca_port_pin_usage[port][pin] != module_id) return -EBUSY;

	/* Check for valid port */
	if (port == 0x01)  return -EINVAL;

	/* Check for valid pins */
	if (port == 0x00) {
		if ((pin > 5) || (pin == 0x00)) return -EINVAL;	
	}
	
	if (port == 0x02) {
		if ((pin == 0x00) || (pin == 0x01))  return -EINVAL;
		if ((pin > 21) || (pin < 13)) return -EINVAL;	
	}
	
	PORT_READ_REG(inca_port_bases[port] + INCA_PORT_PUDSEL_REG,reg);
	reg |= (1 << pin);
	PORT_WRITE_REG(inca_port_bases[port] + INCA_PORT_PUDSEL_REG,reg);

	return OK;
}

/**
 *	inca_port_clear_pudsel - Clear PUDSEL bit for given pin
 *	@port:      Port number
 *	@pin:       Pin to be reserved
 *	@module_id: Module ID to identify the owner of a pin
 *
 *	This function clears the PUDSEL bit for the given pin.
 *
 *  	Return Value:	
 *	@EBUSY = Pin used by another module.
 *	@EINVAL = Invalid port or pin provided.
 *	@OK = OK
 */
int inca_port_clear_pudsel(int port, int pin, int module_id)
{
	u32	reg;

	if (port < 0 || pin < 0) return -EINVAL;
	if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

	if (inca_port_pin_usage[port][pin] != module_id) return -EBUSY;

	/* Check for valid port */
	if (port == 0x01)  return -EINVAL;

	/* Check for valid pins */
	if (port == 0x00) {
		if ((pin > 5) || (pin == 0x00)) return -EINVAL;	
	}
	
	if (port == 0x02) {
		if ((pin == 0x00) || (pin == 0x01))  return -EINVAL;
		if ((pin > 21) || (pin < 13)) return -EINVAL;	
	}
	
	PORT_READ_REG(inca_port_bases[port] + INCA_PORT_PUDSEL_REG,reg);
	reg &= ~(1 << pin);
	PORT_WRITE_REG(inca_port_bases[port] + INCA_PORT_PUDSEL_REG,reg);

	return OK;
}


/**
 *	inca_port_set_puden - Set PUDEN bit for given pin
 *	@port:      Port number
 *	@pin:       Pin to be reserved
 *	@module_id: Module ID to identify the owner of a pin
 *
 *	This function sets the PUDEN bit for the given pin.
 *
 *  	Return Value:	
 *	@EBUSY = Pin used by another module.
 *	@EINVAL = Invalid port or pin provided.
 *	@OK = OK
 */
int inca_port_set_puden(int port, int pin, int module_id)
{
	u32	reg;

	if (port < 0 || pin < 0) return -EINVAL;
	if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

	if (inca_port_pin_usage[port][pin] != module_id) return -EBUSY;

	/* Check for valid port */
	if (port == 0x01)  return -EINVAL;

	/* Check for valid pins */
	if (port == 0x00) {
		if ((pin > 5) || (pin == 0x00)) return -EINVAL;	
	}
	
	if (port == 0x02) {
		if (pin > 21) return -EINVAL;	
	}
	
	PORT_READ_REG(inca_port_bases[port] + INCA_PORT_PUDEN_REG,reg);
	reg |= (1 << pin);
	PORT_WRITE_REG(inca_port_bases[port] + INCA_PORT_PUDEN_REG,reg);

	return OK;
}

/**
 *	inca_port_clear_puden - Disable PUDEN bit for given pin
 *	@port:      Port number
 *	@pin:       Pin to be reserved
 *	@module_id: Module ID to identify the owner of a pin
 *
 *	This function clears the PUDEN bit for the given pin.
 *
 *  	Return Value:	
 *	@EBUSY = Pin used by another module.
 *	@EINVAL = Invalid port or pin provided.
 *	@OK = OK
 */
int inca_port_clear_puden(int port, int pin, int module_id)
{
	u32	reg;

	if (port < 0 || pin < 0) return -EINVAL;
	if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

	if (inca_port_pin_usage[port][pin] != module_id) return -EBUSY;

	/* Check for valid port */
	if (port == 0x01)  return -EINVAL;

	/* Check for valid pins */
	if (port == 0x00) {
		if ((pin > 5) || (pin == 0x00)) return -EINVAL;	
	}
	
	if (port == 0x02) {
		if (pin > 21) return -EINVAL;	
	}
	
	PORT_READ_REG(inca_port_bases[port] + INCA_PORT_PUDEN_REG,reg);
	reg &= ~(1 << pin);
	PORT_WRITE_REG(inca_port_bases[port] + INCA_PORT_PUDEN_REG,reg);

	return OK;
}

/**
 *	inca_port_set_stoff - Enable Schmitt Trigger for given pin
 *	@port:      Port number
 *	@pin:       Pin to be reserved
 *	@module_id: Module ID to identify the owner of a pin
 *
 *	This function enables the Schmitt Trigger for the given pin.
 *
 *  	Return Value:	
 *	@EBUSY = Pin used by another module.
 *	@EINVAL = Invalid port or pin provided.
 *	@OK = OK
 */
int inca_port_set_stoff(int port, int pin, int module_id)
{
	u32	reg;

	if (port < 0 || pin < 0) return -EINVAL;
	if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

	if (inca_port_pin_usage[port][pin] != module_id) return -EBUSY;

	/* Check for valid pins */
	if (port == 0x00) {
		if (pin > 8) return -EINVAL;	
	}

	if (port == 0x01) {
		if (pin > 17) return -EINVAL;	
	}
	
	if (port == 0x02) {
		if (pin > 15) return -EINVAL;	
	}
	
	PORT_READ_REG(inca_port_bases[port] + INCA_PORT_STOFF_REG,reg);
	reg |= (1 << pin);
	PORT_WRITE_REG(inca_port_bases[port] + INCA_PORT_STOFF_REG,reg);

	return OK;
}

/**
 *	inca_port_clear_stoff - Disable Schmitt Trigger for given pin
 *	@port:      Port number
 *	@pin:       Pin to be reserved
 *	@module_id: Module ID to identify the owner of a pin
 *
 *	This function disables the Schmitt Trigger for the given pin.
 *
 *  	Return Value:	
 *	@EBUSY = Pin used by another module.
 *	@EINVAL = Invalid port or pin provided.
 *	@OK = OK
 */
int inca_port_clear_stoff(int port, int pin, int module_id)
{
	u32	reg;

	if (port < 0 || pin < 0) return -EINVAL;
	if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

	if (inca_port_pin_usage[port][pin] != module_id) return -EBUSY;

	/* Check for valid pins */
	if (port == 0x00) {
		if (pin > 8) return -EINVAL;	
	}

	if (port == 0x01) {
		if (pin > 17) return -EINVAL;	
	}
	
	if (port == 0x02) {
		if (pin > 15) return -EINVAL;	
	}
	
	PORT_READ_REG(inca_port_bases[port] + INCA_PORT_STOFF_REG,reg);
	reg &= ~(1 << pin);
	PORT_WRITE_REG(inca_port_bases[port] + INCA_PORT_STOFF_REG,reg);

	return OK;
}

/**
 *	inca_port_set_dir_out - Set direction to output for given pin
 *	@port:      Port number
 *	@pin:       Pin to be reserved
 *	@module_id: Module ID to identify the owner of a pin
 *
 *	This function sets the direction for the given pin to output.
 *
 *  	Return Value:	
 *	@EBUSY = Pin used by another module.
 *	@EINVAL = Invalid port or pin provided.
 *	@OK = OK
 */
int inca_port_set_dir_out(int port, int pin, int module_id)
{
	u32	reg;

	if (port < 0 || pin < 0) return -EINVAL;
	if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

	if (inca_port_pin_usage[port][pin] != module_id) return -EBUSY;

	/* Check for valid port */
	if (port == 0x00) return -EINVAL;

	/* Check for valid pins */
	if (port == 0x01) {
		if (pin > 17) return -EINVAL;	
	}
	
	if (port == 0x02) {
		if (pin > 15) return -EINVAL;	
	}
	
	PORT_READ_REG(inca_port_bases[port] + INCA_PORT_DIR_REG,reg);
	reg |= (1 << pin);
	PORT_WRITE_REG(inca_port_bases[port] + INCA_PORT_DIR_REG,reg);

	return OK;
}

/**
 *	inca_port_set_dir_in - Set direction to input for given pin
 *	@port:      Port number
 *	@pin:       Pin to be reserved
 *	@module_id: Module ID to identify the owner of a pin
 *
 *	This function sets the direction for the given pin to input.
 *
 *  	Return Value:	
 *	@EBUSY = Pin used by another module.
 *	@EINVAL = Invalid port or pin provided.
 *	@OK = OK
 */
int inca_port_set_dir_in(int port, int pin, int module_id)
{
	u32	reg;

	if (port < 0 || pin < 0) return -EINVAL;
	if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

	if (inca_port_pin_usage[port][pin] != module_id) return -EBUSY;

	/* Check for valid port */
	if (port == 0x00) return -EINVAL;

	/* Check for valid pins */
	if (port == 0x01) {
		if (pin > 17) return -EINVAL;	
	}
	
	if (port == 0x02) {
		if (pin > 15) return -EINVAL;	
	}
	
	PORT_READ_REG(inca_port_bases[port] + INCA_PORT_DIR_REG,reg);
	reg &= ~(1 << pin);
	PORT_WRITE_REG(inca_port_bases[port] + INCA_PORT_DIR_REG,reg);

	return OK;
}

/**
 *	inca_port_set_output - Set output bit for given pin
 *	@port:      Port number
 *	@pin:       Pin to be reserved
 *	@module_id: Module ID to identify the owner of a pin
 *
 *	This function sets the output bit to 1 for the given pin.
 *
 *  	Return Value:	
 *	@EBUSY = Pin used by another module.
 *	@EINVAL = Invalid port or pin provided.
 *	@OK = OK
 */
int inca_port_set_output(int port, int pin, int module_id)
{
	u32	reg;

	if (port < 0 || pin <0) return -EINVAL;
	if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

	if (inca_port_pin_usage[port][pin] != module_id) return -EBUSY;
	
	/* Check for valid port */
	if (port == 0x00) return -EINVAL;

	/* Check for valid pins */
	if (port == 0x01) {
		if (pin > 17) return -EINVAL;	
	}
	
	if (port == 0x02) {
		if (pin > 15) return -EINVAL;	
	}
	PORT_READ_REG(inca_port_bases[port] + INCA_PORT_OUT_REG,reg);
	reg |= (1 << pin);
	PORT_WRITE_REG(inca_port_bases[port] + INCA_PORT_OUT_REG,reg);

	return OK;
}

/**
 *	inca_port_clear_output - Clear output bit for given pin
 *	@port:      Port number
 *	@pin:       Pin to be reserved
 *	@module_id: Module ID to identify the owner of a pin
 *
 *	This function sets the output bit to 0 for the given pin.
 *
 *  	Return Value:	
 *	@EBUSY = Pin used by another module.
 *	@EINVAL = Invalid port or pin provided.
 *	@OK = OK
 */
int inca_port_clear_output(int port, int pin, int module_id)
{
	u32	reg;

	if (port < 0 || pin < 0) return -EINVAL;
	if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

	if (inca_port_pin_usage[port][pin] != module_id) return -EBUSY;

	/* Check for valid port */
	if (port == 0x00) return -EINVAL;

	/* Check for valid pins */
	if (port == 0x01) {
		if (pin > 17) return -EINVAL;	
	}
	
	if (port == 0x02) {
		if (pin > 15) return -EINVAL;	
	}
	
	PORT_READ_REG(inca_port_bases[port] + INCA_PORT_OUT_REG,reg);
	reg &= ~(1 << pin);
	PORT_WRITE_REG(inca_port_bases[port] + INCA_PORT_OUT_REG,reg);

	return OK;
}

/**
 *	inca_port_get_input - Get input bit for given pin
 *	@port:      Port number
 *	@pin:       Pin to be reserved
 *	@module_id: Module ID to identify the owner of a pin
 *
 *	This function gets the value of the given pin.
 *
 *  	Return Value:	
 *	@EBUSY = Pin used by another module.
 *	@EINVAL = Invalid port or pin provided.
 *	@0 = Pin is set to 0
 *	@1 = Pin is set to 1
 */
int inca_port_get_input(int port, int pin, int module_id)
{
	u32	reg;

	if (port < 0 || pin < 0) return -EINVAL;
	if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

	if (inca_port_pin_usage[port][pin] != module_id) return -EBUSY;

	/* Check for valid port */
	if (port == 0x00) return -EINVAL;

	/* Check for valid pins */
	if (port == 0x01) {
		if (pin > 17) return -EINVAL;	
	}
	
	if (port == 0x02) {
		if (pin > 15) return -EINVAL;	
	}

	PORT_READ_REG(inca_port_bases[port] + INCA_PORT_IN_REG,reg);
	reg &= (1 << pin);
	if (reg == 0x00) return 0;
	else return 1;
}

/**
 *	inca_port_read_procmem - Create proc file output
 *	@buf:      Buffer to write the string to
 *	@start:    not used (Linux internal)
 *	@offset:   not used (Linux internal)
 *	@count:    not used (Linux internal)
 *	@eof:      Set to 1 when all data is stored in buffer
 *	@data:     not used (Linux internal)
 *
 *	This function creates the output for the port proc file, by reading
 *	all appropriate registers and displaying the content as a table.
 *
 *  	Return Value:	
 *	@len = Lenght of data in buffer
 */
int inca_port_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	long	len = 0;
	int 	t=0;
	u32	bit=0;
	u32	reg=0;

	len = sprintf(buf,"\nInca-IP Port Settings\n");

	len=len+sprintf(buf+len,"         3         2         1         0\n");
	len=len+sprintf(buf+len,"        10987654321098765432109876543210\n");
	len=len+sprintf(buf+len,"----------------------------------------\n");
	len=len+sprintf(buf+len,"P1-OUT: ");
	PORT_READ_REG(INCA_IP_Ports_P1_OUT,reg);
	bit = 0x80000000;
	for (t=0;t<32;t++) {
		if ((reg & bit) > 0) len=len+sprintf(buf+len,"X");
		else len=len+sprintf(buf+len," ");
		bit = bit >> 1;
	}

	len=len+sprintf(buf+len,"\nP2-OUT: ");
	PORT_READ_REG(INCA_IP_Ports_P2_OUT,reg);
	bit = 0x80000000;
	for (t=0;t<32;t++) {
		if ((reg & bit) > 0) len=len+sprintf(buf+len,"X");
		else len=len+sprintf(buf+len," ");
		bit = bit >> 1;
	}

	len=len+sprintf(buf+len,"\nP1-IN:  ");
	PORT_READ_REG(INCA_IP_Ports_P1_IN,reg);
	bit = 0x80000000;
	for (t=0;t<32;t++) {
		if ((reg & bit) > 0) len=len+sprintf(buf+len,"X");
		else len=len+sprintf(buf+len," ");
		bit = bit >> 1;
	}

	len=len+sprintf(buf+len,"\nP2-IN:  ");
	PORT_READ_REG(INCA_IP_Ports_P2_IN,reg);
	bit = 0x80000000;
	for (t=0;t<32;t++) {
		if ((reg & bit) > 0) len=len+sprintf(buf+len,"X");
		else len=len+sprintf(buf+len," ");
		bit = bit >> 1;
	}

	len=len+sprintf(buf+len,"\nP1-DIR: ");
	PORT_READ_REG(INCA_IP_Ports_P1_DIR,reg);
	bit = 0x80000000;
	for (t=0;t<32;t++) {
		if ((reg & bit) > 0) len=len+sprintf(buf+len,"X");
		else len=len+sprintf(buf+len," ");
		bit = bit >> 1;
	}

	len=len+sprintf(buf+len,"\nP2-DIR: ");
	PORT_READ_REG(INCA_IP_Ports_P2_DIR,reg);
	bit = 0x80000000;
	for (t=0;t<32;t++) {
		if ((reg & bit) > 0) len=len+sprintf(buf+len,"X");
		else len=len+sprintf(buf+len," ");
		bit = bit >> 1;
	}

	len=len+sprintf(buf+len,"\nP0-STO: ");
	PORT_READ_REG(INCA_IP_Ports_P0_STOFF,reg);
	bit = 0x80000000;
	for (t=0;t<32;t++) {
		if ((reg & bit) > 0) len=len+sprintf(buf+len,"X");
		else len=len+sprintf(buf+len," ");
		bit = bit >> 1;
	}

	len=len+sprintf(buf+len,"\nP1-STO: ");
	PORT_READ_REG(INCA_IP_Ports_P1_STOFF,reg);
	bit = 0x80000000;
	for (t=0;t<32;t++) {
		if ((reg & bit) > 0) len=len+sprintf(buf+len,"X");
		else len=len+sprintf(buf+len," ");
		bit = bit >> 1;
	}

	len=len+sprintf(buf+len,"\nP2-STO: ");
	PORT_READ_REG(INCA_IP_Ports_P2_STOFF,reg);
	bit = 0x80000000;
	for (t=0;t<32;t++) {
		if ((reg & bit) > 0) len=len+sprintf(buf+len,"X");
		else len=len+sprintf(buf+len," ");
		bit = bit >> 1;
	}

	len=len+sprintf(buf+len,"\nP2-OD:  ");
	PORT_READ_REG(INCA_IP_Ports_P2_OD,reg);
	bit = 0x80000000;
	for (t=0;t<32;t++) {
		if ((reg & bit) > 0) len=len+sprintf(buf+len,"X");
		else len=len+sprintf(buf+len," ");
		bit = bit >> 1;
	}

	len=len+sprintf(buf+len,"\nP0-PUDE:");
	PORT_READ_REG(INCA_IP_Ports_P0_PUDEN,reg);
	bit = 0x80000000;
	for (t=0;t<32;t++) {
		if ((reg & bit) > 0) len=len+sprintf(buf+len,"X");
		else len=len+sprintf(buf+len," ");
		bit = bit >> 1;
	}

	len=len+sprintf(buf+len,"\nP2-PUDE:");
	PORT_READ_REG(INCA_IP_Ports_P2_PUDEN,reg);
	bit = 0x80000000;
	for (t=0;t<32;t++) {
		if ((reg & bit) > 0) len=len+sprintf(buf+len,"X");
		else len=len+sprintf(buf+len," ");
		bit = bit >> 1;
	}

	len=len+sprintf(buf+len,"\nP0-PUDS:");
	PORT_READ_REG(INCA_IP_Ports_P0_PUDSEL,reg);
	bit = 0x80000000;
	for (t=0;t<32;t++) {
		if ((reg & bit) > 0) len=len+sprintf(buf+len,"X");
		else len=len+sprintf(buf+len," ");
		bit = bit >> 1;
	}

	len=len+sprintf(buf+len,"\nP2-PUDS:");
	PORT_READ_REG(INCA_IP_Ports_P2_PUDSEL,reg);
	bit = 0x80000000;
	for (t=0;t<32;t++) {
		if ((reg & bit) > 0) len=len+sprintf(buf+len,"X");
		else len=len+sprintf(buf+len," ");
		bit = bit >> 1;
	}
	
	len=len+sprintf(buf+len,"\n\n");

	*eof = 1;	// No more data available
	return len;
}

static int inca_port_open(struct inode *inode, struct file *filep)
{
	if (down_trylock(&port_sem) != 0) return -EBUSY;

	return OK;
}

static int inca_port_release(struct inode *inode, struct file *filelp)
{
	up(&port_sem);
	
	return OK;
}

static int write_gs_command(int val){
  int ret = 0;
  int disp_command = val;
  if(disp_command&(1<<0)){ 
    PORT_IOC_CALL(ret, PORT1, GS_D0, inca_port_set_output);
  }
  else{
    PORT_IOC_CALL(ret, PORT1, GS_D0, inca_port_clear_output);
  }
  if(disp_command&(1<<1)){
    PORT_IOC_CALL(ret, PORT1, GS_D1, inca_port_set_output);
  }
  else{
    PORT_IOC_CALL(ret, PORT1, GS_D1, inca_port_clear_output);
  }
  if(disp_command&(1<<2)){
    PORT_IOC_CALL(ret, PORT1, GS_D2, inca_port_set_output);
  }
  else{
    PORT_IOC_CALL(ret, PORT1, GS_D2, inca_port_clear_output);
  }
  if(disp_command&(1<<3)){
    PORT_IOC_CALL(ret, PORT1, GS_D3, inca_port_set_output);
  }
  else{
    PORT_IOC_CALL(ret, PORT1, GS_D3, inca_port_clear_output);
  }
  if(disp_command&(1<<4)){
    PORT_IOC_CALL(ret, PORT2, GS_D4, inca_port_set_output);
  }
  else{
    PORT_IOC_CALL(ret, PORT2, GS_D4, inca_port_clear_output);
  }
  if(disp_command&(1<<5)){
    PORT_IOC_CALL(ret, PORT2, GS_D5, inca_port_set_output);
  }
  else{
    PORT_IOC_CALL(ret, PORT2, GS_D5, inca_port_clear_output);
  }
  if(disp_command&(1<<6)){
    PORT_IOC_CALL(ret, PORT2, GS_D6, inca_port_set_output);
  }
  else{
    PORT_IOC_CALL(ret, PORT2, GS_D6, inca_port_clear_output);
  }
  if(disp_command&(1<<7)){
    PORT_IOC_CALL(ret, PORT2, GS_D7, inca_port_set_output);
  }
  else{
    PORT_IOC_CALL(ret, PORT2, GS_D7, inca_port_clear_output);
  }
  PORT_IOC_CALL(ret, PORT2, GS_CS, inca_port_clear_output);
  PORT_IOC_CALL(ret, PORT2, GS_DC, inca_port_clear_output);
  PORT_IOC_CALL(ret, PORT2, GS_RD, inca_port_set_output);
  PORT_IOC_CALL(ret, PORT2, GS_WR, inca_port_clear_output);
  DELAY;
  PORT_IOC_CALL(ret, PORT2, GS_WR, inca_port_set_output);
  PORT_IOC_CALL(ret, PORT2, GS_DC, inca_port_set_output);
  PORT_IOC_CALL(ret, PORT2, GS_CS, inca_port_set_output);
  return ret;
}

static int write_gs_data(int val){
  int ret = 0;
  int disp_data = val;
  if(disp_data&(1<<0)){
    PORT_IOC_CALL(ret, PORT1, GS_D0, inca_port_set_output);
  }
  else{
    PORT_IOC_CALL(ret, PORT1, GS_D0, inca_port_clear_output);
  }
  if(disp_data&(1<<1)){
    PORT_IOC_CALL(ret, PORT1, GS_D1, inca_port_set_output);
  }
  else{
    PORT_IOC_CALL(ret, PORT1, GS_D1, inca_port_clear_output);
  }
  if(disp_data&(1<<2)){
    PORT_IOC_CALL(ret, PORT1, GS_D2, inca_port_set_output);
  }
  else{
    PORT_IOC_CALL(ret, PORT1, GS_D2, inca_port_clear_output);
  }
  if(disp_data&(1<<3)){
    PORT_IOC_CALL(ret, PORT1, GS_D3, inca_port_set_output);
  }
  else{
    PORT_IOC_CALL(ret, PORT1, GS_D3, inca_port_clear_output);
  }
  if(disp_data&(1<<4)){
    PORT_IOC_CALL(ret, PORT2, GS_D4, inca_port_set_output);
  }
  else{
    PORT_IOC_CALL(ret, PORT2, GS_D4, inca_port_clear_output);
  }
  if(disp_data&(1<<5)){
    PORT_IOC_CALL(ret, PORT2, GS_D5, inca_port_set_output);
  }
  else{
    PORT_IOC_CALL(ret, PORT2, GS_D5, inca_port_clear_output);
  }
  if(disp_data&(1<<6)){
    PORT_IOC_CALL(ret, PORT2, GS_D6, inca_port_set_output);
  }
  else{
    PORT_IOC_CALL(ret, PORT2, GS_D6, inca_port_clear_output);
  }
  if(disp_data&(1<<7)){
    PORT_IOC_CALL(ret, PORT2, GS_D7, inca_port_set_output);
  }
  else{
    PORT_IOC_CALL(ret, PORT2, GS_D7, inca_port_clear_output);
  }
  PORT_IOC_CALL(ret, PORT2, GS_CS, inca_port_clear_output);
  PORT_IOC_CALL(ret, PORT2, GS_DC, inca_port_set_output);
  PORT_IOC_CALL(ret, PORT2, GS_RD, inca_port_set_output);
  PORT_IOC_CALL(ret, PORT2, GS_WR, inca_port_clear_output);
  DELAY;
  PORT_IOC_CALL(ret, PORT2, GS_WR, inca_port_set_output);
  PORT_IOC_CALL(ret, PORT2, GS_CS, inca_port_set_output);
  return ret;
}

static inline u32 place_data(const int val){
  u32 reg;

  PORT_READ_REG(inca_port_bases[PORT1] + INCA_PORT_OUT_REG,reg);
  reg = (reg & 0xFFFE1FFF)|((val&0x0000000F)<<13); //zeroing out 13,14,15,16 and placing value
  PORT_WRITE_REG(inca_port_bases[PORT1] + INCA_PORT_OUT_REG,reg);

  PORT_READ_REG(inca_port_bases[PORT2] + INCA_PORT_OUT_REG,reg);
  reg = (reg & 0xFFFF0FFF)|((val&0x000000F0)<<8); //zeroing out 12,13,14,15 and placing value
  return reg;
}

#ifdef DISP_GS_240x128
static inline void write_direct_command(const int val){
  u32 reg;
  
  reg = place_data(val);
  reg &= ~(0x00000640); //clear CS,DC and WR
  reg |= 0x00000004;    //RD
  PORT_WRITE_REG(inca_port_bases[PORT2] + INCA_PORT_OUT_REG,reg);

  PORT_READ_REG(inca_port_bases[PORT2] + INCA_PORT_OUT_REG,reg);
  reg |= 0x00000640; //set CS,DC and WR
  PORT_WRITE_REG(inca_port_bases[PORT2] + INCA_PORT_OUT_REG,reg);
}

static inline void write_direct_data(const int val){
  u32 reg;
  
  reg = place_data(val);
  reg &= ~(0x00000600); //clear CS and WR
  reg |= 0x00000044;    //set DC and RD
  PORT_WRITE_REG(inca_port_bases[PORT2] + INCA_PORT_OUT_REG,reg);

  PORT_READ_REG(inca_port_bases[PORT2] + INCA_PORT_OUT_REG,reg);
  reg |= 0x00000600; //set CS and WR
  PORT_WRITE_REG(inca_port_bases[PORT2] + INCA_PORT_OUT_REG,reg);
}

/*int write_gs_mdata(volatile struct inca_port_ioctl_parm parm){
  int i, j;
  unsigned int on = parm.port;
  unsigned int off = parm.pin;
  
  for(j = 0; j<8; j++)
    for(i = 0; i<30; i++)
      if(parm.disp_block[i]&(1<<j)) write_gs_data(on); else write_gs_data(off);
}*/

static inline int write_gs_mdata(volatile struct inca_port_ioctl_parm parm){
  int i;
  unsigned int on = parm.port;
  unsigned int off = parm.pin;
  unsigned int disp_data = parm.value;
  
  for(i = 0; i<30; i++) //XFLUSHGRAN is 30 (x length of block)
    if(disp_data&(1<<i)) 
      //write_gs_data(on); 
      write_direct_data(on);
    else 
      //write_gs_data(off);
      write_direct_data(off);

  return 0;
}

static inline int write_gs_mdata_2b3p(volatile struct inca_port_ioctl_parm parm){
  int i;
  unsigned int on = parm.port;
  unsigned int off = parm.pin;
  unsigned int disp_data = parm.value;
  unsigned int pb1 = 0;
  unsigned int pb2 = 0;
  
  //2B3P scheme
  for(i = 0; i<30; i++){ //XFLUSHGRAN is 30 (x length of block)
    pb1 = 0; pb2 = 0;
    if(disp_data&(1<<i))
      pb1 = on&0xF8; 
    else 
      pb1 = off&0xF8; 
    i++;
    if(disp_data&(1<<i)){
      pb1 |= (on>>5)&0x07; 
      pb2 = (on<<3)&0xC0;
    } 
    else {
      pb1 |= (off>>5)&0x07;
      pb2 = (off<<3)&0xC0;
    }
    i++;
    if(disp_data&(1<<i))
      pb2 |= (on>>3)&0x1F;
    else
      pb2 |= (off>>3)&0x1F;
    //write_gs_data(pb1); 
    //write_gs_data(pb2);
    write_direct_data(pb1);
    write_direct_data(pb2);
  }
  return 0;
}

static inline int write_gs_dispbytes(volatile struct inca_port_ioctl_parm parm){
  int i;
  unsigned int nbytes = parm.port;
  unsigned int mode = parm.pin;
  unsigned int pb1 = 0;
  unsigned int pb2 = 0;

  if(nbytes > GS_DISP_DATA_SIZE)
    return -1;

  if(mode == 2){ //3b3p
    for(i=0; i<nbytes; i++)
      write_direct_data(parm.disp_bytes[i]);
  }
  else{ //2b3p
    for(i = 0; i<nbytes; i++){
      pb1 = 0; pb2 = 0;
      pb1 = parm.disp_bytes[i]&0xF8; 
      i++;
      pb1 |= (parm.disp_bytes[i]>>5)&0x07; 
      pb2 = (parm.disp_bytes[i]<<3)&0xC0;
      i++;
      pb2 |= (parm.disp_bytes[i]>>3)&0x1F;
      //write_gs_data(pb1); 
      //write_gs_data(pb2);
      write_direct_data(pb1);
      write_direct_data(pb2);
    }
  }
}
#endif

static int inca_port_ioctl(struct inode *inode, struct file *filp, 
					unsigned int cmd, unsigned long arg) 
{
	int ret = 0;
	volatile struct inca_port_ioctl_parm parm;

	int disp_command = 0;
	int disp_data = 0;
	int i;
	
	if (_IOC_TYPE(cmd) != INCA_PORT_IOC_MAGIC) return -EINVAL;
	
	if (_IOC_DIR(cmd) & _IOC_WRITE) {
		if (!access_ok(VERIFY_READ, arg, sizeof(struct inca_port_ioctl_parm))) 
			return -EFAULT;
		ret=copy_from_user((void*)&parm,(void*)arg,sizeof(struct inca_port_ioctl_parm));
	}
	if (_IOC_DIR(cmd) & _IOC_READ) {
		if (!access_ok(VERIFY_WRITE, arg, sizeof(struct inca_port_ioctl_parm))) 
			return -EFAULT;
	}

	switch (cmd) {
		case INCA_PORT_IOCOD:
			if (parm.value == 0x00) {
				PORT_IOC_CALL(ret,parm.port,parm.pin,inca_port_clear_open_drain);
			}
			else {
				PORT_IOC_CALL(ret,parm.port,parm.pin,inca_port_set_open_drain);
			}
			break;
		case INCA_PORT_IOCPUDSEL:
			if (parm.value == 0x00) {
				PORT_IOC_CALL(ret,parm.port,parm.pin,inca_port_clear_pudsel);
			}
			else {
				PORT_IOC_CALL(ret,parm.port,parm.pin,inca_port_set_pudsel);
			}
			break;
		case INCA_PORT_IOCPUDEN:
			if (parm.value == 0x00) {
				PORT_IOC_CALL(ret,parm.port,parm.pin,inca_port_clear_puden);
			}
			else {
				PORT_IOC_CALL(ret,parm.port,parm.pin,inca_port_set_puden);
			}
			break;
		case INCA_PORT_IOCSTOFF:
			if (parm.value == 0x00) {
				PORT_IOC_CALL(ret,parm.port,parm.pin,inca_port_clear_stoff);
			}
			else {
				PORT_IOC_CALL(ret,parm.port,parm.pin,inca_port_set_stoff);
			}
			break;
		case INCA_PORT_IOCDIR:
			if (parm.value == 0x00) {
				PORT_IOC_CALL(ret,parm.port,parm.pin,inca_port_set_dir_in);
			}
			else {
				PORT_IOC_CALL(ret,parm.port,parm.pin,inca_port_set_dir_out);
			}
			break;
		case INCA_PORT_IOCOUTPUT:
			if (parm.value == 0x00) {
				PORT_IOC_CALL(ret,parm.port,parm.pin,inca_port_clear_output);
			}
			else {
				PORT_IOC_CALL(ret,parm.port,parm.pin,inca_port_set_output);
			}
			break;
		case INCA_PORT_IOCINPUT:
			ret=inca_port_reserve_pin(parm.port,parm.pin,PORT_MODULE_ID); 
			if (ret == 0) 
				parm.value=inca_port_get_input(parm.port,parm.pin,PORT_MODULE_ID); 
			ret=inca_port_free_pin(parm.port,parm.pin,PORT_MODULE_ID); 
			copy_to_user((void*)arg,(void*)&parm,sizeof(struct inca_port_ioctl_parm));
			break;
		//////SNOM360 SERIAL DISPLAY 
	        case INCA_PORT_DISPCMD:
		  disp_command = parm.value;
		  PORT_IOC_CALL(ret, parm.port, CS, inca_port_clear_output);
		  PORT_IOC_CALL(ret, parm.port, RS, inca_port_clear_output);
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_clear_output);
		  if (disp_command & (1<<7)){ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_set_output);
		  }
		  else{ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_clear_output);
		  }
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_set_output);
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_clear_output);
		  if (disp_command & (1<<6)){ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_set_output);
		  }
		  else{ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_clear_output);
		  }
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_set_output);
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_clear_output);
		  if (disp_command & (1<<5)){ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_set_output);
		  }
		  else{ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_clear_output);
		  }
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_set_output);
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_clear_output);
		  if (disp_command & (1<<4)){ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_set_output);
		  }
		  else{ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_clear_output);
		  }
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_set_output);
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_clear_output);
		  if (disp_command & (1<<3)){ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_set_output);
		  }
		  else{ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_clear_output);
		  }
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_set_output);
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_clear_output);
		  if (disp_command & (1<<2)){ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_set_output);
		  }
		  else{ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_clear_output);
		  }
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_set_output);
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_clear_output);
		  if (disp_command & (1<<1)){ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_set_output);
		  }
		  else{ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_clear_output);
		  }
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_set_output);
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_clear_output);
		  if (disp_command & (1<<0)){ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_set_output);
		  }
		  else{ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_clear_output);
		  }
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_set_output);
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_clear_output);

		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CS, inca_port_set_output);
		  break;
	        case INCA_PORT_DISPDATA:
		  disp_data = parm.value;
		  PORT_IOC_CALL(ret, parm.port, CS, inca_port_clear_output);
		  PORT_IOC_CALL(ret, parm.port, RS, inca_port_set_output);
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_clear_output);
		  if (disp_data & (1<<7)){ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_set_output);
		  }
		  else{ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_clear_output);
		  }
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_set_output);
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_clear_output);
		  if (disp_data & (1<<6)){ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_set_output);
		  }
		  else{ 
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_clear_output);
		  }
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_set_output);
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_clear_output);
		  if (disp_data & (1<<5)){
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_set_output);
		  }
		  else {
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_clear_output);
		  }
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_set_output);
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_clear_output);
		  if (disp_data & (1<<4)) {
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_set_output);
		  }
		  else {
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_clear_output);
		  }
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_set_output);
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_clear_output);
		  if (disp_data & (1<<3)) {
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_set_output);
		  }
		  else {
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_clear_output);
		  }
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_set_output);
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_clear_output);
		  if (disp_data & (1<<2)) {
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_set_output);
		  }
		  else {
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_clear_output);
		  }
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_set_output);
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_clear_output);
		  if (disp_data & (1<<1)) {
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_set_output);
		  }
		  else {
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_clear_output);
		  }
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_set_output);
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_clear_output);
		  if (disp_data & (1<<0)) {
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_set_output);
		  }
		  else {
		    PORT_IOC_CALL(ret, parm.port, SDA, inca_port_clear_output);
		  }
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_set_output);
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CLK, inca_port_clear_output);
		  SERIAL_DELAY;
		  PORT_IOC_CALL(ret, parm.port, CS, inca_port_set_output);
		  break;
		  ///////////////////////////
#ifdef DISP_GS_240x128
	        case INCA_PORT_GS_DISPCMD:
		  //ret = write_gs_command(parm.value);
		  write_direct_command(parm.value);
		  break;
	        case INCA_PORT_GS_DISPDATA:
		  //ret = write_gs_data(parm.value);
		  write_direct_data(parm.value);
		  break;
	        case INCA_PORT_GS_MDISPDATA:
		  write_gs_mdata(parm);
		  break;
	        case INCA_PORT_GS_MDISPDATA_2B3P:
		  write_gs_mdata_2b3p(parm);
		  break;
	        case INCA_PORT_GS_DISPBYTES:
		  write_gs_dispbytes(parm);
		  break;
#endif
		  ///////////////////////////
	        default:
			ret = -EINVAL;
	}

	return ret;
}

static struct file_operations port_fops = {
	open    : inca_port_open,
	release : inca_port_release,
	ioctl   : inca_port_ioctl
};

/**
 *	inca_port_init - Initialize port structures
 *
 *	This function initializes the internal data structures of the driver
 *	and will create the proc file entry and device.
 *
 *  	Return Value:	
 *	@OK = OK
 */
int __init inca_port_init(void)
{
	int t = 0;
	int i = 0;
	int err = 0;
	u32 pins = 0;

	printk("Inca-IP Port Initialization\n");

	sema_init(&port_sem,1);
	
	for (t = 0; t<MAX_PORTS; t++) {
		for (i = 0;i<PINS_PER_PORT;i++) inca_port_pin_usage[t][i] = -1;
	}

	/* Check ports for available pins */
	for (t = 0; t<MAX_PORTS; t++) {
		pins = inca_mux_gpio_pins(t);
		for (i = 0;i<PINS_PER_PORT;i++) {
			if ((pins & 0x00000001) > 0) 
				inca_port_pin_usage[t][i] = 0;
			pins = pins >> 1;
		}
	}
	
   	/* register port device */
   	err = register_chrdev(0,"inca-port", &port_fops);
	if (err > 0) port_major = err;
	else {
		printk("inca-port: Error! Could not register port device. #%d\n",err);
		return err;
	}
	
	/* Create proc file */ 
	create_proc_read_entry("driver/inca_port", 0, NULL, inca_port_read_procmem, NULL);

	return OK;
}

module_init(inca_port_init);

EXPORT_SYMBOL(inca_port_reserve_pin);
EXPORT_SYMBOL(inca_port_free_pin);
EXPORT_SYMBOL(inca_port_set_open_drain);
EXPORT_SYMBOL(inca_port_clear_open_drain);
EXPORT_SYMBOL(inca_port_set_pudsel);
EXPORT_SYMBOL(inca_port_clear_pudsel);
EXPORT_SYMBOL(inca_port_set_puden);
EXPORT_SYMBOL(inca_port_clear_puden);
EXPORT_SYMBOL(inca_port_set_stoff);
EXPORT_SYMBOL(inca_port_clear_stoff);
EXPORT_SYMBOL(inca_port_set_dir_out);
EXPORT_SYMBOL(inca_port_set_dir_in);
EXPORT_SYMBOL(inca_port_set_output);
EXPORT_SYMBOL(inca_port_clear_output);
EXPORT_SYMBOL(inca_port_get_input);

