/* pseudo_lan.c: based on isa_skeleton.c
 *
 *	Written 1993-94 by Donald Becker.
 *
 *	Copyright 1993 United States Government as represented by the
 *	Director, National Security Agency.
 *
 *	This software may be used and distributed according to the terms
 *	of the GNU General Public License, incorporated herein by reference.
 *
 *	The author may be reached as becker@scyld.com, or C/O
 *	Scyld Computing Corporation
 *	410 Severn Ave., Suite 210
 *	Annapolis MD 21403
 *
 *	This file is an outline for writing a network device driver for the
 *	the Linux operating system.
 *
 *	To write (or understand) a driver, have a look at the "loopback.c" file to
 *	get a feel of what is going on, and then use the code below as a skeleton
 *	for the new driver.
 *
 */

/*  static const char *version = */
/*  	"pseudo_lan v0.1\n"; */

/*
 *  Sources:
 *	List your sources of programming information to document that
 *	the driver is your own creation, and give due credit to others
 *	that contributed to the work. Remember that GNU project code
 *	cannot use proprietary or trade secret information. Interface
 *	definitions are generally considered non-copyrightable to the
 *	extent that the same names and structures must be used to be
 *	compatible.
 *
 *	Finally, keep in mind that the Linux kernel is has an API, not
 *	ABI. Proprietary object-code-only distributions are not permitted
 *	under the GPL.
 */

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <linux/spinlock.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <linux/errno.h>
#include <linux/init.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/mii.h>

#include "inca_mdio.h"

/*
 * The name of the card. Is used for messages and in the requests for
 * io regions, irqs and dma channels
 */
static const char* cardname[] = { "incaip_lan", "incaip_pc" };

/* First, a few definitions that the brave might change. */

/* use 0 for production, 1 for verification, >2 for debug */
#ifndef NET_DEBUG
#define NET_DEBUG 2
#endif
/*  static unsigned int net_debug = NET_DEBUG; */

#define MY_TX_TIMEOUT  ((400*HZ)/1000)

/* Information that need to be kept for each board. */
struct net_local {
    struct net_device_stats stats;
    long open_time;			/* Useless example local info. */

    /* Tx control lock.  This protects the transmit buffer ring
     * state along with the "tx full" state of the driver.  This
     * means all netif_queue flow control actions are protected
     * by this lock as well.
     */
    spinlock_t lock;
};

/* Index to functions, as function prototypes. */

extern int netcard_probe(struct net_device *dev);

static int	net_open(struct net_device *dev);
static int	net_send_packet(struct sk_buff *skb, struct net_device *dev);
static int	net_close(struct net_device *dev);
static struct	net_device_stats *net_get_stats(struct net_device *dev);
static void     net_tx_timeout(struct net_device *dev);
static int      net_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd);


/*
 * Check for a network adaptor of this type, and return '0' iff one exists.
 * If dev->base_addr == 0, probe all likely locations.
 * If dev->base_addr == 1, always return failure.
 * If dev->base_addr == 2, allocate space for the device and return success
 * (detachable devices only).
 */
int __init 
netcard_probe(struct net_device *dev)
{
    struct net_local *np = (struct net_local *)dev->priv;

    SET_MODULE_OWNER(dev);

    /* Initialize the device structure. */
    if (dev->priv == NULL) {
        dev->priv = kmalloc(sizeof(struct net_local), GFP_KERNEL);
        if (dev->priv == NULL)
            return -ENOMEM;
    }

    memset(dev->priv, 0, sizeof(struct net_local));

    np = (struct net_local *)dev->priv;
    spin_lock_init(&np->lock);

    dev->open		= net_open;
    dev->stop		= net_close;
    dev->hard_start_xmit	= net_send_packet;
    dev->get_stats		= net_get_stats;
    dev->do_ioctl       = net_ioctl;

    dev->tx_timeout		= &net_tx_timeout;
    dev->watchdog_timeo	= MY_TX_TIMEOUT; 

    /* Fill in the fields of the device structure with ethernet values. */
    ether_setup(dev);

    return 0;
}

static void net_tx_timeout(struct net_device *dev)
{
	struct net_local *np = (struct net_local *)dev->priv;

	printk(KERN_WARNING "%s: transmit timed out\n", dev->name);

	np->stats.tx_errors++;

	/* If we have space available to accept new transmit
	 * requests, wake up the queueing layer.  This would
	 * be the case if the chipset_init() call above just
	 * flushes out the tx queue and empties it.
	 *
	 * If instead, the tx queue is retained then the
	 * netif_wake_queue() call should be placed in the
	 * TX completion interrupt handler of the driver instead
	 * of here.
	 */
        /* we have never space available, since we represent only PHYs */
/*  	if (!tx_full(dev)) */
/*  		netif_wake_queue(dev); */
}

static int      net_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
    struct mii_ioctl_data *mii_data;

    if (ifr == NULL) return -EINVAL;

    mii_data = (struct mii_ioctl_data *)(&ifr->ifr_data);

    switch (cmd)
    {
        /* mii-diag sends this ioctl prior the use of the following two SIOCDEVPRIVATE+1 and SIOCDEVPRIVATE+2 */
        /* mii-diag expects a non error return; */
	case SIOCDEVPRIVATE:	/* binary compat, remove in 2.5 */
	case SIOCGMIIPHY:
            mii_data->phy_id = dev->base_addr; /* phy-address */
            break;

            /* read mii register; this value conforms with the usage in mii-diag.c */
	case SIOCDEVPRIVATE + 1:/* binary compat, remove in 2.5 */
	case SIOCGMIIREG:
            mii_data->val_out = mdio_read(mii_data->phy_id, mii_data->reg_num);
            break;

            /* write mii register; this value conforms with the usage in mii-diag.c */
	case SIOCDEVPRIVATE + 2:/* binary compat, remove in 2.5 */
	case SIOCSMIIREG:
            if (!capable(CAP_NET_ADMIN)) // only administrators are allowed to write mdio registers
                return -EPERM;

            mdio_write(mii_data->phy_id, mii_data->reg_num, mii_data->val_in);
            break;

        default:
            return -ENOTTY;
    }

    return 0;
}

/*
 * Open/initialize the board. This is called (in the current kernel)
 * sometime after booting when the 'ifconfig' program is run.
 *
 * This routine should set everything up anew at each open, even
 * registers that "should" only need to be set once at boot, so that
 * there is non-reboot way to recover if something goes wrong.
 */
static int
net_open(struct net_device *dev)
{
	struct net_local *np = (struct net_local *)dev->priv;

	np->open_time = jiffies;

	/* We are now ready to accept transmit requeusts from
	 * the queueing layer of the networking.
	 */
        /* we never accept any requests */
/*  	netif_start_queue(dev); */

	return 0;
}

/* This will only be invoked if your driver is _not_ in XOFF state.
 * What this means is that you need not check it, and that this
 * invariant will hold if you make sure that the netif_*_queue()
 * calls are done at the proper times.
 */
static int net_send_packet(struct sk_buff *skb, struct net_device *dev)
{
/*  	struct net_local *np = (struct net_local *)dev->priv; */

        printk(__FILE__ __FUNCTION__ " shouldn't get here !\n");
        netif_stop_queue(dev);  /* try to stop upper layer to call us */
        
	return 1;               /* packet not sent */
}

/* The inverse routine to net_open(). */
static int
net_close(struct net_device *dev)
{
	struct net_local *lp = (struct net_local *)dev->priv;

	lp->open_time = 0;

	netif_stop_queue(dev);

	return 0;

}

/*
 * Get the current statistics.
 * This may be called with the card open or closed.
 */
static struct net_device_stats *net_get_stats(struct net_device *dev)
{
	struct net_local *lp = (struct net_local *)dev->priv;

	return &lp->stats;
}


static struct net_device pseudolan_device[2];
static unsigned long pseudolan_phy_addr[2] = { 6, 5 }; /* PHY-Addr LAN/PC */

MODULE_LICENSE("GPL");

int __init pseudo_lan_init(void)
{
	int result, i;

        for (i = 0; i < sizeof(pseudolan_device)/sizeof(pseudolan_device[0]); i++)
        {
            /* Copy the parameters from insmod into the device structure. */
            strcpy(pseudolan_device[i].name, cardname[i]);
            pseudolan_device[i].base_addr = pseudolan_phy_addr[i];
            pseudolan_device[i].irq       = 0;
            pseudolan_device[i].dma       = 0;
            pseudolan_device[i].mem_start = 0;
            pseudolan_device[i].init      = netcard_probe;
            if ((result = register_netdev(&pseudolan_device[i])) != 0)
            {
                printk("Failed to register pseudolan device %d\n", i);
		return result;
            }
        }

	return 0;
}

void __exit pseudo_lan_cleanup(void)
{
    int i;

	/* No need to check MOD_IN_USE, as sys_delete_module() checks. */
    for (i = 0; i < sizeof(pseudolan_device)/sizeof(pseudolan_device[0]); i++)
    {
	unregister_netdev(&pseudolan_device[i]);
        if (pseudolan_device[i].priv)
            kfree(pseudolan_device[i].priv);
    }

}

module_init(pseudo_lan_init);
module_exit(pseudo_lan_cleanup);



/*
 * Local variables:
 *  compile-command:
 *	gcc -D__KERNEL__ -Wall -Wstrict-prototypes -Wwrite-strings
 *	-Wredundant-decls -O2 -m486 -c skeleton.c
 *  version-control: t
 *  kept-new-versions: 5
 *  tab-width: 4
 *  c-indent-level: 4
 * End:
 */
