#include <linux/module.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/slab.h> // former "malloc.h"
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/proc_fs.h>
#include <linux/ioport.h>

#include <asm/incaip/inca-ip.h>
#include <asm/incaip/mux.h>

#include <asm/incaip/pwm.h>
#include "pwm_defs.h"

/*>>>>>>>>>>>>>>>>>>>>>>>>> Module Specific Stuff <<<<<<<<<<<<<<<<<<<<<<<<<<<*/
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Andreas Schmidt");
MODULE_DESCRIPTION("Inca-IP Pulse Width Modulator (PWM) Driver");
/*
 * Parameter Definitions for Module
 */
static int pwm1 = -1;		// Availability of PWM1 (0 or 1) 
MODULE_PARM(pwm1, "i");
MODULE_PARM_DESC(pwm1, "Availability of PWM1 (0 or 1)");

static int pwm2 = -1;		// Availability of PWM2 (0 or 1) 
MODULE_PARM(pwm2, "i");
MODULE_PARM_DESC(pwm2, "Availability of PWM2 (0 or 1)");

static char *pwm1_name = "pwm1";
MODULE_PARM(pwm1_name, "s");
MODULE_PARM_DESC(pwm1_name, "Name of PWM1 Device");

static char *pwm2_name = "pwm2";
MODULE_PARM(pwm2_name, "s");
MODULE_PARM_DESC(pwm2_name, "Name of PWM2 Device");

static int pwm1_value = 0;	// Initial value for PWM1
MODULE_PARM(pwm1_value, "i");
MODULE_PARM_DESC(pwm1_value, "Initial value for PWM1 (0..255)");

static int pwm2_value = 0;	// Initial value for PWM2
MODULE_PARM(pwm2_value, "i");
MODULE_PARM_DESC(pwm2_value, "Initial value for PWM2 (0..255)");

static int major = 0;	// Major ID of PWM device (default: 0=dynamic)
MODULE_PARM(major, "i");
MODULE_PARM_DESC(major, "Major ID of PWM device (default: 0=dynamic)");

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>> Local Variables <<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
static char *dev_name[MAX_PWM]  = { NULL, NULL };
static int  *dev_avail[MAX_PWM] = { &pwm1, &pwm2 };
static int   dev_id[MAX_PWM]    = {0, 1};

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> PROC Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
static struct proc_dir_entry *inca_pwm_proc[MAX_PWM];

/****************************************************************************
Description :
   Converts an ascii string of decimal digits to a positive integer. 
Arguments   : 
   *s - Pointer to the ascii string.
Return      :
   >= 0     - converted integer value
   -EINVAL  - unexpected character in ascii string
Remarks     :
****************************************************************************/
static int ascii2int(char* s)
{
	int	temp_val = 0;

	while (*s >= '0' && *s <= '9') {
		temp_val = temp_val * 10 + (*s - '0');
		s++;
	}
	
	if (*s != 0xA) { // string should end with line feed (0xA)
		printk(KERN_ERR "unexpected character in ascii string!\n");
		return -EINVAL;
	}
	return temp_val;
}

static int inca_pwm_read_procmem(char *buf, char **start, off_t offset, 
					int count, int *eof, void *data)
{
	long	len = 0;
	int 	id  = *((int *) data);
	
	len += sprintf(buf+len,"PWM_device=/dev/%s\n", dev_name[id]);
	len += sprintf(buf+len,"PWM_value=%d\n", inca_pwm_get(id));

	*eof = 1;	// No more data available

	return len;
}

static int inca_pwm_write_procmem(struct file *file, const char *buffer,
					unsigned long count, void *data)
{
	char*	ibuf = (char*) buffer;
	int 	val = 0;
	int 	id  = *((int *) data);

	val = ascii2int(ibuf);
	
	if (val >= 0 && val < 256) {
		inca_pwm_set(id, val);
	}
	else {
		printk(KERN_ERR "enter a value in the range 0..255!\n"); 
		return -EINVAL;
	}
	
	return count;
	
}

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> Device Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
struct file_operations pwm_fops = {
	read  : inca_pwm_read,
	write : inca_pwm_write,
	ioctl : inca_pwm_ioctl
};

static ssize_t inca_pwm_read(struct file *filp, char *buf,
					 size_t count, loff_t *f_pos)
{
	struct inode *inode = INODE_FROM_F(filp);
	int id = MINOR(inode->i_rdev);
	char pwm_val;
	
	if (id < MAX_PWM) {
		pwm_val = inca_pwm_get(id);
	}
	else {
		printk(KERN_ERR "PWM ID out of range!\n");
		return -EINVAL;
	}
	
	count = 1;
	if (copy_to_user(buf, &pwm_val, count)) {
		return -EFAULT;
	}
	return count;
}

static ssize_t inca_pwm_write(struct file *filp, const char *buf,
					 size_t count, loff_t *f_pos)
{
	struct inode *inode = INODE_FROM_F(filp);
	int id = MINOR(inode->i_rdev);
	char pwm_val;
	
	if (count != 1) {
		return -EINVAL;
	}
	
	if (copy_from_user(&pwm_val, buf, count)) {
		return -EFAULT;
	}
	
	if (id < MAX_PWM) {
		inca_pwm_set(id, pwm_val);
	}
	else {
		printk(KERN_ERR "PWM ID out of range!\n");
		return -EINVAL;
	}
	return count;
}

static int inca_pwm_ioctl(struct inode *inode, struct file *filp, 
				unsigned int cmd, unsigned long arg) 
{
	int err;
	
	switch (cmd) {
		default:
			err = -EINVAL;
	}
	return err;
}


/*>>>>>>>>>>>>>>>>>>>>>>>>>>>> Public Functions <<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
/****************************************************************************
Description :
   This function sets the Pulse Width value of a PWM device. 
Arguments   : 
   id    - ID of the addressed PWM device
   value - PWM value in the range 0..255
Return      :
   0        - OK
   -EINVAL  - error
Remarks     :
****************************************************************************/
int inca_pwm_set(int id, u8 value) {
	u32 temp;
	
	if (id < MAX_PWM) {
		if (*dev_avail[id]) {
			temp = *INCA_IP_TSF_PWM12 & ~(0xFFFF << 16*id);
			*INCA_IP_TSF_PWM12 = temp | (value << 16*id);
		}
		else {
			printk(KERN_ERR "Device %s not available!\n", 
								dev_name[id]);
			return -EINVAL;
		}
	}
	else {
		printk(KERN_ERR "PWM ID out of range!\n");
		return -EINVAL;
	}
	return 0;
}

/****************************************************************************
Description :
   This function gets the Pulse Width value of a PWM device. 
Arguments   : 
   id    - ID of the addressed PWM device
Return      :
   >= 0     - current PWM value
   -EINVAL  - error
Remarks     :
****************************************************************************/
int inca_pwm_get(int id) {
	if (id < MAX_PWM) {
		if (*dev_avail[id]) {
			return (int) ((*INCA_IP_TSF_PWM12 >> 16*id) & 0xFFFF);
		}
		else {
			printk(KERN_ERR "Device %s not available!\n", 
								dev_name[id]);
			return -EINVAL;
		}
	}
	else {
		printk(KERN_ERR "PWM ID out of range!\n");
		return -EINVAL;
		
	}
}

/*>>>>>>>>>>>>>>>>>>>>>>>>>>> Module Init/Exit <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/****************************************************************************
Description :
   This function initializes the Pulse Width Modulator (PWM) hardware of the
   INCA-IP device. This function is called when the module gets inserted
   (via insmod).  
Arguments   :
Return      :
Remarks     :
****************************************************************************/
static int __init inca_pwm_init(void)
{
	int i, err = 0, res;
	unsigned long flags;
	char tmp_string[27];
	
	dev_name[0] = pwm1_name;
	dev_name[1] = pwm2_name;
	
	/* Check module parameters */
	if (pwm1 == -1) pwm1 = inca_mux_pwm1_pins(); // this is *dev_avail[0]
	if (pwm2 == -1) pwm2 = inca_mux_pwm2_pins(); // this is *dev_avail[1]

	printk(KERN_NOTICE "Device availability: %s=%d and %s=%d\n",
		 dev_name[0], *dev_avail[0], dev_name[1], *dev_avail[1]);

	if (pwm1 != 1 && pwm2 != 1) {	// no PWM devices requested by user! 
		printk(KERN_ERR "No PWM devices requested by user!\n");
		return -EINVAL;
	}
	if ((unsigned) pwm1_value > 255 || (unsigned) pwm2_value > 255) {
		printk(KERN_ERR "PWM value(s) out of range 0..255!\n");
		return -EINVAL;
	}

	/* allocate Linux resources */
	if (request_region((u32) INCA_IP_TSF_PWM12, 0x04, 
				"INCA-IP Pulse Width Modulator") == NULL) {
		printk(KERN_ERR "request_region() failed!\n");
		err = -EBUSY;
		goto FAIL_REQUEST_REGION;
	}
	
	/* enable the clock for Terminal Specific Functions in the 
						Power Management Unit */
#ifndef DONE_ELSEWHERE	
 	save_flags(flags); cli();
	*INCA_IP_PMU_PM_GEN |= INCA_IP_PMU_PM_GEN_EN9;   
 	restore_flags(flags);
#endif
	/* Initialize the PWMs */
	inca_pwm_set(0, pwm1_value);
	inca_pwm_set(1, pwm2_value);
	
	/* enable the PWM on the INCA-IP */
 	save_flags(flags); cli();
	*INCA_IP_TSF_TSF_CONF |= INCA_IP_TSF_TSF_CONF_PWMEN;
 	restore_flags(flags);
	
   	/* register PWM devices */
	res = register_chrdev(major, "pwm", &pwm_fops);
	if (res < 0) {
		printk(KERN_ERR "Could not register PWM device!\n");
		goto FAIL_REGISTER_DEVICE;
	}
	if (major == 0) major = res;

	printk(KERN_NOTICE "PWM device registered. major: %d.\n", major);
	
	SET_MODULE_OWNER(&pwm_fops);
		
	/* create PROC entries */
	for (i = 0; i < MAX_PWM; i++) {
		if (*dev_avail[i]) {
			sprintf(tmp_string, "driver/%s", dev_name[i]);
			inca_pwm_proc[i] = create_proc_entry(tmp_string, 0, 0);
			if (!inca_pwm_proc[i]) {
				printk(KERN_ERR "could not create "
					"/proc/driver/%s!\n", dev_name[i]);
				err = -ENOENT;
				goto FAIL_CREATE_PROC;
			}
			inca_pwm_proc[i]->read_proc  = inca_pwm_read_procmem;
			inca_pwm_proc[i]->write_proc = inca_pwm_write_procmem;
			inca_pwm_proc[i]->data = &dev_id[i];
			printk(KERN_NOTICE "/proc/driver/%s created!\n", 
								dev_name[i]);
		}
	}
	
	/* success */
	printk(KERN_NOTICE "INCA-IP Pulse Width Modulator initialized\n");
	
	return 0; 


	/* if initialization fails, release all resources aquired so far! */
FAIL_CREATE_PROC:
	for (; i > 0; i--) 
		if (*dev_avail[i-1]) {
			sprintf(tmp_string, "driver/%s", dev_name[i]);
			remove_proc_entry(tmp_string, 0);
		}
FAIL_REGISTER_DEVICE:
	unregister_chrdev(major, "pwm");
			
 	save_flags(flags); cli();
	*INCA_IP_TSF_TSF_CONF    &= ~INCA_IP_TSF_TSF_CONF_PWMEN;
 	restore_flags(flags);
	
	for (i = 0; i < MAX_PWM; i++) inca_pwm_set(i, 0);

	release_region((u32) INCA_IP_TSF_PWM12, 0x04);
	
FAIL_REQUEST_REGION:
	return err;
}

/****************************************************************************
Description :
   This function releases the Pulse Width Modulator (PWM) hardware of the
   INCA-IP device. This function is called when the module gets removed
   (via rmmod).  
Arguments   :
Return      :
Remarks     :
****************************************************************************/
static void __exit inca_pwm_exit(void)
{
	int i;
	unsigned long flags;
	char tmp_string[27];
	
	/* reset PWMs */
	for (i = 0; i < MAX_PWM; i++) inca_pwm_set(i, 0);

	/* disable the PWM module */
 	save_flags(flags); cli();
	*INCA_IP_TSF_TSF_CONF    &= ~INCA_IP_TSF_TSF_CONF_PWMEN;
 	restore_flags(flags);

	/* Release aquired resources (in inverse order) */
	for (i = 0; i < MAX_PWM; i++) 
		if (*dev_avail[i]) {
			sprintf(tmp_string, "driver/%s", dev_name[i]);
			remove_proc_entry(tmp_string, 0);
			printk(KERN_NOTICE "/proc/driver/%s removed!\n",
								dev_name[i]);
		}
		
	unregister_chrdev(major, "pwm");
	
	release_region((u32) INCA_IP_TSF_PWM12, 0x04);
	
}

#ifdef MODULE
module_init(inca_pwm_init);
module_exit(inca_pwm_exit);
#endif
