/*
  Cipher Feedback Block mode
 
  Written by David Bryson <mutex@mac.com>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software

   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <linux/string.h>
#include <linux/sched.h>

#if defined(CIPHER_BITS_128)
# define BS 16
#else
# define BS 8
#endif


/*  
 * These functions only use the XOR operator on the data, so no
 * endianness problems should occur.
 */

static int CIPHER_NAME(_cfb_encrypt)(struct cipher_context *cx, 
				const u8 *in_blk, u8 *out_blk, int size, 
				int atomic, const u8 *_iv)
{
	const u32 *iv = (const u32 *) _iv;
	int SS = cx->ci->segmentsize/8;
#ifdef HW_ENCRYPT
	u8 tmp[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
# if defined(CIPHER_BITS_64) /* DES / 3DES Cipher */
	*INCA_IP_DES_3DES_DES_CONTROLR = (*INCA_IP_DES_3DES_DES_CONTROLR &
			~INCA_IP_DES_3DES_DES_CONTROLR_E_D) | /* encrypt */
			INCA_IP_DES_3DES_DES_CONTROLR_O(3) | /* CFB Mode */
			INCA_IP_DES_3DES_DES_CONTROLR_F(
				SS ==  8 ? 1 :
				SS ==  4 ? 2 :
				SS ==  2 ? 3 :
				SS ==  1 ? 4 : 1); /* feedback size */

	*INCA_IP_DES_3DES_DES_IVHR = iv[0];
	*INCA_IP_DES_3DES_DES_IVLR = iv[1];
#  ifdef CRYPTO_DEBUG
	printk("Initial value IVH=%08X IVL=%08X\n", iv[0], iv[1]);
#  endif	
# elif defined(CIPHER_BITS_128) /* AES Cipher */
	*INCA_IP_AES_AES_CONTROLR = (*INCA_IP_AES_AES_CONTROLR &
			~INCA_IP_AES_AES_CONTROLR_E_D) | /* encrypt */
			INCA_IP_AES_AES_CONTROLR_O(3) | /* CFB Mode */
			INCA_IP_AES_AES_CONTROLR_F(
				SS == 16 ? 0 :
				SS ==  8 ? 1 :
				SS ==  4 ? 2 :
				SS ==  2 ? 3 :
				SS ==  1 ? 4 : 0); /* feedback size */

	*INCA_IP_AES_AES_IV3R = iv[0];
	*INCA_IP_AES_AES_IV2R = iv[1];
	*INCA_IP_AES_AES_IV1R = iv[2];
	*INCA_IP_AES_AES_IV0R = iv[3];
# else
#  error unsupported cipher blocksize
# endif	
	for (size -= SS; size >= 0; size -= SS) {
//		printk(KERN_NOTICE "@(in_blk)=%08X\n", (u32) in_blk);
		CIPHER_NAME(_encrypt)(cx, in_blk, out_blk, SS, atomic);
		in_blk += SS; out_blk += SS;

# if LINUX_VERSION_CODE <= KERNEL_VERSION(2,5,0)
		if (!atomic && current->need_resched)
			schedule ();
# else
		if (!atomic)
			cond_resched ();
# endif
	} /* for */
	if (size != -SS) {
		memcpy(tmp, in_blk, size+SS);
		CIPHER_NAME(_encrypt)(cx, tmp, out_blk, SS, atomic);
	} /* if */

#else /* Software implementaion */
	u8 tmp[BS];

	if (BS != SS) {
		printk(KERN_ERR "Segment size (SS) < block size (BS)"
				" not supported in software CFB mode!");
		return -EINVAL;
	}

	((u32 *)tmp)[0] = iv[0];
	((u32 *)tmp)[1] = iv[1];
# ifdef CIPHER_BITS_128
	((u32 *)tmp)[2] = iv[2];
	((u32 *)tmp)[3] = iv[3];
# endif
	for (size -= BS; size >= 0; size -= BS) {
		CIPHER_NAME(_encrypt)(cx, tmp, tmp, BS, atomic);
		((u32 *)tmp)[0] ^= ((u32 *)in_blk)[0];
		((u32 *)tmp)[1] ^= ((u32 *)in_blk)[1];
# ifdef CIPHER_BITS_128
		((u32 *)tmp)[2] ^= ((u32 *)in_blk)[2];
		((u32 *)tmp)[3] ^= ((u32 *)in_blk)[3];
# endif
		memcpy(out_blk, tmp, sizeof(tmp));
		in_blk += BS; out_blk += BS;

# if LINUX_VERSION_CODE <= KERNEL_VERSION(2,5,0)
		if (!atomic && current->need_resched)
			schedule();
# else
		if (!atomic)
			cond_resched();
# endif	       
	} /* end for */

	if (size != BS) {
		memset(tmp, 0, sizeof(tmp));
		memcpy(tmp, in_blk, size+BS);
		CIPHER_NAME(_encrypt)(cx, tmp, tmp, BS, atomic);

		((u32 *)tmp)[0] ^= ((u32 *)in_blk)[0 - BS/4];
		((u32 *)tmp)[1] ^= ((u32 *)in_blk)[1 - BS/4];
# ifdef CIPHER_BITS_128
		((u32 *)tmp)[2] ^= ((u32 *)in_blk)[2 - BS/4];
		((u32 *)tmp)[3] ^= ((u32 *)in_blk)[3 - BS/4];
# endif
	} /* end if */
#endif

	return 0;
}

static int CIPHER_NAME(_cfb_decrypt)(struct cipher_context *cx, 
				const u8 *in_blk, u8 *out_blk, int size, 
				int atomic, const u8 *_iv)
{
	const u32 *iv = (const u32 *) _iv;
	int SS = cx->ci->segmentsize/8;
#ifdef HW_DECRYPT
	u8 tmp[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
# if defined(CIPHER_BITS_64) /* DES / 3DES Cipher */
	*INCA_IP_DES_3DES_DES_CONTROLR |=
			INCA_IP_DES_3DES_DES_CONTROLR_E_D | /* decrypt */
			INCA_IP_DES_3DES_DES_CONTROLR_O(3) | /* CFB Mode */
			INCA_IP_AES_AES_CONTROLR_F(
				SS ==  8 ? 1 :
				SS ==  4 ? 2 :
				SS ==  2 ? 3 :
				SS ==  1 ? 4 : 1); /* feedback size */


      *INCA_IP_DES_3DES_DES_IVHR = iv[0];
      *INCA_IP_DES_3DES_DES_IVLR = iv[1];
#  ifdef CRYPTO_DEBUG
      printk("Initial value IVH=%08X IVL=%08X\n", iv[0], iv[1]);
#  endif	
# elif defined(CIPHER_BITS_128) /* AES Cipher */
	*INCA_IP_AES_AES_CONTROLR |=
			INCA_IP_AES_AES_CONTROLR_E_D | /* decrypt */
			INCA_IP_AES_AES_CONTROLR_O(3) | /* CFB Mode */
			INCA_IP_AES_AES_CONTROLR_F(
				SS == 16 ? 0 :
				SS ==  8 ? 1 :
				SS ==  4 ? 2 :
				SS ==  2 ? 3 :
				SS ==  1 ? 4 : 0); /* feedback size */
			
	/* Processing of decryption key already started in aes_set_key() ! */
	while ((*INCA_IP_AES_AES_CONTROLR & INCA_IP_AES_AES_CONTROLR_KV) == 0)
#  ifdef CRYPTO_DEBUG
		printk(KERN_NOTICE "processing decryption key!\n")
#  endif	
		;
			
	*INCA_IP_AES_AES_IV3R = iv[0];
	*INCA_IP_AES_AES_IV2R = iv[1];
	*INCA_IP_AES_AES_IV1R = iv[2];
	*INCA_IP_AES_AES_IV0R = iv[3];
# else
#  error unsupported cipher blocksize
# endif	
	
	for (size -= SS; size >= 0; size -= SS) {
//		printk(KERN_NOTICE "@(in_blk)=%08X\n", (u32) in_blk);
		CIPHER_NAME(_decrypt)(cx, in_blk, out_blk, SS, atomic);
//		CIPHER_NAME(_encrypt)(cx, in_blk, out_blk, SS, atomic);
		in_blk += SS; out_blk += SS;

# if LINUX_VERSION_CODE <= KERNEL_VERSION(2,5,0)
		if (!atomic && current->need_resched)
			schedule ();
# else
		if (!atomic)
			cond_resched ();
# endif
	}

	if (size != -SS) { 
		CIPHER_NAME(_decrypt)(cx, in_blk, tmp, SS, atomic);
//		CIPHER_NAME(_encrypt)(cx, in_blk, tmp, SS, atomic);
		memcpy(out_blk, tmp, size+SS);
	}
#else /* Software implementation */
	u8 tmp[BS];

	if (BS != SS) {
		printk(KERN_ERR "Segment size (SS) < block size (BS)"
				" not supported in software CFB mode!");
		return -EINVAL;
	}

	((u32 *)tmp)[0] = iv[0];
	((u32 *)tmp)[1] = iv[1];
# ifdef CIPHER_BITS_128
	((u32 *)tmp)[2] = iv[2];
	((u32 *)tmp)[3] = iv[3];
# endif
	for (size -= BS; size >= 0; size -= BS) {
//		CIPHER_NAME(_decrypt)(cx, tmp, tmp, BS, atomic);
		CIPHER_NAME(_encrypt)(cx, tmp, tmp, BS, atomic);
		((u32 *)tmp)[0] ^= ((u32 *)in_blk)[0];
		((u32 *)tmp)[1] ^= ((u32 *)in_blk)[1];
# ifdef CIPHER_BITS_128
		((u32 *)tmp)[2] ^= ((u32 *)in_blk)[2];
		((u32 *)tmp)[3] ^= ((u32 *)in_blk)[3];
# endif
		memcpy(out_blk, tmp, sizeof(tmp));
		memcpy(tmp, in_blk, sizeof(tmp));
		in_blk += BS; out_blk += BS;

# if LINUX_VERSION_CODE <= KERNEL_VERSION(2,5,0)
		if (!atomic && current->need_resched)
			schedule();
# else
		if (!atomic)
			cond_resched();
# endif	       
	} /* end for */

	if (size != -BS) { 
//??		CIPHER_NAME(_encrypt)(cx, tmp, tmp, BS, atomic);
		CIPHER_NAME(_decrypt)(cx, tmp, tmp, BS, atomic);
		((u32 *)tmp)[0] ^= ((u32 *)in_blk)[0];
		((u32 *)tmp)[1] ^= ((u32 *)in_blk)[1];
# ifdef CIPHER_BITS_128
		((u32 *)tmp)[2] ^= ((u32 *)in_blk)[2];
		((u32 *)tmp)[3] ^= ((u32 *)in_blk)[3];
# endif
		memcpy(out_blk, tmp, size+BS);
	} /* end if */
#endif

	return 0;
}

#undef BS
