/*
 * Driver for the INCAIP Switch
 * Copyright (C) 2003 Gary Jennejohn >gj@denx.de>
 *
 * Based on:
 *
 * isa-skeleton.c: A network driver outline for linux.
 *
 *	Written 1993-94 by Donald Becker.
 *
 *	Copyright 1993 United States Government as represented by the
 *	Director, National Security Agency.
 *
 *	This software may be used and distributed according to the terms
 *	of the GNU General Public License, incorporated herein by reference.
 *
 *	The author may be reached as becker@scyld.com, or C/O
 *	Scyld Computing Corporation
 *	410 Severn Ave., Suite 210
 *	Annapolis MD 21403
 *
 */

static const char *version = "incaipsw Ver. 0.1\n";

/*
 *  Sources:
 *	List your sources of programming information to document that
 *	the driver is your own creation, and give due credit to others
 *	that contributed to the work. Remember that GNU project code
 *	cannot use proprietary or trade secret information. Interface
 *	definitions are generally considered non-copyrightable to the
 *	extent that the same names and structures must be used to be
 *	compatible.
 *
 *	Finally, keep in mind that the Linux kernel is has an API, not
 *	ABI. Proprietary object-code-only distributions are not permitted
 *	under the GPL.
 */

#include <linux/config.h>

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <linux/spinlock.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <linux/errno.h>
#include <linux/init.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <asm/incaip/inca-ip.h>
#include <asm/incaip/irq.h>
#include <asm/incaip/dma.h>
/* TEST */
#include <net/tcp.h>

/* Well, I really don't like it, but for now it's the fastest solution */
#include <../arch/mips/infineon/incaip/switch/vxWorks_type.h>
#include <../arch/mips/infineon/incaip/switch/inca_switch_access.h>
#include <../arch/mips/infineon/incaip/switch/inca_switch_settings.h>

/*
 * The name of the card. Is used for messages and in the requests for
 * io regions, irqs and dma channels
 */
static const char* cardname = "incaipsw";

/* First, a few definitions that the brave might change. */

/* use 0 for production, 1 for verification, >2 for debug */
#ifndef NET_DEBUG
#define NET_DEBUG 2
#endif
static unsigned int incaipsw_debug = NET_DEBUG;

/* #define INCAIPSW_DEBUG  */

/*  #define INCASW_RX_DEBUG */

#define MY_TX_TIMEOUT  ((400*HZ)/1000)

/* Information that need to be kept for each board. */
struct incaipsw_local {
	struct net_device_stats stats;
	long open_time;			/* Useless example local info. */

	/* Tx control lock.  This protects the transmit buffer ring
	 * state along with the "tx full" state of the driver.  This
	 * means all netif_queue flow control actions are protected
	 * by this lock as well.
	 */
	spinlock_t lock;
};


#define NUM_RX_DESC 90
#define NUM_TX_DESC	15
#define PKTSIZE_ALIGN	1520
#define INCAIP_SW_DMACHAN 0

#define SWITCH_UNICAST	1
#define SWITCH_MULTICAST	2
#define SWITCH_BROADCAST	4

#define MULTIPLE_TX_ATTEMPTS   
#if defined MULTIPLE_TX_ATTEMPTS
#define TX_MAX_ATTEMPTS 20  /* number of transmission attempts, 1 ... n */
#endif

#if 0
#define SW_WRITE_REG(reg, value)   \
         *((volatile u32*)reg) = (u32)value;\
         asm("nop;nop;nop;nop;nop");\
         asm("nop;nop;nop;nop;nop");\
         asm("nop;nop;nop;nop;nop");\
         *((volatile u32*)reg) = (u32)value;

#define SW_READ_REG(reg, value)    \
         value = (u32)*((volatile u32*)reg);\
         asm("nop;nop;nop;nop;nop");\
         asm("nop;nop;nop;nop;nop");\
         asm("nop;nop;nop;nop;nop");\
         value = (u32)*((volatile u32*)reg);

#define RX_CPU_NO_BUFFERS    0x00008000
#define RX_CPU_NO_PB_NUMBERS 0x00004000
#define RX_LAN_NO_BUFFERS    0x00000800
#define RX_LAN_NO_PB_NUMBERS 0x00000400
#define RX_PC_NO_BUFFERS     0x00000080
#define RX_PC_NO_PB_NUMBERS  0x00000040
#define MAC_TABLE_CHANGE     0x00000010
#define GBL_RX_WM3           0x00000001
#endif

/* Index to functions, as function prototypes. */

extern int	incaipsw_probe(struct net_device *dev);

static int	incaipsw_open(struct net_device *dev);
static int	incaipsw_send_packet(struct sk_buff *skb, struct net_device *dev);
static void	incaipsw_interrupt(int irq, void *dev_id, struct pt_regs *regs);
static int	incaipsw_close(struct net_device *dev);
static struct	net_device_stats *incaipsw_get_stats(struct net_device *dev);
static void	incaipsw_set_multicast_list(struct net_device *dev);
static int 	incaipsw_set_mac_address(struct net_device *dev, void *addr);
static void     incaipsw_tx_timeout(struct net_device *dev);
static unsigned long incaip_sw_allocbuf(void *, unsigned long *);
static unsigned long incaip_sw_rcv(void *, void *, unsigned long, unsigned long);
static unsigned long incaip_sw_txc(void *, unsigned long);


/* Example routines you must write ;->. */
#define tx_done(dev) 1
static int	hardware_send_packet(struct net_device *dev, char *buf, 
                                 int length, struct sk_buff *skb);
static void 	chipset_init(struct net_device *dev, int startp);

static void incaipsw_dma_init(void)
{
	incaDMA_Init();
}

static int tx_full(struct net_device *dev)
{
	return(0);
}

static int hardware_send_packet(struct net_device *dev, char *buf, 
                                int length, struct sk_buff *skb)
{
#ifdef MULTIPLE_TX_ATTEMPTS	
	unsigned int retry_count = 0;  /* tx attempts counter */
#endif
#if 0
register unsigned long regValue;
#endif		
#ifndef INCA_DMA_USE_COPYING
    struct sk_buff *skb_aligned;

    if ((unsigned long)skb->data & 3) /* unaligned data buffer ? */
    {
/*          printk("unaligned skb->data (0x%08X->0x%08X)\n", */
/*                 skb, skb->data); */
        skb_aligned = dev_alloc_skb(skb->len + 4);
        if (!skb_aligned)
        {
            printk(KERN_NOTICE"%s: memory squeeze, dropping tx packet\n",
                   dev->name);
            dev_kfree_skb(skb);
            return -1;
        }

/*          printk("skb_aligned->data(1) (0x%08X)= 0x%08X\n", skb_aligned, skb_aligned->data); */
        if ((unsigned long)skb_aligned->data & 3)
        {
            skb_reserve(skb_aligned, 4 - ((unsigned long)skb_aligned->data & 3));
/*              printk("skb_aligned->data(2) = 0x%08X\n", skb_aligned->data); */
        }
        memcpy(skb_aligned->data, skb->data, skb->len);
        dev_kfree_skb(skb);
        skb = skb_aligned;
        skb_aligned = 0;
        buf = skb->data;
    }
#endif
/*      printk("Tx-Send: skb=0x%08lX\n", (unsigned long)skb); */
#ifdef MULTIPLE_TX_ATTEMPTS		
	/* perform TX_MAX_ATTEMPTS to send the data */
    while( (retry_count < TX_MAX_ATTEMPTS) &&
		   incaDMA_Transmit(INCAIP_SW_DMACHAN, buf, length, 1, (unsigned long)skb ) )
	{
		/* stop the queue to cause backpressure towards net-layer */
		if( !netif_queue_stopped(dev) )
			netif_stop_queue(dev);  
		retry_count++;        
    }
    /* restart the queue if necessary */
    if( netif_queue_stopped(dev) )
        netif_wake_queue(dev);  		
	
    if(retry_count >= TX_MAX_ATTEMPTS)
    {
        return -1; /* return error if all attempts failed */
    }
#else
	
    if( incaDMA_Transmit(INCAIP_SW_DMACHAN, buf, length, 1, (unsigned long)skb ))  
	{
        return -1;
	}		
#endif

#if 0
    SW_READ_REG(0xbf1047f8, regValue);
	regValue&= 0x0002EEFF;
    if(regValue!=0x00000000)
	{

/*	    printk("SWITCH_IRNCR1: 0x%08X\n", regValue);  */
     
		if(regValue & RX_CPU_NO_BUFFERS)
		{
		    printk("RX_CPU_NO_BUFFERS\n");
		}
	    if(regValue & RX_CPU_NO_PB_NUMBERS)
		{
		    printk("RX_CPU_NO_PB_NUMBERS\n");
		}
	    if(regValue & RX_LAN_NO_BUFFERS)
		{
		    printk("RX_LAN_NO_BUFFERS\n");
		}		
	    if(regValue & RX_LAN_NO_PB_NUMBERS)
		{
		    printk("RX_LAN_NO_PB_NUMBERS\n");
		}		
	    if(regValue & RX_PC_NO_BUFFERS)
		{
		    printk("RX_PC_NO_BUFFERS\n");
		}		
	    if(regValue & RX_PC_NO_PB_NUMBERS)
		{
		    printk("RX_PC_NO_PB_NUMBERS\n");
		}
	    if(regValue & MAC_TABLE_CHANGE)
		{
		    printk("MAC_TABLE_CHANGE\n");
            sysSwMacShowTable();
		}
	    if(regValue & GBL_RX_WM3)
		{
		    printk("GBL_RX_WM3\n");
		}				
		SW_WRITE_REG(0xbf1047f8, regValue);
	}
#endif
    return 0;
}

int inca_switch_control_init(void);
static void chipset_init(struct net_device *dev, int startp)
{
    sw21553_MacTableEntry tempMacTableEntry;

#ifdef CONFIG_INCAIP_SWITCH_CONTROL
    /* reset switch */
    inca_switch_control_init();
#endif
    
    /* Make sure the CPU port is disabled for now. We
     * don't want packets to get stacked for us until
     * we enable DMA and are prepared to receive them.
     * startp == 0: disable forwarding
     * startp != 0: enable forwarding
     */
    if (startp)
    {
        sw21553_SwUpdSpPtPortState(INCA_IP_SWITCH_PORT_SELECT_CPU, 
                                   INCA_IP_Switch_SP_PT_STATE_FORWARD);
    }
    else
    {
        sw21553_SwUpdSpPtPortState(INCA_IP_SWITCH_PORT_SELECT_CPU, 
                                   INCA_IP_Switch_SP_PT_STATE_DISABLED);
    }

/*    set the own MAC address into the switch MAC table
      --> setup the own MAC address into MAC table manually
          (no learning)
*/    
    memcpy(tempMacTableEntry.MacAddr, dev->dev_addr, 6);
    sw21553_SwMacSetTableEntryCntrl( INCA_IP_SWITCH_PORT_SELECT_CPU_V,   /* CPU voice port */
                                     1,                                  /* critical */
                                     0,                                  /* static entry */ 
                                     &tempMacTableEntry.Cntrl);
   /* add own MAC address, overwrite existing */
   sw21553_SwMacAddEntry(tempMacTableEntry, 1) ;
    
    /* enable CRC GEN. */
    sw21553_SwIoSetControl(eSwIoUpdArlCtl_CG, (void *)1) ;
    
    /* disable ADD TAG */
    sw21553_SwIoSetControl(eSwIoUpdMacHdCtl_ADD, (void *)0) ;
	
}

static unsigned char my_ethaddr[MAX_ADDR_LEN];
/* need to get the ether addr from u-boot */
static int __init ethaddr_setup(char *line)
{
	char *ep;
	int i;

	memset(my_ethaddr, 0, MAX_ADDR_LEN);
	/* there should really be routines to do this stuff */
	for (i = 0; i < 6; i++)
	{
		my_ethaddr[i] = line ? simple_strtoul(line, &ep, 16) : 0;
		if (line)
			line = (*ep) ? ep+1 : ep;
	}
	return 0;
}
__setup("ethaddr=", ethaddr_setup);

/*
 * The INCAIP Switch is always there - it's in the CPU.
 */
int __init 
incaipsw_probe(struct net_device *dev)
{
	struct incaipsw_local *np;
	static unsigned version_printed;
	int i, retval;
	static int init_done = 0;

	/* otherwise we return success for all possible interfaces! */
	if (init_done != 0)
		return 1;

	SET_MODULE_OWNER(dev);

	if (incaipsw_debug  &&  version_printed++ == 0)
		printk(KERN_DEBUG "%s", version);

	printk(KERN_INFO "%s: %s, ", dev->name, cardname);

	/* this should be handled by ethaddr_setup() */
	/* TEST */
	retval = 0;
	for (i = 0; i < 6; i++)
		retval += my_ethaddr[i];
	/* ethaddr not set in u-boot ? */
	if (retval == 0)
	{
		dev->dev_addr[0] = 0x00;
		dev->dev_addr[1] = 0x20;
		dev->dev_addr[2] = 0xda;
		dev->dev_addr[3] = 0x86;
		dev->dev_addr[4] = 0x23;
		dev->dev_addr[5] = 0x74;
	}
	else
		for (i = 0; i < 6; i++)
			dev->dev_addr[i] = my_ethaddr[i];
	/* print it out */
	for (i = 0; i < 6; i++)
		printk(" %2.2x", dev->dev_addr[i]);
	printk("\n");

	dev->irq = INT_NUM_IM1_IRL7;

	incaipsw_dma_init();

	/* let the DMA driver allocate space for us */
	incaDMA_InitRxChannel(INCAIP_SW_DMACHAN, NUM_RX_DESC, PKTSIZE_ALIGN, dev,
		incaip_sw_allocbuf, incaip_sw_rcv);
	incaDMA_InitTxChannel(INCAIP_SW_DMACHAN, NUM_TX_DESC, 3, dev, incaip_sw_txc);

    
    sw21553_PhyInit();   
#ifdef CONFIG_INCAIP_SWITCH_AMDIX
    sw21553_AmdixInit();
#endif
    
	/* Set MAC address.
	 */
        sw21553_SwSetPMacMacAddr(dev->dev_addr);

	/* enable spanning tree forwarding, enable the CPU port */
	/* ST_PT:
	      CPS (CPU port status)   0x3 (forwarding)
	      LPS (LAN port status)   0x3 (forwarding)
	      PPS (PC port status)    0x3 (forwarding)
	*/
/*  	SW_WRITE_REG(INCA_IP_Switch_ST_PT,0x3f); */
    sw21553_SwUpdSpPtPortState(INCA_IP_SWITCH_PORT_SELECT_CPU, 
                               INCA_IP_Switch_SP_PT_STATE_FORWARD);

	/* Do not forward unknown packets to LAN and PC */
	//SW_WRITE_REG(INCA_IP_Switch_UN_DEST,0x01fc);

	/* Initialize the device structure. */
	if (dev->priv == NULL) {
		dev->priv = kmalloc(sizeof(struct incaipsw_local), GFP_KERNEL);
		if (dev->priv == NULL)
			return -ENOMEM;
	}

	memset(dev->priv, 0, sizeof(struct incaipsw_local));

	np = (struct incaipsw_local *)dev->priv;
	spin_lock_init(&np->lock);

#if 0 /* needed? */
	/* Grab the region so that no one else tries to probe our ioports. */
	request_region(ioaddr, NETCARD_IO_EXTENT, cardname);
#endif

	dev->open	= incaipsw_open;
	dev->stop	= incaipsw_close;
	dev->hard_start_xmit = incaipsw_send_packet;
	dev->get_stats	= incaipsw_get_stats;
	dev->set_multicast_list = incaipsw_set_multicast_list;
 	dev->set_mac_address = incaipsw_set_mac_address;
     dev->tx_timeout = incaipsw_tx_timeout;
     dev->watchdog_timeo	= MY_TX_TIMEOUT; 

	/* Fill in the fields of the device structure with ethernet values. */
	ether_setup(dev);

	init_done = 1;

	return 0;
}

static void incaipsw_tx_timeout(struct net_device *dev)
{
	struct incaipsw_local *np = (struct incaipsw_local *)dev->priv;

	printk(KERN_WARNING "%s: transmit timed out, %s?\n", dev->name,
	       tx_done(dev) ? "IRQ conflict" : "network cable problem");

	/* Try to restart the adaptor. */
	chipset_init(dev, 0);

	np->stats.tx_errors++;

	/* If we have space available to accept new transmit
	 * requests, wake up the queueing layer.  This would
	 * be the case if the chipset_init() call above just
	 * flushes out the tx queue and empties it.
	 *
	 * If instead, the tx queue is retained then the
	 * netif_wake_queue() call should be placed in the
	 * TX completion interrupt handler of the driver instead
	 * of here.
	 */
	if (!tx_full(dev))
		netif_wake_queue(dev);
}

/*
 * Open/initialize the board. This is called (in the current kernel)
 * sometime after booting when the 'ifconfig' program is run.
 *
 * This routine should set everything up anew at each open, even
 * registers that "should" only need to be set once at boot, so that
 * there is non-reboot way to recover if something goes wrong.
 */
static int
incaipsw_open(struct net_device *dev)
{
	struct incaipsw_local *np = (struct incaipsw_local *)dev->priv;
#ifdef INCAIPSW_DEBUG
printk("%s called\n", __FUNCTION__);
#endif

	/*
	 * This is used if the interrupt line can turned off (shared).
	 * See 3c503.c for an example of selecting the IRQ at config-time.
	 */
	if (request_irq(dev->irq, &incaipsw_interrupt, 0, cardname, dev)) {
		return -EAGAIN;
	}


    
	/* Reset the hardware here. Don't forget to set the station address. */
	chipset_init(dev, 1);
	np->open_time = jiffies;

    sw21553_PhyStart();    
    
	/* We are now ready to accept transmit requeusts from
	 * the queueing layer of the networking.
	 */
	netif_start_queue(dev);

	return 0;
}

/* This will only be invoked if your driver is _not_ in XOFF state.
 * What this means is that you need not check it, and that this
 * invariant will hold if you make sure that the netif_*_queue()
 * calls are done at the proper times.
 */
static int incaipsw_send_packet(struct sk_buff *skb, struct net_device *dev)
{
	struct incaipsw_local *np = (struct incaipsw_local *)dev->priv;
	short length = ETH_ZLEN < skb->len ? skb->len : ETH_ZLEN;
	unsigned char *buf = (unsigned char *)KSEG1ADDR(skb->data);
#ifdef INCAIP_SW_TXDEBUG
	struct tcphdr *th;
#endif
#ifdef INCAIPSW_TXDEBUG
	int command;
#endif

#ifdef INCAIPSW_DEBUG
        printk("%s called\n", __FUNCTION__);
#endif
	/* If some error occurs while trying to transmit this
	 * packet, you should return '1' from this function.
	 * In such a case you _may not_ do anything to the
	 * SKB, it is still owned by the network queueing
	 * layer when an error is returned.  This means you
	 * may not modify any SKB fields, you may not free
	 * the SKB, etc.
	 */

	/* This is the case for older hardware which takes
	 * a single transmit buffer at a time, and it is
	 * just written to the device via PIO.
	 *
	 * No spin locking is needed since there is no TX complete
	 * event.  If by chance your card does have a TX complete
	 * hardware IRQ then you may need to utilize np->lock here.
	 */
#ifdef INCAIP_SW_TXDEBUG
        if (skb->nh.iph->protocol == 6)
        {
            printk("6 ");
            printk("Sack: %#x ", skb->h.th->ack);
            printk("Spsh: %#x ", skb->h.th->psh);
            printk("Srst: %#x ", skb->h.th->rst);
            printk("Ssyn: %#x ", skb->h.th->syn);
            printk("Sfin: %#x ", skb->h.th->fin);
            printk("\n");
        }
#endif
#ifdef INCAIPSW_TXDEBUG
        printk("TX: ");
        for (command = 0; command < skb->len; command++)
        {
            printk("0x%02X ", buf[command]);
            if(((command+1)%16) == 0)printk("\n");
        }
        printk("\n");
#endif
#ifndef CONFIG_MIPS_UNCACHED
	dma_cache_wback_inv((unsigned long)skb->data, length);
#endif
	if( hardware_send_packet(dev, buf, length, skb) != 0 )
	{
	    /* de6wmigr: To be clarified which value to be passed 
		   to upper layer in case of error */
		return 1;	
	}
	np->stats.tx_bytes += skb->len;
	np->stats.tx_packets++;

	dev->trans_start = jiffies;

#ifdef INCA_DMA_USE_COPYING
	dev_kfree_skb(skb);
#endif

	return 0;
}

/*
 * The typical workload of the driver:
 * Handle the network interface interrupts.
 *
 * With this hardware all the real work is handled by the DMA driver,
 * so there;s nothing to do here.
 *
 */
static void incaipsw_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
#ifdef INCAIPSW_DEBUG
	printk("%s\n", __FUNCTION__);
#endif
	return;
}

/* The inverse routine to incaipsw_open(). */
static int
incaipsw_close(struct net_device *dev)
{
	struct incaipsw_local *lp = (struct incaipsw_local *)dev->priv;

	lp->open_time = 0;

	netif_stop_queue(dev);

	/* Flush the Tx and disable Rx here. */

	/* Disable forwarding to the CPU port.
	 */
        sw21553_SwUpdSpPtPortState(INCA_IP_SWITCH_PORT_SELECT_CPU, 
                                   INCA_IP_Switch_SP_PT_STATE_DISABLED);

	free_irq(dev->irq, dev);

	/* Update the statistics here. */

	return 0;

}

/*
 * Get the current statistics.
 * This may be called with the card open or closed.
 */
static struct net_device_stats *incaipsw_get_stats(struct net_device *dev)
{
	struct incaipsw_local *lp = (struct incaipsw_local *)dev->priv;

	lp->stats.rx_dropped=sw21553_Read_CPU_MIB_RxDropped();
	return &lp->stats;
}

int incaipsw_set_mac_address(struct net_device *dev, void *addr)
{
	struct sockaddr *saddr=addr;
	struct sw21553_mac_table_entry mac_entry;

	if (dev->flags & IFF_UP) /* can't act on a running interface */
		return -EBUSY;

	// Copy MAC address to device structure
	memcpy(dev->dev_addr, saddr->sa_data, ETH_ALEN);
	/* Copy MAC address to MAC entry structure*/
	memcpy(mac_entry.MacAddr, dev->dev_addr, 6);
	/* Clear all entries */
	mac_entry.Cntrl=0;
	/* Delete entry from table */
	sw21553_SwMacClearEntryByMac(mac_entry);
	/* Set control value for entry (CPU-V port, critical, static) */
	sw21553_SwMacSetTableEntryCntrl(INCA_IP_SWITCH_PORT_SELECT_CPU_V,1,0,
                                     &mac_entry.Cntrl);
	/* Add new MAC address */
	sw21553_SwMacAddEntry(mac_entry, 1) ;
	sw21553_SwSetPMacMacAddr(dev->dev_addr);

	return 0;
}

/*
 * Set or clear the multicast filter for this adaptor.
 * num_addrs == -1	Promiscuous mode, receive all packets
 * num_addrs == 0	Normal mode, clear multicast list
 * num_addrs > 0	Multicast mode, receive normal and MC packets,
 *			and do best-effort filtering.
 */
#define HW_MAX_ADDRS 255
static void incaipsw_set_multicast_list(struct net_device *dev)
{
	struct dev_mc_list *mc_ptr;
	struct sw21553_mac_table_entry mac_entry;

	if (dev->flags&IFF_PROMISC) {
		/* Enable promiscuous mode */
		sw21553_SwUpdUnDestPorts(INCA_IP_SWITCH_PORT_SELECT_CPU, \
					SWITCH_UNICAST | SWITCH_MULTICAST | SWITCH_BROADCAST);
		/* Disable Learning for PC and LAN ports */
		sw21553_SwSetArlControl(0,INCA_IP_Switch_ARL_CTL_ENL|0x1c);

	}
	else if((dev->flags&IFF_ALLMULTI) || dev->mc_count > HW_MAX_ADDRS)
	{
		/* Disable promiscuous mode, use normal mode. */
		sw21553_SwUpdUnDestPorts(INCA_IP_SWITCH_PORT_SELECT_CPU, \
					SWITCH_MULTICAST | SWITCH_BROADCAST);
		/* Enable Learning for PC and LAN ports */
		sw21553_SwSetArlControl(0x5,INCA_IP_Switch_ARL_CTL_ENL|0x1c);
	}
	else
	{
		/* Only enable broadcast frames */
		sw21553_SwUpdUnDestPorts(INCA_IP_SWITCH_PORT_SELECT_CPU, \
							SWITCH_BROADCAST);
   
		/* Enable Learning for PC and LAN ports */
		sw21553_SwSetArlControl(0x5,INCA_IP_Switch_ARL_CTL_ENL|0x1c);
		/* Clear MAC table */
		sw21553_SwMacResetMacTable();
		/* Add own MAC address */
	     memcpy(mac_entry.MacAddr, dev->dev_addr, 6);
		sw21553_SwMacSetTableEntryCntrl( INCA_IP_SWITCH_PORT_SELECT_CPU_V,   /* CPU voice port */
                                     1,                                  /* critical */
                                     0,                                  /* static entry */ 
                                     &mac_entry.Cntrl);
		sw21553_SwMacAddEntry(mac_entry, 1) ;
		sw21553_SwSetPMacMacAddr(dev->dev_addr);

		if(dev->mc_count) {
			mac_entry.Cntrl=0;
			/* Walk the address list, and load the filter */
			for (mc_ptr=dev->mc_list;mc_ptr;mc_ptr=mc_ptr->next) {
				memcpy(mac_entry.MacAddr,mc_ptr->dmi_addr,6);
				sw21553_SwMacSetTableEntryCntrl( INCA_IP_SWITCH_PORT_SELECT_CPU_V,   /* CPU voice port */
									  1,                                  /* critical */
									  0,                                  /* static entry */ 
									  &mac_entry.Cntrl);
				sw21553_SwMacAddEntry(mac_entry,1);
			}
		}
	}

}

static unsigned long incaip_sw_allocbuf(void *vdev, unsigned long *info)
{
    struct sk_buff *skb;
	struct net_device *dev = (struct net_device *)vdev;
#ifdef INCAIPSW_DEBUG
	printk("%s called\n", __FUNCTION__);
#endif

	skb = dev_alloc_skb(PKTSIZE_ALIGN + 4);
	if (skb == NULL) {
            printk(KERN_NOTICE "%s: Memory squeeze, dropping packet.\n",
                   dev->name);
            return 0;
	}

        *info = (unsigned long) skb;

        if ((unsigned long)skb->data & 3)
        {
            printk("incaip_sw_allocbuf: unaligned data pointer\n");
            return 0;
        }
/*          skb_reserve(skb, 2); */

/*          printk("RX: alloc skb=0x%08lx\n", (unsigned long)skb); */

	return (unsigned long)skb->data;
}

static unsigned long incaip_sw_rcv(void *vdev, void *buf, unsigned long size, unsigned long info)
{
	/* Malloc up new buffer. */
	struct sk_buff *skb;
	struct net_device *dev = (struct net_device *)vdev;
	struct incaipsw_local *lp = (struct incaipsw_local *)dev->priv;
#ifdef INCA_DMA_USE_COPYING
#ifndef CONFIG_MIPS_UNCACHED
	unsigned char *lbuf = (unsigned char *)buf;
#else
	unsigned char *lbuf = (unsigned char *)KSEG1ADDR(buf);
#endif
	unsigned char *ldat;
#endif /* INCA_DMA_USE_COPYING */

#ifdef INCASW_RX_DEBUG
	int command;
#endif

#ifdef INCAIPSW_DEBUG
	printk("%s called\n", __FUNCTION__);
#endif

	lp->stats.rx_bytes+=size;
	
#ifdef INCA_DMA_USE_COPYING
	skb = dev_alloc_skb(size + 2);	/* +2 to allow for alignment */
	if (skb == NULL) {
		printk(KERN_NOTICE "%s: Memory squeeze, dropping packet.\n",
			   dev->name);
		lp->stats.rx_dropped++;
		return -ENOMEM;
	}

	/* Make sure IP header is aligned on 16 byte boundary */
	skb_reserve(skb, 2);
#else
        skb = (struct sk_buff *)info;
/*          printk("RX: skb=0x%08lx\n", info); */
        if (skb->data != buf)
            printk("incaip_sw_rcv: buf(%08lX) does not match skb->data(0x%08lx)\n",
                   (unsigned long) buf, (unsigned long)skb->data);
#endif

	skb->dev = dev;

	skb_put(skb, size);

#ifdef INCA_DMA_USE_COPYING
	ldat = skb->data;
	/* 'skb->data' points to the start of sk_buff data area. */
	memcpy(ldat, lbuf, size);
#endif

#ifdef INCASW_RX_DEBUG
printk("RX: (0x%08lX)\n", (unsigned long)&skb->data[0]);
for (command = 0; command < size; command++)
{
/*  printk("0x%02X ", lbuf[command]); */
printk("0x%02X ", skb->data[command]);
if(((command+1)%16) == 0)printk("\n");
}
printk("\n");
#endif

        skb->protocol=eth_type_trans(skb,dev);
	skb->pkt_type = PACKET_HOST;
#ifdef INCAIP_SW_RXDEBUG
if (ldat[14] == 0x45 && ldat[23] == 6)
{
printk("6 ");
printk("Rack: %#x ", ((struct tcphdr *)&ldat[34])->ack);
printk("Rpsh: %#x ", ((struct tcphdr *)&ldat[34])->psh);
printk("Rrst: %#x ", ((struct tcphdr *)&ldat[34])->rst);
printk("Rsyn: %#x ", ((struct tcphdr *)&ldat[34])->syn);
printk("Rfin: %#x ", ((struct tcphdr *)&ldat[34])->fin);
printk("\n");
for (command = 0; command < size; command++)
{
printk("0x%02X ", ldat[command]);
if(((command+1)%16) == 0)printk("\n");
}
printk("\n");
}
#endif
	netif_rx(skb);
	dev->last_rx = jiffies;
	lp->stats.rx_packets++;
	return 0;
}

static unsigned long incaip_sw_txc(void *vdev, unsigned long info)
{
    struct net_device *dev = (struct net_device *)vdev;
    struct sk_buff *skb = (struct sk_buff *)info;

#ifdef INCAIPSW_DEBUG
	printk("%s called\n", __FUNCTION__);
#endif

#ifndef INCA_DMA_USE_COPYING
        if (skb)
            dev_kfree_skb(skb);
#endif

	return 0;
}


#ifdef MODULE

static struct net_device this_device;
MODULE_LICENSE("GPL");

int init_module(void)
{
	int result;

	/* the INCAIP switch is always there, so really no need to probe */
	this_device.init      = incaipsw_probe;

	if ((result = register_netdev(&this_device)) != 0)
		return result;

	return 0;
}

void
cleanup_module(void)
{
    /* Disable forwarding to the CPU port.
     */
/*  	SW_WRITE_REG(INCA_IP_Switch_ST_PT,0xf); */
    sw21553_SwUpdSpPtPortState(INCA_IP_SWITCH_PORT_SELECT_CPU, 
                               INCA_IP_Switch_SP_PT_STATE_DISABLED);

	/* No need to check MOD_IN_USE, as sys_delete_module() checks. */
	unregister_netdev(&this_device);

	/*
	 * If we don't do this, we can't re-insmod it later.
	 * Release irq here, when you have jumpered versions and
	 * allocate them in incaipsw_probe().
	 */
        /* irq already released in incaipsw_close */
/*  	free_irq(this_device.irq, dev); */

	/* Close DMA channels.
	 */
	incaDMA_CloseRxChannel(0);
	incaDMA_CloseTxChannel(0);

	if (this_device.priv)
		kfree(this_device.priv);
}

#endif /* MODULE */
