/*
 * incaip_ssc.h
 *
 * Common defines/structures which are used both in the kernel and potentially
 * by applications.
 *
 * Copyright (c) 2003 Gary Jennejohn <gj@denx.de>
 *
 */
#ifndef INCAIP_SSC_H
#define INCAIP_SSC_H

/* various error statistics */

struct incaip_ssc_statistics
{
	unsigned int  rxErr;      /* RX error */
	unsigned int  phErr;      /* Phase error */       
	unsigned int  brErr;      /* Baudrate error */       
	unsigned int  txErr;      /* TX error */       
	unsigned int  txOvErr;    /* TX Overflow error */         
	unsigned int  rxOvErr;    /* RX Overflow error */         
	unsigned int  rxUnErr;    /* RX Underrun error */
	unsigned int  rxBytes;
	unsigned int  txBytes;
};

/* Macros to configure SSC hardware */
/* loopBack: */
#define SSC_LOOPBACK_OFF         0
#define SSC_LOOPBACK_ON          1
/* clockPolarity: */
#define SSC_IDLE_CLOCK_LOW       0
#define SSC_IDLE_CLOCK_HIGH      1
/* clockPhase: */
#define SSC_TX_ON_LEADING_EDGE   0
#define SSC_TX_ON_TRAILING_EDGE  1
/* headingControl: */
#define SSC_LSB_FIRST            0
#define SSC_MSB_FIRST            1
/* dataWidth: */
#define SSC_MIN_DATA_WIDTH       2
#define SSC_MAX_DATA_WIDTH       16
/* master/slave mode select */
#define SSC_MASTER_MODE          1
#define SSC_SLAVE_MODE           0
/* full/half duplex mode select */
#define SSC_FULL_DUPLEX          0
#define SSC_HALF_DUPLEX          1

/* default values for SSC configuration */
#define SSC_DEFAULT_AUTO_RESET      0 /* disable */
#define SSC_DEFAULT_BR_ERR_DETECT   1 /* enable */
#define SSC_DEFAULT_PH_ERR_DETECT   0 /* disable */
#define SSC_DEFAULT_RX_ERR_DETECT   1 /* enable */
#define SSC_DEFAULT_TX_ERR_DETECT   0 /* disable */
#define SSC_DEFAULT_LOOP_BACK       SSC_LOOPBACK_OFF
#define SSC_DEFAULT_CLOCK_POLARITY  SSC_IDLE_CLOCK_HIGH
#define SSC_DEFAULT_CLOCK_PHASE     SSC_TX_ON_TRAILING_EDGE
#define SSC_DEFAULT_HEADING_CONTROL SSC_MSB_FIRST
#define SSC_DEFAULT_DATA_WIDTH      16 /* bits */
#define SSC_DEFAULT_DUPLEX_MODE     SSC_FULL_DUPLEX
#define SSC_DEFAULT_USE_FIFO	    1 /* enable */
#define SSC_DEFAULT_BAUDRATE	    1000000

struct incaip_ssc_hwopts {
	unsigned int  brErrDetect    :1;   /* Baudrate Error detection */
	unsigned int  phErrDetect    :1;   /* Phase Error detection */
	unsigned int  rxErrDetect    :1;   /* Receive Error detection */
	unsigned int  txErrDetect    :1;   /* Transmit Error detection */
	unsigned int  autoReset      :1;   /* Automatic reset on a baudrate
								      error */
	unsigned int  loopBack       :1;   /* Loopback mode */
	unsigned int  clockPolarity  :1;   /* Idle clock is high or low */
	unsigned int  clockPhase     :1;   /* Tx on trailing or leading edge*/
	unsigned int  headingControl :1;   /* LSB first or MSB first */
	unsigned int  dataWidth      :5;   /* from 2 up to 16 bits */
	unsigned int  masterSelect   :1;   /* Master or Slave mode */
	unsigned int  duplexMode     :1;   /* Full or half duplex mode */
	unsigned int  useFIFO	     :1;   /* whether or not to use the FIFO */
};

#ifdef __KERNEL__

/* kernel interface for fops, read and write */
extern int incaip_ssc_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
extern int incaip_ssc_open(struct inode *, struct file *);
extern int incaip_ssc_close(struct inode *, struct file *);
extern ssize_t incaip_ssc_kernel_read(int, char *, size_t);
extern ssize_t incaip_ssc_kernel_write(int, const char *, size_t);

/* the register layout of an SSC */
struct incaip_ssc_regs
{
	volatile unsigned long  ssc_clc;        /*0x0000*/
	volatile unsigned long  ssc_rsvd1[3];   /* for mapping */   
	volatile unsigned long  ssc_con;        /*0x0010*/
	volatile unsigned long  ssc_br;         /*0x0014*/
	volatile unsigned long  ssc_rsvd2[2];   /* for mapping */   
	volatile unsigned long  ssc_tbuf;       /*0x0020*/
	volatile unsigned long  ssc_rbuf;       /*0x0024*/
	volatile unsigned long  ssc_rsvd3[2];   /* for mapping */   
	volatile unsigned long  ssc_rxfcon;     /*0x0030*/
	volatile unsigned long  ssc_txfcon;     /*0x0034*/
	volatile unsigned long  ssc_fstat;      /*0x0038*/
	volatile unsigned long  ssc_rsvd4;      /* for mapping */   
	volatile unsigned long  ssc_whbcon;
};

/* fill levels, masks and shifts for the FIFOs */
#define INCAIP_SSC_TXFIFO_FULL 16
#define INCAIP_SSC_TXFIFO_SHIFT 8
#define INCAIP_SSC_TXFIFO_MASK 0x3f00
#define INCAIP_SSC_TXFIFO_FL 8
#define INCAIP_SSC_RXFIFO_MASK 0x3f
#define INCAIP_SSC_RXFIFO_FL 1

struct incaip_ssc_port
{
	struct incaip_ssc_regs *regs;
	struct incaip_ssc_hwopts opts;
	struct incaip_ssc_statistics stats;
	/* wait queue for incaip_ssc_read() */
	wait_queue_head_t rwait, pwait;
	char port_is_open; /* exclusive open  - boolean */
	char dma_active;
	int elem_size; /* shift for element */
	/* buffer and pointers to the read/write position */
#define RX_BUFSIZE 5000
#define TX_BUFSIZE 5000
	unsigned short *rxbuf; /* buffer for RX */
	unsigned short *rxrp, *rxwp; /* pointers to read/write pos. */
	unsigned short *txbuf; /* buffer for TX */
	unsigned short *txrp, *txwp; /* pointers to read/write pos. */
	unsigned int baud;
	/* each channel has its own interrupts */
	/* transmit/receive/error interrupt bit */
	unsigned int txirq, rxirq, erirq;
	/* and its own bit to enable MRST */
	unsigned int mrst;
	/* a workaround for slave mode to send all 1's  - boolean */
	char disable_tx;
	/* round index to verify that transmit buffer is not filled without  */
	/* receive buffer is taken when the hardware FIFOs are not used */
	unsigned long runIdxTxIrq;
	unsigned long runIdxRxIrq;
	/* boolean - disable the transmitter */
	volatile char bDisTransmit;
	/* boolean - state of the receive interrupt (enabled/disabled) */
	volatile char bRxIrqEnabled;
#ifdef INCAIP_SSC_USEDMA
	char dma_freecnt; /* how many DMA channels are free? Only SSC1! */
#endif
};
#define SSC_IS_MASTER(p) ((p)->opts.masterSelect == SSC_MASTER_MODE)
/*
 * protect against inadvertant attempts to flush the non-existent
 * FIFOs on SSC1 -> results in a bus error! This makes sense because
 * that's an access to a non-existent address.
 */
#define SSC_USE_FIFO(p) ((p) == &isp[0] && (p)->opts.useFIFO == 1)
#endif /* __KERNEL__ */
 
/* defines for ioctl's */
/* read out the statistics */
#define SSC_STATS_READ _IOR('S', 1, struct incaip_ssc_statistics)
/* clear the statistics */
#define SSC_STATS_RESET _IO('S', 2)
/* set the buadrate */
#define SSC_SET_BAUD _IOW('S', 3, unsigned int)
/* get the current baudrate */
#define SSC_GET_BAUD _IOR('S', 4, unsigned int)
/* set hardware options */
#define SSC_SET_HWOPTS _IOW('S', 5, struct incaip_ssc_hwopts)
/* get the current hardware options */
#define SSC_GET_HWOPTS _IOR('S', 6, struct incaip_ssc_hwopts)
/* find out whether the driver supports DMA */
#define SSC_GET_DMA _IOR('S', 7, int)
/* get number of bytes in the write FIFO */
/* although there is a FIONREAD, there is no FIONWRITE */
#define SSC_FIONWRITE _IOR('S', 8, int)
/* flush both FIFOs */
#define SSC_FIFOFLUSH _IO('S', 9)
/* flush only the read FIFO */
#define SSC_FIFORFLUSH _IO('S', 10)
/* flush only the write FIFO */
#define SSC_FIFOWFLUSH _IO('S', 11)

#endif /* INCAIP_SSC_H */

